/*
 *comm.c - communication software.
 */

#include <parallel.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "comm.h"

/*Preprocessed messages*/
char mesg_end[ 10 ];
char mesg_fire[ 10 ];
char mesg_stopped[ 10 ];
char mesg_eo_stopped[ 10 ];     /*current EO is stopped*/
char mesg_na[ 10 ];		/*There is no agenda activations*/

/*
 *StartEOs - start up all of the EOs
 */
StartEOs()
{
    extern int Start_tasks();
    int ct = 0;
    struct defeo *eo;
    struct defeo *eo_in;
    struct defrs *rs;
    struct eo_id *eos;
    struct eo_id *out;

    for (eo=imgr.defeo_lst; eo != NULL; eo = eo->next) {
	ct++;

	for (eos=eo->out_boxes;eos != NULL; eos = eos->next) {

	    if ((eo_in = GetEO_by_name(eos->name)) == NULL) {
		sprintf(error_sb, "Expert object %s missing.", eos->name);
		EH(NULLSCANNER, error_sb);
		return(NO);
		} /*if*/

	    eos->id = eo_in->id;
	    eos->sema = eo_in->recv;
	    }/*for*/

	for (eos = eo->in_boxes; eos != NULL; eos = eos->next) {

	    if ((eo_in = GetEO_by_name(eos->name)) == NULL) {
		sprintf(error_sb, "Expert object %s missing.", eos->name);
		EH(NULLSCANNER, error_sb);
		return(NO);
		} /*if*/

	    for (out=eo_in->out_boxes;out != NULL;out = out->next) 
		if (out->id == eo->id)
		    break;

	    if (out == NULL) {
		EH(NULLSCANNER, "Missing out box.");
		return(NO);
		}

	    eos->box = out->box;
	    eos->id = eo_in->id;
	    } /*for*/

	    
	for (rs=eo->defrs_locals; rs != NULL; rs = rs->next)
	    ct++;

	} /*for*/

    imgr.barrier = barrier_create(ct, TASK_BLOCK);

    printf("Starting %d processes\n", ct);

    task_init( 200000, ct, Start_tasks, 1024, 0);

} /*StartEOs*/

/*
 *Start_tasks
 */
Start_tasks()
{
    struct defeo *eo;
    struct defrs *rs;
    extern void EO_Exec();

    semaphore_init(imgr.print, TASK_BLOCK);
    semaphore_signal(imgr.print);

    for (eo=imgr.defeo_lst; eo != NULL; eo = eo->next) { /*for*/

	for(rs=eo->defrs_locals; rs != NULL; rs = rs->next)
	    rs->eo = eo;

	if (task_start(EO_Exec, 4096, 1, eo->id) == NULL) { 
	    EH(NULLSCANNER, "task_start failure ... se: quit");
	    exit(1);
	    } /*if*/

	} /*for*/

} /*Start_tasks*/

/*
 *EO_MailRSs
 */
void EO_MailRSs(rs, mesg)
struct defrs *rs;
char *mesg;
{
    for (;rs != NULL; rs = rs->next) 
	SendMesg(rs->in_box, mesg, rs->in_sema, "EO to RS");

} /*EO_MailRSs*/

/*
 *RS_ProcessMessage
 */
RS_ProcessMessage(rs, in, mesg)
struct defrs *rs;
struct mailbox *in;
char *mesg;
{
    int type;
    int id;
    char token[ MAX_MESG_LEN ];
    struct fact *fact;

    sscanf(mesg, "%d", &type);

    switch (type) {

	case MESG_STOPPED :
		in->status = MESG_EO_STOPPED;
		rs->status = EO_STOPPED;

		rs->out_box->status = MESG_EO_STOPPED;
		rs->in_box->status = MESG_EO_STOPPED;

		break;
			
	case MESG_END : 
		if (in->status != MESG_EO_STOPPED)
		    in->status = MESG_START;
		return(NO);
		break;
     
	case MESG_EO_STOPPED :
		in->status = MESG_EO_STOPPED;
		return(NO);
		break;

	case MESG_RETRACT :
		sscanf(mesg, "%*d %*d %d", &id);
		fact = GetFact(rs, id);
		RetractFlow(rs, fact);
		RemoveFact(rs, fact);
		break;

	case MESG_ASSERT :
		strcpy(token, strpbrk(mesg, "|") + 1);
		sscanf(mesg, "%*d %*d %d ", &id);
		Assert(rs, token, id);
		break;

	case MESG_FIRE :
		FireRule(rs); 
		break;

 	default :
		EH(&rs->in, "MESSAGE OUT OF RANGE IN RS");

	} /*switch*/

    return(YES);
 
} /*RS_ProcessMessage*/

/*
 *SetState_Boxes
 */
void SetState_Boxes(eo_boxes, state)
struct eo_id *eo_boxes;
int state;
{
    struct eo_id *boxes;

    boxes = eo_boxes;

    for (; eo_boxes != NULL; eo_boxes = eo_boxes->next) { 
	spin_lock(eo_boxes->box->lock);
	eo_boxes->box->status = state;
	spin_unlock(eo_boxes->box->lock);
	} /*for*/

} /*SetState_Boxes*/

/*
 *EO_ProcessMessage
 */
EO_ProcessMessage(eo, in, mesg)
struct defeo *eo;
struct mailbox *in;
char *mesg;
{
    char token[ MAX_MESG_LEN ];
    int add = NO;
    int eo_id;
    int rs_p, rule_p;
    int type;
    struct eo_id *boxes;

    semaphore_wait(imgr.print);
    printf("EO IS PROCESSING: <%s> \n", mesg);
    semaphore_signal(imgr.print);

    sscanf(mesg, "%d", &type);

    switch ((int)type) {  /*switch*/
	case MESG_RTF :
			sscanf(mesg, "%*d %d %d %s", &rs_p, &rule_p, token);

			if (eo->rs_box == NULL) 
			    add = YES;
	     else if ((eo->rs_priority == rs_p) && (eo->rule_priority < rule_p))
			    add = YES;
			else if (eo->rs_priority < rs_p)
			    add = YES;

			if (add == YES) {
			    eo->rs_box = in;
			    eo->rs_priority = rs_p;
			    eo->rule_priority = rule_p;
			    } /*if*/

			break;

	case MESG_NA :
			break;

	case MESG_STOPPED :
			in->status = MESG_EO_STOPPED;
			eo->status = EO_STOPPED;
			SetState_Boxes(eo->in_boxes, MESG_EO_STOPPED);
			SetState_Boxes(eo->out_boxes, MESG_EO_STOPPED);
			EO_MailRSs(eo->defrs_locals, mesg);
			break;

	case MESG_EO_STOPPED :
			in->status = MESG_EO_STOPPED;
			return(NO);
			break;
			
	case MESG_END :
			if (in->status != MESG_EO_STOPPED)
			    in->status = MESG_END;

			return(NO);
			break;

	case MESG_ASSERT :
	case MESG_RETRACT :
			sscanf(mesg, "%*d %d", &eo_id);
			if (eo_id == eo->id) 
			    EO_MailRSs(eo->defrs_locals, mesg);
		else for (boxes=eo->out_boxes;boxes != NULL; boxes=boxes->next)
		    	    if (eo_id == boxes->id) { 
				SendMesg(boxes->box, mesg, boxes->sema, "Forward to EO");
				break;
				} /*if*/

			break;

	default :
			EH(&eo->in, "EO: ProcessMessage out of range.");
			break;
	} /*switch*/


    return(YES);
 
} /*EO_ProcessMessage*/

/*
 *FireRule
 */
FireRule(rs)
struct defrs *rs;
{
    struct activation *temp;

    rs->eo->total_rules++;

    if (imgr.watch == YES) {

	semaphore_wait(imgr.print);
	printf("\n\n");
	printf("FIRE %4d %s", rs->Agenda.list->id, rs->Agenda.list->name);
	PrintIds(rs->Agenda.list->variables);
	printf("\n\n");
	semaphore_signal(imgr.print);

	} /*if*/

    /*Save the activation so it can be given back (deallocated)*/
    temp = rs->Agenda.list;

    rs->Agenda.list = rs->Agenda.list->next;

    ExecuteRHS(rs, temp->rhs->rhs->next, temp->variables);

    SendMesg(rs->out_box, mesg_end, rs->out_sema, "Rule has fired");

    ReturnAct(temp);

    return(YES);

} /*FireRule*/

/*
 *RecvMesg
 */
char *RecvMesg(box, sema, comment)
struct mailbox *box;
SEMAPHORE *sema;
char *comment;
{
    char mesg[ MAX_MESG_LEN ];

    semaphore_wait(sema);

    Retrieve(box, mesg);

    return(MakeString(mesg));

} /*RecvMesg*/

/*
 *SendMesg
 */
SendMesg(box, mesg, sema, comment)
struct mailbox *box;
char *mesg;
SEMAPHORE *sema;
char *comment;
{

    Deliver(box, mesg);

    semaphore_signal(sema);

} /*RS_SendMesg*/

/*
 *RemoveMessage
 */
char **RemoveMessage(buffer)
char **buffer;
{
    char **new, **tmp;
    int ct = 0;

    if (*buffer == NULL)
	return(buffer);

    new = (char **)MemAlloc(sizeof(char **) * BufSize(buffer));

    tmp = buffer + 1;

    while (*tmp != NULL)
	new[ct++] = *(tmp++);

    new[ct] = NULL;

    FREEMEM(buffer[0]); /*Give back the message*/
    FREEMEM(buffer);    /*Free its storage pointer*/

    return(new);

} /*RemoveMessage*/

/*
 *AddMessage
 */
char **AddMessage(buffer, lst)
char **buffer;
char *lst;
{
    char **new;
    int ct = 0;

    new = (char **)MemAlloc(sizeof(char **) * (BufSize(buffer) + 2));

    for (;*buffer != NULL; ++buffer)
	new[ct++] = *buffer;

    new[ct++] = lst;
    new[ct] = NULL;

    FREEMEM(buffer);

    return(new);

} /*AddMessage*/

/*
 *CreateMesg
 */
char *CreateMesg(pele)
struct element *pele;
{
    char mesg_create[ MAX_MESG_LEN ];
    int offset = 0;

    for (; pele != NULL ; pele = pele->next) {

	if (pele->type == WORD)
	    sprintf(mesg_create + offset, " %s", pele->string);
	else
	    sprintf(mesg_create + offset, " %6.2f", pele->numeric);

	for (;mesg_create[offset] != '\0';offset++);
	} /*for*/

    return(MakeString(mesg_create));

} /*CreateMesg*/

/*
 *CreateToken
 */
char *CreateToken(type, eo_id, string1, value1, value2)
int type;
char *string1;
int value1, value2;
{
    char mesg[ MAX_MESG_LEN ];

    switch (type) {

	case MESG_RETRACT : 
			sprintf(mesg, "%d %d %d", MESG_RETRACT, eo_id, value1);
			return(MakeString(mesg));
			break;

	case MESG_ASSERT :
		/*string1 is the fact string to be asserted*/
		sprintf(mesg, "%d %d %d | %s", MESG_ASSERT,eo_id,value1, string1);
		return(MakeString(mesg));
		break;

	case MESG_RTF :
		sprintf(mesg, "%d %d %d %s", MESG_RTF, value1, value2, string1);
		return(MakeString(mesg));
		break;

	default :
		EH(NULLSCANNER, "Case out of bounds in CreateToken");
		break;

	} /*switch*/

    return(NULL);

} /*CreateToken*/

/*
 *Deliver
 */
Deliver(box, mesg)
struct mailbox *box;
char *mesg;
{

    spin_lock(box->lock);

    if (box->status != MESG_EO_STOPPED)
	box->mesgs = AddMessage(box->mesgs, mesg);
    else
	FREEMEM(mesg);

    spin_unlock(box->lock);

    return(YES);

} /*Deliver*/

/*
 *Retrieve
 */
Retrieve(box, mesg)
struct mailbox *box;
char mesg[];
{

    spin_lock(box->lock);

    if ((box->status == MESG_EO_STOPPED) && (*box->mesgs == NULL))
	strcpy(mesg, mesg_eo_stopped);
    else {
	strcpy(mesg, box->mesgs[0]);
	box->mesgs = RemoveMessage(box->mesgs);
	} /*else*/

    spin_unlock(box->lock);

    return(YES);

} /*Retrieve*/

/*
 *IsMesgs
 */
IsMesgs(box)
struct mailbox *box;
{
    int cond = YES;

    spin_lock(box->lock);

    if (box->status != MESG_EO_STOPPED)
	if (*box->mesgs == NULL)
	    cond = NO;

    spin_unlock(box->lock);

    return(cond);

} /*IsMesgs*/

/*
 *RS_Exec
 */
void RS_Exec(id, name)
int id;
char *name;
{
    char *mesg;
    struct defrs *rs;
    struct defeo *eo;

    for (eo=imgr.defeo_lst; eo != NULL; eo = eo->next) 
	if (eo->id == id)
	    break;

    for (rs=eo->defrs_locals; rs != NULL; rs = rs->next)
	if (rs->name == name)
	     break;

    barrier(imgr.barrier);

    while (rs->status != EO_STOPPED) {

	/*First get all messages from the input box*/
	while ((mesg = RecvMesg(rs->in_box, rs->in_sema, \
					"To Rule Set from the EO")) != NULL)
	    if (RS_ProcessMessage(rs, rs->in_box, mesg) == NO)
		break;

	if (rs->status == EO_STOPPED)
	    break;

	/*Now see what is at the front of the rule set's agenda */
	if (rs->Agenda.list == NULL) 
	    SendMesg(rs->out_box, mesg_na, rs->out_sema, \
			"To: EO: from RS: front of agenda -> nothing");
	else  {
	    mesg = CreateToken(MESG_RTF, 0, rs->Agenda.list->name,rs->priority,\
				rs->Agenda.list->priority);
	    SendMesg(rs->out_box, mesg, rs->out_sema, \
				"To: EO: from RS: from RS rule ready to fire");
	    } /*else*/

	SendMesg(rs->out_box, mesg_end, rs->out_sema, "RS: Send End Mesg");

	while ((mesg = RecvMesg(rs->in_box, rs->in_sema, \
				"To: RS: from eo: Should I fire")) != NULL)
	    if (RS_ProcessMessage(rs, rs->in_box, mesg) == NO) 
		break;

	} /*while*/

    return(YES);

} /*RS_Exec*/

/*
 *EO_Exec
 */
void EO_Exec(id)
int id;
{
    char **facts;
    char *mesg;
    int ct;
    int done = YES;
    int go;
    int wait;
    struct defeo *eo;
    struct deffacts *stmt;
    struct defrs *rs, *fired;
    struct eo_id *eos;
    struct mailbox *box;

  
    /*since you cannot pass pointers to structures are the Multimax,
      I just pass the EO's id and have it look up its own structure*/
    for (eo=imgr.defeo_lst; eo != NULL; eo = eo->next)
	if (eo->id == id) 
	    break;

    if (eo == NULL) {
	EH(&eo->in, "EO not found.");
	return;
	} /*if*/

    /*set all of the rule sets to point to the right eo: needed at runtime*/
    for (rs=eo->defrs_locals; rs != NULL; rs = rs->next)
	task_start(RS_Exec, 1024, 2, eo->id, rs->name);

    barrier(imgr.barrier);

    mesg = CreateToken(MESG_ASSERT, eo->id, "initial-fact", 1, 0); /*mesg*/

    EO_MailRSs(eo->defrs_locals, mesg); /*EO_MailRSs*/

    /*For each expert object assert the new facts from the deffacts stmts*/
    for (stmt=eo->deffacts; stmt != NULL; stmt = stmt->next)
	for (facts=stmt->facts,ct = 0; facts[ct] != NULL; ++ct)  {
     mesg = CreateToken(MESG_ASSERT, eo->id, facts[ct], COMPUTE_FACT_ID(eo), 0);
	    EO_MailRSs(eo->defrs_locals, mesg);
	    } /*for*/

    eo->status = EO_READY;

    /*Send the end message*/
    EO_MailRSs(eo->defrs_locals, mesg_end);

    while (eo->status != EO_STOPPED) {

	/*Wait for the rule sets responses: what is ready to fire*/
	eo->rs_box = NULL;
	fired = NULL;

	/*Find the highest rule in the agenda*/
	for (rs=eo->defrs_locals; rs != NULL; rs = rs->next) {
	    while ((mesg=RecvMesg(rs->out_box, rs->out_sema, \
				"eo: agenda processing from RSs")) != NULL)
		if (EO_ProcessMessage(eo, rs->out_box, mesg) == NO)
		    break;
	    } /*for*/

	if (imgr.watch) {
	    semaphore_wait(imgr.print);
	    DisplayFacts(eo->defrs_locals);
	    semaphore_signal(imgr.print);
	    }/*if*/

	for (rs=eo->defrs_locals; rs != NULL; rs = rs->next) {
	    if (rs->out_box == eo->rs_box) {
		SendMesg(rs->in_box, mesg_fire, rs->in_sema, "EO to RS: Fire rule");
		fired = rs;
		} /*if*/

	    SendMesg(rs->in_box, mesg_end, rs->in_sema, "EO to RS: Mesg end");

	    } /*for*/

	/*If fired then get the new messages from the fired rule set*/
	if (fired != NULL) 
	    while ((mesg=RecvMesg(fired->out_box, fired->out_sema, "To: EO from RS: Rule fired")) != NULL)
		if (EO_ProcessMessage(eo, fired->out_box, mesg) == NO)
		    break;

	/*tell all of the other EOS that all messages have been sent*/
	for (eos = eo->out_boxes; eos != NULL; eos = eos->next) 
	     SendMesg(eos->box, mesg_end, eos->sema, "EO: Mesg end to other EOS");


	/*Now get all new messages from the EOs*/
	for (eos = eo->in_boxes; eos != NULL; eos = eos->next)  {

	    eos->status = MESG_START;
	    /*here if we have a stopped EO and no messages then we simulate
	      a message*/
	    if((eos->box->status == MESG_EO_STOPPED)&&(*eos->box->mesgs== NULL))
		semaphore_signal(eo->recv);

	    } /*for*/

	go = YES;

	while (go == YES) {

	    go = NO;

	    /*query each in box for a message*/
	    for (eos=eo->in_boxes; eos != NULL; eos = eos->next)
		if (eos->status == MESG_START) 
		    break;

	    if (eos != NULL) {
		go = YES;

		semaphore_wait(eo->recv);
		semaphore_signal(eo->recv);

		for (eos=eo->in_boxes; eos != NULL; eos = eos->next)
		    if (eos->status == MESG_START) {
			if (IsMesgs(eos->box) == YES) {

			    mesg = RecvMesg(eos->box, eo->recv, "To EO: from another EO");

			    if (EO_ProcessMessage(eo, eos->box, mesg) == NO)
				eos->status = MESG_END;

			    break; /*Only want to fetch one message*/
			    } /*if*/

			} /*if MESG_START*/
		} /*if not null*/

	    } /*while*/

	/*Send the end message*/
	EO_MailRSs(eo->defrs_locals, mesg_end);

	} /*cycle*/

    return(YES);

} /*EO_Exec*/

