/*
 *compare.c -- Holds the functions used for pattern comparison.
 */

#include <parallel.h>
#include <stdio.h>
#include "common.h"

/*
 *LocalBind - return a variable of the appropriate type.
 */
struct values *LocalBind(name, opcode, values)
char *name;
int opcode;
struct values *values;
{

    for (; values != NULL; values = values->next) 
	if ((values->name == name) && (values->opcode == opcode))
	    break;

    return(values);

} /*LocalBind*/

/*
 *CreateValue
 */
struct values *CreateValue(fact, pattern)
struct element *fact;
struct ruleslot *pattern; /*pattern*/
{
    struct values *new;
	
    new = allocatestruct(values);
    new->next = NULL;
    new->opcode = pattern->opcode;
    new->type = fact->type;
    new->name = pattern->string;
    new->oe.origin = NULL;

    if (new->type == NUMERIC)
	new->numeric = fact->numeric;
    else
	new->string = fact->string;

     return(new);

} /*CreateValue*/   

/*
 *CompareSBIND
 */
CompareSBIND(fact, value)
struct element *fact;
struct values *value;
{
    if (fact->type != value->type)
	return(NO);

    if (fact->type == NUMERIC)
	return((fact->numeric == value->numeric) ? YES : NO);
    else
	return((fact->string == value->string) ? YES : NO);

} /*CompareSBIND*/

/*
 *Compare - Compares a pattern to fact to a pattern.
 */
struct values *Compare(pattern, fact, bindings)
struct ruleslot *pattern; /*pattern being compared*/
struct element *fact; /*fact being compared to the pattern*/
struct values *bindings; /*local bindings applicable to this state*/
{
    struct values *pvalues;
    struct values *pv;
    int result = 0; /*Used for logical connectors*/

    if ((pattern == NULL) && (fact == NULL)) {
	return(ValuesCopy(bindings, ALL));
	}
    else if ((pattern == NULL) || (fact == NULL)) /*small line between success*/
	return(NULL); /*match failed*/

    result = YES;

    /*Compare the literal word of numeric*/

    switch (pattern->opcode) {
	     case WORD :

		if (fact->string != pattern->string)
		    result =  NO;

		break;

	    case NUMERIC :

		if (fact->type != NUMERIC)
		    result =  NO;
		else if (fact->numeric != pattern->numeric)
		    result =  NO;

		break;
		
	    case SFREE :

		break; /*WORD or NUMERIC*/

	    case SBIND :

		if ((pvalues=LocalBind(pattern->string, SBIND, bindings)) == NULL) {
		    for (pv=bindings; pv->next != NULL; pv=pv->next);
		    pv->next = CreateValue(fact, pattern);
		    }
		else if (CompareSBIND(fact,pvalues) == NO)  {
		    result = NO;
		    break;
		    }

		break; /*SBIND*/

	    default :
		printf("COMPARE CASE OUT OF BOUNDS\n");

	    } /*switch*/

    if (result == YES)  
	return(Compare(pattern->next, fact->next, bindings));
    else
	return(NULL);

} /*Compare*/
