/*
 *eo.c
 */

#include <parallel.h>
#include <stdio.h>
#include "common.h"
#include "comm.h"

/*
 *SetUp_EOs
 */
SetUp_EOs(eos)
struct defeo *eos;
{
    struct defeo *eo;
    struct eo_id *setids;

    for (; eos != NULL; eos = eos->next) {

	eos->status = EO_INIT;
	eos->total_rules = 0;
	eos->rs_box = NULL;

	/*Give ids to all of the outboxes eo names*/
	for (setids= eos->out_boxes; setids != NULL; setids=setids->next) {
	    eo = GetEO_by_name(setids->name);
	    setids->id = setids->box->id = eo->id;
	    setids->sema = eo->recv; /*set the right semaphore: receivers*/
	    } /*for*/

	} /*for*/

    return( YES );

} /*SetUp_EOs*/

/*
 *GetEO_by_id
 */
struct defeo *GetEO_by_id(id)
int id;
{
    struct defeo *eo_trav;

    for (eo_trav = imgr.defeo_lst; eo_trav != NULL; eo_trav = eo_trav->next)
	if (eo_trav->id == id)
	    break; /*if*/

    return(eo_trav);

} /*GetEO_by_id*/

/*
 *GetEO_by_name
 */
struct defeo *GetEO_by_name(name)
char *name;
{
    struct defeo *eo_trav;

    for (eo_trav = imgr.defeo_lst; eo_trav != NULL; eo_trav = eo_trav->next)
	if (strcmp(eo_trav->name, name) == 0)
	    break;

    return(eo_trav);

} /*GetEO_by_name*/

/*
 *CreateInBoxes
 */
CreateInBoxes(eos)
struct defeo *eos;
{
    struct eo_id *out;
  
    for (;eos != NULL; eos = eos->next)
	for(out = eos->out_boxes;out != NULL; out = out->next)
	    AddEO_to_EO_Boxes(out->name, eos->name, IN_BOX);

} /*CreateInBoxes*/

/*
 *AddEO_to_EO_Boxes
 */
AddEO_to_EO_Boxes(eo_name, eo_new, direction)
char *eo_name;
char *eo_new;
int direction;
{
    struct defeo *eo;
    struct eo_id *trav;
    struct eo_id *new;

    if ((eo = GetEO_by_name(eo_name)) == NULL) {
	sprintf(error_sb, "Expert object <%s> not found.", eo_name);
	EH(NULLSCANNER, error_sb);
	return(NO);
	} /*if*/

    trav = (direction == IN_BOX) ? eo->in_boxes : eo->out_boxes;

    /*Does the name already exist*/
    for (; trav != NULL; trav=trav->next)
	if (strcmp(eo_new, trav->name) == 0)
	    return(NO);

    new = allocatestruct(eo_id);

    if (direction == OUT_BOX) {
	new->box = CreateMailbox();
	new->box->id = -1;
	new->box->name = eo_new;
	} /*if*/
		
    new->status = MESG_START;
    new->name = eo_new;
    new->id = new->box->id;

    /*add the mailbox to the EO*/
    new->next = (direction == IN_BOX) ? eo->in_boxes : eo->out_boxes;

    if (direction == IN_BOX)
	eo->in_boxes = new;
    else
	eo->out_boxes = new;

    return(YES);

} /*AddEO_to_EO_Boxes*/

/*
 * RS_Exists
 */
RS_Exists(name)
char *name;
{
    struct defrs *trav;

    for (trav = imgr.defrs_lst; trav != NULL; trav = trav->next)
	 if (name == trav->name)
	     return(YES);

    return(NO);

} /*RS_Exists*/
