/*
 *execute.c - contains all of the functions necessary for the evaluation of
 *            the right hand side of a rule.
 */

#include <parallel.h>
#include <stdio.h>
#include "common.h"
#include "comm.h"

/*
 *ExecuteRHS - Executes all the functions on the RHS of a rule.
 */
ExecuteRHS(rs, commands, locals)
struct defrs *rs;
struct lr *commands;
struct values *locals;
{
    for (; commands != NULL; commands = commands->next) 
	rhs[commands->locals->whoset](rs, commands->locals->next, locals);
	
} /*ExecuteRHS*/

/*
 *ExecuteStop - shut down the expert object.
 */
ExecuteStop(rs, command, locals)
struct defrs *rs;
struct values *command; /*list of variables to retract*/
struct values *locals;  /*Local variables list bound: state*/
{

    SendMesg(rs->out_box, mesg_stopped, rs->out_sema, "RHS: Execute Stop");

    return(YES);

} /*ExecuteStop*/

/*
 *ExecuteRetract - executes the retract function.
 */
ExecuteRetract(rs, command, locals)
struct defrs *rs;
struct values *command; /*list of variables to retract*/
struct values *locals;  /*Local variables list bound: state*/
{
    char *mesg;
    struct defeo *eo;
    struct eo_id *eo_ids;
    struct eo_id *eos;
    struct values *pv;

    eos = command->oe.eo_id;

    /*For each variable listed, retract the fact it is bound to*/
    for (;command != NULL; command = command->next) {

	if ((pv=LocalBind(command->string, SBIND, locals)) == NULL) {
	    sprintf(error_sb,"Retract Failed: ?%s not found.", command->string);
	    EH(&rs->in, error_sb);
	    return(NO);
	    }

	if (pv->oe.origin == NULL) {
sprintf(error_sb, "Retract Failed: ?%s not bound to a fact.", command->name);
	    EH(&rs->in, error_sb);
	    return(NO);
	    } /*if*/

	for (eo_ids=eos; eo_ids != NULL; eo_ids = eo_ids->next) {
	 eo = GetEO_by_name(eo_ids->name);
	 mesg = CreateToken(MESG_RETRACT, eo->id, SBNULL, pv->oe.origin->id, 0);
	 SendMesg(rs->out_box, mesg, rs->out_sema, "RHS: Retract");
	 } /*for*/

        } /*for each command argument*/

    return(YES);

} /*ExecuteRetract*/

/*
 *ExecutePrintOut - Executes the RHS command "printout"
 */
ExecutePrintOut(rs, command, locals)
struct defrs *rs;
struct values *command;
struct values *locals;
{
    struct values *pv;

    semaphore_wait(imgr.print);

    for (;command != NULL; command=command->next) {

	switch (command->opcode) {
	    case NUMERIC :
		printf("%f", command->numeric);
		break;

	    case WORD :
		printf("%s", command->string);
		break;

	    case SBIND :
		if ((pv=LocalBind(command->string, SBIND, locals)) == NULL) {
		    PRINTOUT_ERROR:
sprintf(error_sb,"Variable %s not found for PRINTOUT command.",command->string);
		    EH(&rs->in, error_sb);
		    goto EXIT_PRINTOUT;
		    }

		if (pv->type == NUMERIC)
		    printf("%f", pv->numeric);
		else
		    printf("%s", pv->string);
		break;
	    default :
		EH(&rs->in, "Case in ExecutePrintOut out of range");
		goto EXIT_PRINTOUT;

	    } /*switch*/

	printf(" ");

	} /*for each element of the printout*/

    printf("\n");

EXIT_PRINTOUT:

    semaphore_signal(imgr.print);
    
} /*ExecutePrintOut*/

/*
 *ExecuteAssert - executes the RHS command ExecuteAssert
 */
ExecuteAssert(rs, command, locals)
struct defrs *rs;
struct values *command; /*command arguments*/
struct values *locals; /*local variables list*/
{
    char *fact;
    char *mesg;
    struct defeo *eo;
    struct element *pelenew;
    struct element *pelehead;
    struct eo_id *eos;
    struct values *pv; /*return values from LocalBind*/

    pelehead = NULL;

    do {

     if (command->opcode == EO_IDS) {

	fact = CreateMesg(pelehead);

	for (eos=command->oe.eo_id; eos != NULL; eos = eos->next) {

 	 eo = GetEO_by_name(eos->name);

	 mesg=CreateToken(MESG_ASSERT, eo->id, fact, COMPUTE_FACT_ID(rs->eo),0);

	 SendMesg(rs->out_box, mesg, rs->out_sema, "RHS: Assert");

	 } /*for*/

	break;
	} /*if*/

	if (pelehead == NULL) 
	    pelenew = pelehead = allocatestruct(element);
	else {
	    pelenew->next = allocatestruct(element);
	    pelenew = pelenew->next;
	    } /*else*/

	/*create the appropriate elements*/

	switch (command->opcode) {

	    case WORD :
		pelenew->type = WORD;
		pelenew->string = command->string;
		break;

	    case NUMERIC :
		pelenew->type = NUMERIC;
		pelenew->numeric = command->numeric;
		break;

	    case SBIND :
		if ((pv=LocalBind(command->string, SBIND, locals)) == NULL) {
		    MISSING_NAMES:
sprintf(error_sb,"Variable %s not found for ASSERT command.", command->string);
		    EH(&rs->in, error_sb);
		    goto ASSERT_ERROR;
		    }

		pelenew->type = pv->type;

		if (pelenew->type == WORD)
		    pelenew->string = pv->string;
		else
		    pelenew->numeric = pv->numeric;

		break;

	    default :
		EH(&rs->in, "ASSERT FIELD OUT OF BOUNDS IN SWITCH");
		return(NO);
		break;
	    } /*switch*/

	command = command->next;
	} while (command != NULL);

	pelenew->next = NULL;
		
	return(YES);
		
ASSERT_ERROR: /*Error, clean up*/
	pelenew->next = NULL;
	ReturnElements(pelehead);
	return(NO);

} /*ExecuteAssert*/

/*
 *RetractFlow - retracts a fact from the working memory.
 */
RetractFlow(rs, fact)
struct defrs *rs;
struct fact *fact; /*fact that we are retracting*/
{
    struct match *pms; /*Pattern Matches*/
    struct datalink *flow; /*links from each pattern*/

    CleanAgenda(rs, fact->id);
    
    /*For each pattern the fact matched*/
    for (pms=fact->list; pms != NULL; pms=pms->next) 
	for (flow=pms->slot->outflow; flow != NULL; flow = flow->next) 
	    RemoveId(flow->path, flow->side, fact->id);

    return(YES);

} /*RetractFlow*/

/*
 *RemoveId
 */
RemoveId(des, side, id)
struct decidernode *des;
int side;
int id;
{
    struct lr *head, *binds, *last, *temp;

    for (;des->outflow != NULL;des=des->outflow){

	head = binds = (side == RHS) ? des->rhs : des->lhs;

	/*For each binding list in the side*/
	for (last=NULL; binds != NULL; ) 
	    if (AnyId(id, binds->locals) == NO) {
		last = binds;
		binds = binds->next;
		}
	    else {

		/*Remove locals pair*/

		temp = binds;

		if (last == NULL) 
		    binds = head = binds->next;
		else 
		    binds = last->next = binds->next;

		FreeLr(temp);

		} /*Each binding list*/

	if (side == RHS)
	    des->rhs = head;
	else
	    des->lhs = head;

	side = des->side; /*Next side we are entering from*/
    } /*for*/

} /*RemoveId*/
	
/*
 *AnyId - checks a token list to see if a particular system id exists.
 */
AnyId(id, values)
int id; /*ID we are looking for*/
struct values *values; /*List of values*/
{

    for (;values != NULL; values = values->next) {
	if (values->opcode != SPECIAL)
	    return(NO);

	if (values->whoset == id)
	    return(YES);
	} /*for*/

    return(NO);

} /*AnyId*/
