/*
 *init.c - contains the code for the initialization of the system when it
 *         starts up.
 */

#include <parallel.h>
#include <stdio.h>
#include "common.h"
#include "init.h"
#include "comm.h"

/*
 *Init()
 */
Init()
{
    int i;

    /*allocates memory for the system*/
    if (share_malloc_init( SYSTEM_MEMORY ) != 0) {
	EH(NULLSCANNER, "NO SYSTEM MEMORY: share_malloc_init");
	exit(0);
	} /*if*/

    /*Set the mode for the interface manager*/
    imgr.watch = YES ;
    imgr.lock = spin_create( PAR_UNLOCKED );

    if ((imgr.print = semaphore_create( TASK_BLOCK )) == NULL)
	EH(NULLSCANNER, "Semaphore print failed in creation.");

    sprintf(mesg_end, "%d", MESG_END );
    sprintf(mesg_stopped, "%d", MESG_STOPPED );
    sprintf(mesg_eo_stopped, "%d", MESG_EO_STOPPED);
    sprintf(mesg_na, "%d", MESG_NA );
    sprintf(mesg_fire, "%d", MESG_FIRE );

    /*Always initialize the hash table first, before all other initializations*/
    for (i=0; i < HASHSIZE; i++) {
	ht[i] = NULL;
	ht_locks[i] = spin_create( PAR_UNLOCKED );
	} /*for*/

    rhs[ RHSASSERT ] = ExecuteAssert;
    rhs[ RHSPRINTOUT ] = ExecutePrintOut;
    rhs[ RHSRETRACT ] = ExecuteRetract;
    rhs[ RHSSTOP ] = ExecuteStop;

    imgr.error = NO;

    imgr.eo_ids = 1;
    imgr.current_eo = NULL;
    imgr.defeo_lst = NULL;

    imgr.current_rs = NULL;
    imgr.defrs_lst = NULL;


    ASSERT = HashString("assert");
    DECLARE = HashString("declare");
    DEFEO = HashString("defeo");
    DEFFACTS = HashString("deffacts");
    DEFRS = HashString("defrs");
    DEFRULE = HashString("defrule");
    PRINTOUT = HashString("printout");
    PRIORITY = HashString("priority");
    RETRACT = HashString("retract");
    SBNULL = HashString("");
    SBAND = HashString("and");
    STOP = HashString("stop");

} /*Init*/

/*
 *CreateRS (see struct definition for comments about defrs declaration)
 */
struct defrs *CreateRS()
{
    struct defrs *new;

    new = allocatestruct(defrs);

    new->priority = RS_PRIORITY_DEFAULT;
    new->status = EO_READY;

    InitScanner(&new->in);

    new->Agenda.id = 0;
    new->Agenda.list = NULL;

    new->out_box = CreateMailbox();
    new->in_box = CreateMailbox();

    new->out_sema = semaphore_create( TASK_BLOCK );
    new->in_sema = semaphore_create( TASK_BLOCK );

    new->factlist = NULL;
    new->bindings = NULL;
    new->patterns = NULL;
    new->defrules = NULL;
    new->next = NULL;

    return(new);

} /*CreateRS*/

/*
 *CreateEO
 */
struct defeo *CreateEO()
{
    struct defeo *new;

    if (imgr.current_eo != NULL)
	imgr.current_eo->defrs_locals = imgr.defrs_lst;

    imgr.defrs_lst = NULL;
    imgr.current_rs = NULL;

    new = allocatestruct(defeo);

    InitScanner(&new->in);

    new->defrs_locals = NULL;
    new->total_rules = 0;
    new->factid = 1;
    new->status = EO_INIT;
    new->deffacts = NULL;
    new->rs_box = NULL;

    new->id = imgr.eo_ids++;

    new->in_boxes = NULL;
    new->recv = semaphore_create( TASK_BLOCK );

    new->out_boxes = NULL;

    return(new);

} /*CreateEO*/

/*
 *CreateMailbox
 */
struct mailbox *CreateMailbox()
{
    struct mailbox *new;
    static box_id = 1000;

    new = allocatestruct(mailbox);

    new->lock = spin_create( PAR_UNLOCKED );
    new->status = MESG_START;
    new->box_id = box_id++;

    new->mesgs = INIT_BUF;
    *new->mesgs = NULL;

    return(new);

} /*CreateMailbox*/

/*
 *InitScanner
 */
InitScanner(in)
struct scanner *in;
{

    in->laCt = in->laFill = in->laEmpty = 0;
    in->ln = 1;
    in->input = NULL;
    in->ct = 0;
    in->fn = NULL;
    in->ch = EOF;

} /*InitScanner*/
