/*
 *print.c - contains all of the functions responsible for the printing of data
 *          to the user's screen.
 */

#include <parallel.h>
#include <stdio.h>
#include "common.h"
#include "comm.h"

/*
 *PrintAgenda
 */
PrintAgenda(eo)
struct defeo *eo;
{
    struct activation *act;
    struct defrs *rs;

    printf("===============Agenda EO:<%10s>==============\n", eo->name);

  for (rs=eo->defrs_locals; rs != NULL; rs = rs->next) {

    printf("---------------Agenda RS:<%10s>--------------\n", rs->name);

    for (act = rs->Agenda.list; act != NULL; act = act->next) {
	printf("Rule Name: %s; ", act->rhs->rhs->locals->name);
	printf("Activation id: %d; ", act->id);
	printf("Priority %6.2f.",  act->priority);
	PrintIds(act->variables);
	printf("\n");
	} /*for*/
    } /*for*/

    printf("------------------------------------\n");

} /*PrintAgenda*/ 

/*
 *PrintIds
 */
PrintIds(locals)
struct values *locals;
{
    char SP = ' ';

    printf(" [");

    for (;locals != NULL; locals=locals->next) 
	if ((locals->opcode == SPECIAL) && (locals->whoset >= 0)) {
	    printf("%cf-%d", SP, locals->whoset);
	    SP = ',';
	    } /*if*/

    printf(" ]");

} /*PrintIds*/

/*
 *DisplayFacts -- Prints out the fact list for a rule set.
 */
DisplayFacts(rs)
struct defrs *rs;
{
    struct fact *pfacttrav;
    struct element *peletrav;

    printf("=======================================================\n");
    printf("Rule set: <%s>; Expert Object: <%s>.\n ", rs->name, rs->eo->name);
    printf("Blackboard contains:\n");
    printf("=======================================================\n");

    for(pfacttrav = rs->blackboard;pfacttrav != NULL; pfacttrav=pfacttrav->next) {

	printf("Id: %d; ", pfacttrav->id);

	for (peletrav=pfacttrav->entry;peletrav != NULL;peletrav=peletrav->next)
	    switch(peletrav->type) {
		case WORD :
			printf("%s ", peletrav->string);
			break;

		case NUMERIC :
			printf("%f ", peletrav->numeric);
			printf(" ");
			break;

		default :
			EH(NULLSCANNER, "Display facts: Unknown fact type");
			break;
		} /*switch*/

	printf("\n");
	} /*for*/

    printf("=======================================================\n");

    return(YES);

} /*DisplayFacts*/

/*
 *PrintMesgs
 */
PrintMesgs(mesgs)
char **mesgs;
{
    if (*mesgs == NULL)
	printf("No messages.\n");
    else for (;*mesgs != NULL;++mesgs)
	printf("mesg: <%s> \n", *mesgs);

    return(YES);

} /*PrintMesgs*/

/*
 *PrintBoxes
 */
PrintBoxes(boxes)
struct eo_id *boxes;
{
    for (; boxes != NULL; boxes = boxes->next) {

	printf("Box name <%s>; id <%d>\n", boxes->name, boxes->id);

	PrintBox(boxes->box);

	} /*for*/

} /*PrintBoxes*/

/*
 *PrintBox
 */
PrintBox(box)
struct mailbox *box;
{

	printf("eo <%s>: id <%d>: status: ", box->name, box->id);

	switch (box->status) {
		case MESG_EO_STOPPED :
				printf("eo stopped.\n");
				break;

		case MESG_STOPPED :
				printf("stopped.\n");
				break;

		case MESG_INIT :
				printf("initialization.\n");
				break;

		case MESG_END :
				printf("received.\n");
				break;

		case MESG_START :
				printf("in progress.\n");
				break;

		default :
				printf("unknown: se.\n");
				break;
		} /*switch*/

	PrintMesgs(box->mesgs);
	printf("----------------------------done.\n");

    return(YES);

} /*PrintBox*/

/*
 *PrintSummary
 */
PrintSummary()
{
    struct defeo *eos;
    long ct;

    printf("--------------------Execution Summary--------------------\n");

    for (eos=imgr.defeo_lst; eos != NULL; eos = eos->next) {
	printf("Expert Object <%s>: ", eos->name);

	printf(" status <");
	switch (eos->status) {
		case EO_READY :
			printf("ready");
			break;

		case EO_STOPPED :
			printf("stopped");
			break;

		case EO_INIT :
			printf("initialized");
			break;

		case EO_PROCESSING :
			printf("running");
			break;

		case EO_WAITING :
			printf("waiting");
			break;

		default :
			printf("unknown");
			break;
		}

	ct = eos->total_rules;
	printf("> : %d rule%c fired.\n", ct, (ct == 1) ? ' ': 's');

	} /*for*/

    printf("---------------------------------------------------------\n");

    return(YES);

} /*PrintSummary*/
