/*
 *assert.c - code for the assert function.
 */

#include <parallel.h>
#include <stdio.h>
#include "common.h"
#include "comm.h"

/*
 *ProcessNewString - Take a character string as input and parse it into
 *		     an internal fact format.
 */
struct element *ProcessNewString(string)
char *string;
{
    struct element *pelenew;
    struct element *peleret;
    struct scanner in;

    /*turn on the scanner string mode*/

    /*now have the scanner use this input string*/
    InitScanner(&in);
    in.input = string;
    in.ln = -1;
    in.ch = GetScanChar(&in); /*prime the scanner*/
    
    /*peleret is the pointer to the start of the fact structure*/
    peleret = NULL;

    /*For each token in the assert string*/
    for (GetToken(&in, YES); in.token.opcode != QUIT; GetToken(&in, YES)) {

	    if (peleret != NULL) {
		pelenew->next = allocatestruct(element);
		pelenew = pelenew->next;
		}
	    else
		pelenew = peleret = allocatestruct(element);

	    pelenew->type =in.token.opcode;

	    /*facts only have two types of elements*/
	    switch (in.token.opcode) {
		case NUMERIC :
			pelenew->numeric = in.token.numeric;
			break;

		case WORD :
			pelenew->string =in.token.string;
			break;

		default :
			EH(&in, "Illegal assert format.");
			goto PNSEXIT;
			break;
		} /*switch*/

	} /*for length of assert*/

    if (peleret != NULL)
	pelenew->next = NULL;

PNSEXIT :
    return(peleret);

} /*ProcessNewString*/

/*
 *assert
 *      Given a fact as a character string, this function asserts it into
 *      the working memory.
 */
Assert(rs, string, id)
struct defrs *rs;
char *string;
int id;
{

    if (Assertx(rs, ProcessNewString(string), id) == NULL)
	return(NO); /*the string already existed*/

    return(YES); /*assert did take place*/

} /*Assert*/

/*
 *Assertx - assert a fact from the main system.
 */
struct fact *Assertx(rs, newfact, id)
struct defrs *rs;
struct element *newfact;
int id;
{
    struct fact *new;

    if (newfact == NULL)
	return(NULL);

    /*first see if the fact already exists, if it does we do not assert it
      again since it is already known
     */
    if (FactExists(rs, newfact) == YES) 
	return(NULL);

    /*now we create the fact*/
    new = allocatestruct(fact);

    /*
      fact id is computed either as the given id (computed in another EO) or
      as a computed EO which is a fact id for this EO, with this EO's id as
      the least significant positions of the id (see documentation)
     */

    new->id = id;
    new->entry = newfact;
    new->next = NULL;

    new->next = rs->blackboard;
    rs->blackboard = new;

    new->list = Dataflow(rs, new);

    return(new); /*The fact has been asserted*/

} /*Assert a fact*/


/*
 *RemoveFact - removes a fact from the fact list.
 */
RemoveFact(rs, fact)
struct defrs *rs;
struct fact *fact;
{

    if (fact == rs->blackboard) {

	rs->blackboard = rs->blackboard->next;

	if (rs->blackboard != NULL)
	    rs->blackboard->prev = NULL;
	} /*if*/
    else {
	if (fact->next != NULL) 
	    fact->next->prev = fact->prev;

	fact->prev->next = fact->next;
	} /*else*/
	    
    ReturnElements(fact->entry);

    FREEMEM(fact);

 } /*For each command*/

/*
 *AddFacts - Add a new Deffacts statement to the db.
 *      cond -> 0 : Add deffacts statement to its structure.
 *	cond -> 1 : Add the name to the list of known deffact statements.
 *      cond -> 2 : Add the parsed facts to the deffacts structure.
 */
AddFacts(cond, name, stmt, facts)
int cond;
char *name; 
char *stmt; 
char **facts;
{
    struct deffacts *newfacts;

    newfacts=imgr.current_eo->deffacts;

    for (; newfacts != NULL; newfacts = newfacts->next)
	if (newfacts->name == name) {
	    if (cond != 1)
		break;

	    sprintf(error_sb, "Deffact %s redeclared.", name);
	    EH(NULLSCANNER, error_sb);
	    return(NO);
	    } /*if*/
    
    switch (cond) {
	case 0 : /*add stmt*/
		newfacts->stmt = MemAlloc(strlen(stmt) + 1);
		strcpy(newfacts->stmt, stmt);
		break;

	case 1 : /*add name*/
		newfacts = allocatestruct(deffacts);
		newfacts->name = name;
		newfacts->next = imgr.current_eo->deffacts;
		imgr.current_eo->deffacts = newfacts;
		break;

	case 2 : /*add parsed facts*/
		newfacts->facts = facts;
		break;
	} /*switch*/

    return(YES);

} /*AddFacts*/

/*
 *FactExists - does a fact exist in the blackboard.
 */
FactExists(rs, fact)
struct defrs *rs;
struct element *fact;
{
    struct fact *pfacttmp;

    for (pfacttmp=rs->blackboard; pfacttmp != NULL; pfacttmp=pfacttmp->next)
	if (Equal(fact, pfacttmp->entry) == YES)
	    return( YES );

    return( NO );

} /*FactExists*/

/*
 *GetFact
 */
struct fact *GetFact(rs, fact_id)
struct defrs *rs;
int fact_id;
{
    struct fact *trav;

    for (trav = rs->blackboard; trav != NULL; trav = trav->next)
	if (trav->id == fact_id)
	    break;

    return(trav);

} /*GetFact*/
