/*
 *scanner.c - This file contains the scanner routines.
 */

#include <parallel.h>
#include <stdio.h>
#include "common.h"

GetIA(in)
struct scanner *in;
{
    if (in->input[in->ct] == EOS)
	return(EOF);

    return((int)in->input[in->ct++]);

} /*GetIA*/

UnGetScanChar(in, ch, fp)
int ch;
FILE *fp;
struct scanner *in;
{
    if (in->fn == NULL)
	in->ct--;
    else
	ungetc(ch, fp);
}

GetScanChar(in)
struct scanner *in;
{
    if (in->fn == NULL)
	return(GetIA(in));
    else
	return(getc(in->fp));

} /*GetScanChar*/

char *GetWord(in)
struct scanner *in;
{
    char wordsb[ MAXWORDLEN ];
    int ct = 0;		/*counts the number of letters read*/

    for(; IsWord(in->ch); in->ch = GetScanChar(in))
	wordsb[ct++] = in->ch;

    wordsb[ct] = EOS;

    return(HashString(wordsb));

} /*GetWord*/

/*
 *GetToken -- gets the next token from the input buffer.
 */
GetToken(in, la)
struct scanner *in;
int la; 
{
    char wordsb[ MAXWORDLEN ];
    float fin;
    int ct;
    int sign;

    if (la == YES ) {
	if (in->laCt > 0) {

	    in->token.string = in->LaTokens[in->laEmpty].string;
	    in->token.numeric = in->LaTokens[in->laEmpty].numeric;
	    in->token.opcode = in->LaTokens[in->laEmpty].opcode;
	    in->laCt--;
	    in->laEmpty++;

	    if (in->laEmpty >= LASIZE)
		in->laEmpty = 0;

	    return(0);
	    } /*if*/
	} /*la == YES*/

    DoForever {

	if (in->ch == EOL) {
	    in->ln = in->ln + 1;
	    in->ch = GetScanChar(in);
	    } /*if*/
	else if (IsFiller(in->ch)) {
	    while (IsFiller(in->ch))
		in->ch = GetScanChar(in);
	    } /*else*/
	else if (in->ch == ';') {
	    while ((in->ch != EOL) && (in->ch != EOF))
		in->ch = GetScanChar(in);
	    } /*else if*/
	else 
	    break; /*Found a valid parsing token*/
	} /*Take care of nontoken strings*/

    if (IsAlpha(in->ch)) {
	in->token.opcode = WORD;
	in->token.string = GetWord(in);
	} /*if*/
    else switch (in->ch) {
	case ',' :
		in->token.opcode = COMMA;
		in->ch = GetScanChar(in);
		break;

	case '<' :
		if ((in->ch = GetScanChar(in)) != '-') {
		    EH(in, "Unknown format '<????'");
		    return(NO);
		    }

		in->ch = GetScanChar(in);
		in->token.opcode =  BINDER;
		break;

	case '"' :
		for (ct=0; (wordsb[ct] = GetScanChar(in)) != '"';) {

		    if (wordsb[ct] == EOF) {
			EH(in, "End of file found in string");
			return(0);
			} /*if*/

		    if (wordsb[ct] == EOL)
			EH(in, "Line return found in string");

		    ct++;

		    if (ct >= MAXWORDLEN)
			EH(in, "String length limit exceeded.");
		    } /*for*/

		in->ch = GetScanChar(in);
		wordsb[ct] = EOS;
		in->token.opcode = WORD;
		in->token.string = MakeString(wordsb);
		break;

	case '(' :
		in->token.opcode = LPAREN;
		in->ch = GetScanChar(in);
		break;

	case ')' :
		in->token.opcode = RPAREN;
		in->ch = GetScanChar(in);
		break;

	case EOF :
		in->token.opcode = QUIT;
		break;

	case '=' :

		if ((in->ch = GetScanChar(in)) != '>') {
		    EH(in, "Illegal form found '='");
		    break;
		    } /*if*/

		in->token.opcode = SEPARATOR;
		in->ch = GetScanChar(in);

		break;

	case '?' :
		in->ch = GetScanChar(in);

		if (IsAlpha(in->ch)) {
		    in->token.string = GetWord(in);
		    in->token.opcode = SBIND;
		    } /*if*/
		else
		    in->token.opcode = SFREE;

		break;
	default :
		sign = 1;

		if (IsNumeric(in->ch) || (in->ch == '-')) {

		    if (in->ch == '-') {
			sign = -1;
			in->ch = GetScanChar(in);
			} /*ch == '-'*/

		    if (in->ch == '>') {
			in->token.opcode = SENDER;
			in->ch = GetScanChar(in);
			}
		    else if (IsNumeric(in->ch)) {

		        in->token.opcode = NUMERIC; 

			UnGetScanChar(in, in->ch, in->fp);

			if (in->fn == NULL) 
			    sscanf(in->input + in->ct, "%f", &fin);
			else 
			    fscanf(in->fp, "%f", &fin);

			in->token.numeric = fin * sign;

			/*Parse off the read float*/
			for (in->ch=GetScanChar(in); IsNumeric(in->ch);in->ch=GetScanChar(in));

			if (in->ch == '.') {
			   in->ch = GetScanChar(in);

			    if (!(IsNumeric(in->ch)))
				break;

			    for (; IsNumeric(in->ch);in->ch=GetScanChar(in));

			    if ((in->ch == 'e') || (in->ch == 'E')) 
				in->ch = GetScanChar(in);

			    for (; IsNumeric(in->ch);in->ch=GetScanChar(in));

			    } /*ch == '.'*/
			} /*IsNumeric*/
		    else {
			EH(in, "-???? is not supported");
			goto GETTOKEN_ERROR;
			}
		    } /*Is numeric or starts with '-' */
		else {
		    if ((in->ch == EOS) && (in->fp == NULL)) {
			in->token.opcode == EOF;
			break;
			}

	    sprintf(error_sb, "Unknownin->character found octal %o\n", in->ch);
		    EH(in, error_sb);
		    goto GETTOKEN_ERROR;
		    } /*Not Numeric, nor Production*/
		break;
	    } /*switch*/

    if (la == SAVE) {
	if (in->laCt == LASIZE)		/*DEBUGGING*/
	    EH(in, "Look Ahead buffer is overextended by one.");

	in->LaTokens[in->laFill].string = in->token.string;
	in->LaTokens[in->laFill].numeric = in->token.numeric;
	in->LaTokens[in->laFill].opcode = in->token.opcode;

	in->laFill++;
	if (in->laFill >= LASIZE)
	    in->laFill = 0;

	in->laCt++; }

    in->ch =in->ch;

    return(YES);

GETTOKEN_ERROR:

    EH(in, "Unknown input token format found.");
    return(NO);

} /*GetToken*/
