/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: assert.c - code for the assert function.
 *
 * Date: 6-16-87
 *
 */

#include <stdio.h>
#include "common.h"

/*
 *ProcessNewString - Take a character string as input and parse it into
 *		     an internal fact format.
 */
struct element *ProcessNewString(string)
char *string;
{
    struct element *pelenew;
    struct element *peleret;

    char *filename;
    char *oldinput;
    char oldch;
    int oldct;

    /*turn on the scanner string mode*/
    filename = imgr.fn;
    oldinput = imgr.input;
    oldct = imgr.ct;
    oldch = ch;

    /*now have the scanner use this input string*/
    imgr.fn = NULL;
    imgr.ct = 0;
    imgr.input = string;

    ch = GetScanChar(); /*prime the scanner*/
    
    /*peleret is the pointer to the start of the fact structure*/
    peleret = NULL;

    /*For each token in the assert string*/
    for (GetToken(YES); token.opcode != QUIT; GetToken(YES)) {

	    if (peleret != NULL) {
		pelenew->next = allocatestruct(element);
		pelenew = pelenew->next;
		}
	    else
		pelenew = peleret = allocatestruct(element);

	    pelenew->type = token.opcode;

	    /*facts only have two types of elements*/
	    switch (token.opcode) {
		case NUMERIC :
			pelenew->numeric = token.numeric;
			break;

		case WORD :
			pelenew->string = token.string;
			break;

		default :
			EH("Illegal assert format.");
			goto PNSEXIT;
			break;
		} /*switch*/

	} /*for length of assert*/

    if (peleret != NULL)
	pelenew->next = NULL;

    /*Restore the old state of the scanner*/
PNSEXIT :
    imgr.fn = filename;
    imgr.input = oldinput;
    imgr.ct = oldct;
    ch = oldch;

    if (imgr.error == YES) {
	ReturnElements(peleret);
	peleret = NULL;
	}

    return(peleret);

} /*ProcessNewString*/

/*
 *Assert 
 *	Given a fact as a character string, this function asserts it into
 *      the working memory.
 */
Assert(eo, string, id)
struct defeo *eo;
char *string;
int id;
{

    if (Assertx(eo, ProcessNewString(string), id) == NULL)
	return(NO); /*the string already existed*/

    return(YES); /*assert did take place*/

} /*Assert*/

/*
 *Assertx - assert a fact from the main system.
 */
struct fact *Assertx(eo, newfact, id)
struct defeo *eo;
struct element *newfact;
int id;
{
    struct fact *new;

    if (newfact == NULL)
	return(NULL);

    /*first see if the fact already exists, if it does we do not assert it
      again since it is already known
     */
    if (FactExists(eo, newfact) == YES) {
	return(NULL); /*Didn't assert the fact: it existed already*/
	} /*FactExists already (value)*/

    /*now we create the fact*/
    new = allocatestruct(fact);

    /*fact id is computed either as the given id (computed in another EO) or
      as a computed EO which is a fact id for this EO, with this EO's id as
      the least significant positions of the id (see documentation)*/

    new->id = (id < 0) ? COMPUTE_FACT_ID(eo) : id;
    new->entry = newfact;
    new->next = NULL;
    new->list = NULL;

    if (eo->factnew != NULL) {
	eo->factnew->next = new;
	new->prev = eo->factnew;
	eo->factnew = new;
	}
    else  {
	new->prev = NULL;
	eo->factlist = eo->factnew = new;
	}

    if (eo->factptr == NULL)
	eo->factptr = eo->factnew;

    return(new); /*The fact has been asserted*/

} /*Assert a fact*/

