/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: comm.c - communication software.
 *
 * Date: 6-16-87
 *
 */

#include <stdio.h>
#include <string.h>
#include "common.h"
#include "comm.h"

char mesg[ MAX_MESG_LEN ];
char mesg_create[ MAX_MESG_LEN ];
char mesg_start[ 10 ];
char mesg_end[ 10 ];
char mesg_stopped[ 10 ];

/*
 *StartComm
 */
StartComm(eo)
struct defeo *eo;
{
    struct mailbox *trav;
    char *mesg;
    int done = YES;
    int go;

    if (eo->status == EO_STOPPED)
	return(NO);

    if (eo->status == EO_INIT) {
	ProcessFacts(eo);
	eo->status = EO_READY;
	} /*if*/

    /* This is a critical section: and in parallel should have a lock */
    for (trav=eo->in_boxes; trav != NULL; trav=trav->next) {

	if (trav->status == MESG_STOPPED)
	    continue;

       /*If we are still expecting more messages then see if they have arrived*/
	if (trav->status != MESG_END) {

	    go = NO;

	    while ((mesg = RecvMesg(trav)) != NULL) {
		go = YES;
		ProcessMessage(eo, mesg, trav);
		} /*while*/

	    if (trav->status != MESG_END)
		done = NO;

	    if (go == YES) 
		ProcessFacts(eo);

	    }  /*if*/

	} /*for*/

    /*ProcessAgenda resets all of the in boxes to MESG_INIT*/
    if (done == YES)
	ProcessAgenda(eo);

/*
 The second condition above says that if no EO's fired a rule in the
 last round of execution, and this EO's in boxes are all still at
 the message initialization state, then check to see if this EO
 has any rules that are ready to fire in its distributed agenda.
 */

    return(YES);

} /*StartComm*/

/*
 *ProcessMessage
 */
ProcessMessage(eo, mesg, in)
struct defeo *eo;
char *mesg;
struct mailbox *in;
{
    int type;
    int id;
    char token[ MAX_MESG_LEN ];
    struct fact *fact;

    sscanf(mesg, "%d", &type);

    switch ((int)type) {

	case MESG_STOPPED :
			in->status = MESG_STOPPED;
			break;
			
	case MESG_END :
			in->status = MESG_END;
			break;

	case MESG_START :
			in->status = MESG_START;
			break;

	case MESG_RETRACT :
			imgr.status = SYSTEM_CONT;
			sscanf(mesg, "%*d %d", &id);
			fact = GetFact(eo, id);
			RetractFlow(eo, fact);
			RemoveFact(eo, fact);
			break;

	case MESG_ASSERT :
			imgr.status = SYSTEM_CONT;
			strcpy(token, strpbrk(mesg, "|") + 1);
			sscanf(mesg, "%*d %d ", &id);
			Assert(eo, token, id);
			break;

	} /*switch*/
 
} /*ProcessMessage*/

/*
 *ProcessAgenda
 */
ProcessAgenda(eo)
struct defeo *eo;
{
    struct mailbox *trav;
    struct defrs *rs_agenda;
    struct defrs *rs_trav;

    /* conflict resolution */
    rs_agenda = NULL;

    for (rs_trav = eo->defrs_locals; rs_trav != NULL; rs_trav = rs_trav->next)
      if (rs_trav->Agenda.list != NULL) {

	imgr.status = SYSTEM_CONT;

	if (rs_agenda == NULL) {
	    imgr.stmt_name = rs_trav->Agenda.list->rhs->rhs->locals->name;
	    rs_agenda = rs_trav;
	    } 
else if (rs_agenda->Agenda.list->priority < rs_trav->Agenda.list->priority) {
	    imgr.stmt_name = rs_trav->Agenda.list->rhs->rhs->locals->name;
	    rs_agenda = rs_trav;
	    }

	if (imgr.watch == YES)
	    PrintAgenda_RS(rs_trav);

	} /*Agenda still has entries*/


    /*This had to be done here since the firing of the rule can change
      the state of the expert object in a permanent way: <stopped>.
    */
    eo->status = EO_READY;

    if (DEBUG)
	printf("AGENDA HAS BEEN APPLIED\n");

    if (rs_agenda != NULL) 
	FireRule(eo, rs_agenda);

    /*send to all peers an end of message indicator*/
    for (trav=eo->out_boxes; trav != NULL; trav=trav->next) {
	SendMesg(eo, mesg_end, trav->eo_id);
	/*reset out_boxes*/
	trav->status = MESG_INIT;
	} /*for*/
	
    /*reset in-mail boxes*/
    for (trav=eo->in_boxes; trav != NULL; trav=trav->next)
	if (trav->status != MESG_STOPPED)
		trav->status = MESG_INIT;

    if (imgr.watch)
	DisplayFacts(eo);
	
} /*ProcessAgenda*/

/*
 *FireRule
 */
FireRule(eo, rs)
struct defeo *eo;
struct defrs *rs;
{
    struct activation *temp;

    imgr.total_rules++;
    eo->total_rules++;

    if (imgr.watch == YES) {
	printf("FIRE %4d %s", rs->Agenda.list->id, imgr.stmt_name);
	PrintIds(rs->Agenda.list->variables);
	printf("\n");
	}

    /*Save the activation so it can be given back (deallocated)*/
    temp = rs->Agenda.list;

    rs->Agenda.list = rs->Agenda.list->next;

    ExecuteRHS(eo, temp->rhs->rhs->next, temp->variables);

    ReturnAct(temp);

    return(YES);

} /*FireRule*/

/*
 *GetBox
 */
struct mailbox *GetBox(eo, whose, type)
struct defeo *eo;
struct eo_id *whose;
int type;
{
    struct mailbox *search;

    search = (type == IN_BOX) ? eo->in_boxes : eo->out_boxes;

    for (;search != NULL; search = search->next) 
	if (CompareEO_ID(whose, search->eo_name, search->eo_id))
	    break;

    return(search);

} /*GetBox*/

/*
 *RecvMesg
 */
char *RecvMesg(box)
struct mailbox *box;
{
    char *mesg;

    if (DEBUG) {
	printf("RecvMesg: messages in box\n");
	PrintMesgs(box->mesgs);
	}

    if (box->status == MESG_INIT) {
	if (NotEmpty(box))
	    box->status = MESG_START;
	else
	    return(NULL);
	}

    if (box->status == MESG_END)
	return(NULL);

    if ((mesg = box->mesgs[0]) == NULL)
	return(NULL);

    box->mesgs = RemoveMessage(box->mesgs);

    return(mesg);

} /*RecvMesg*/

/*
 *SendMesg
 */
SendMesg(eo, token, eo_id)
struct defeo *eo;
char *token; 
int eo_id; 
{
    struct defeo *to;
    struct eo_id temp;
    struct mailbox *box;
    int type;

    temp.status = NUMERIC;
    temp.ni.id = eo_id;

    if ((box = GetBox(eo, &temp, OUT_BOX)) == NULL) {
	EH("Out_box missing.");
	exit(0);
	}

    if (box->status == MESG_INIT) {
	box->status = MESG_START;
	SendMesg(eo, mesg_start, eo_id);
	}

    if ((to = GetEO_by_name(&temp)) == NULL) {
	EH("SendMesg -> Unknown EO.");
	exit(0);
	} /*if*/

    temp.ni.id = eo->id;

    if ((box = GetBox(to, &temp, IN_BOX)) ==   NULL) {
	EH("SE: Missing in box.");
	exit(0);
	} /*if*/

    box->mesgs = AddMessage(box->mesgs, token);

    if (DEBUG) {
	printf("SendMesg: messages in box\n");
	PrintMesgs(box->mesgs);
	}

    return(YES);

} /*SendMesg*/

/* 
 *All buffers in the system are composed of two dimensional arrays similar
 *to the extern char **environ array in the UNIX environment.
 *
 *Each buffer is NULL terminated, so it must have one at the end of each
 *list with a null pointer.
 *
 *Example: 
 *   To hold the days of the week we allocate an array of 8 pointers
 *   to character arrays, each day of the week has a pointer from 0 - 6, and
 *   array pointer 7 points to NULL, to indicate an end of list (like LISP).
 */

/*
 *BufSize
 */
int BufSize(buffer)
char **buffer;
{
    int ct = 0;

    for (;*buffer != NULL; ++buffer, ++ct);

    return(ct);

} /*BufSize*/

/*
 *RemoveMessage
 */
char **RemoveMessage(buffer)
char **buffer;
{
    char **new, **tmp;
    int ct = 0;

    if (*buffer == NULL)
	return(buffer);

    new = (char **)MemAlloc(sizeof(char **) * BufSize(buffer));

    tmp = buffer + 1;

    while (*tmp != NULL)
	new[ct++] = *(tmp++);

    new[ct] = NULL;

    free(buffer);

    return(new);

} /*RemoveMessage*/

/*
 *AddMessage
 */
char **AddMessage(buffer, lst)
char **buffer;
char *lst;
{
    char **new;
    int ct = 0;

    new = (char **)MemAlloc(sizeof(char **) * (BufSize(buffer) + 2));

    for (;*buffer != NULL; ++buffer)
	new[ct++] = *buffer;

    new[ct++] = lst;
    new[ct] = NULL;

    free(buffer);

    return(new);

} /*AddMessage*/

/*
 *GetEO_by_name
 */
struct defeo *GetEO_by_name(name)
struct eo_id *name;
{
    struct defeo *eo_trav;

    for (eo_trav = imgr.defeo_lst; eo_trav != NULL; eo_trav = eo_trav->next)
	if (CompareEO_ID(name, eo_trav->name, eo_trav->id))
	    break;

    return(eo_trav);

} /*GetEO_by_name*/

/*
 *CreateMesg
 */
char *CreateMesg(pele)
struct element *pele;
{
    int offset = 0;

    for (; pele != NULL ; pele = pele->next) {

	if (pele->type == WORD)
	    sprintf(mesg_create + offset, " %s", pele->string);
	else
	    sprintf(mesg_create + offset, " %6.2f", pele->numeric);

	for (;mesg_create[offset] != '\0';offset++);
	} /*for*/

    return(MakeString(mesg_create));

} /*CreateMesg*/

/*
 *CreateToken
 */
char *CreateToken(type, fact, factid)
int type;
struct element *fact;
int factid;
{
    char *new;

    switch (type) {
	case MESG_RETRACT : 
			sprintf(mesg, "%d %d", MESG_RETRACT, factid);
			return(MakeString(mesg));
			break;

	case MESG_ASSERT :
			new = CreateMesg(fact);
			sprintf(mesg, "%d %d |%s", MESG_ASSERT, factid, new);
			return(MakeString(mesg));
			break;

	default :
			EH("Case out of bounds in CreateToken");
			break;

	} /*switch*/

    return(NULL);

} /*CreateToken*/

/*
 *GetEO_ID
 */
GetEO_ID(name)
char *name;
{
    struct defeo *eos;

    for (eos=imgr.defeo_lst; eos != NULL; eos = eos->next)
	if (strcmp(eos->name, name) == 0)
	    return(eos->id);

    return(-1);

} /*GetEO_ID*/
