/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: dataflow.c - contains the code to drive a fact through the net.
 *
 * Date: 6-16-87
 *
 */

#include <stdio.h>
#include "common.h"

/*
 *Dataflow_EO
 */
struct match *Dataflow_EO(eo, fact)
struct defeo *eo;
struct fact *fact;
{
    struct defrs *rs;
    struct match *matchs, *pmatchHead, *trav;

    pmatchHead = NULL;

    for (rs=eo->defrs_locals; rs != NULL; rs = rs->next)
	if (pmatchHead == NULL) 
	    trav = pmatchHead = Dataflow_RS(rs, fact);
	else { /*make a list of all patterns that match*/
	    matchs = Dataflow_RS(rs, fact);
	    for (; trav->next != NULL; trav = trav->next);
	    trav->next = matchs;
	    }

    return(pmatchHead);

} /*Dataflow_EO*/

/*
 *Dataflow_RS - drives a fact.
 */
struct match *Dataflow_RS(rs, fact)
struct defrs *rs;
struct fact *fact;
{
    struct pattern *trav;
    struct datalink *links;
    struct lr *lrs;
    struct match *matchs, *pmatchHead;
    struct values *bindings;

    pmatchHead = NULL;

    for (trav = rs->patterns; trav != NULL; trav = trav->next) {

	bindings = allocatestruct(values);
	bindings->opcode = SPECIAL;
	bindings->oe.origin = fact;
	bindings->whoset = fact->id;
	bindings->next = NULL;

	Compare(rs, trav->pattern, fact->entry, bindings);

	ReturnValues(bindings);

	if (rs->bindings == NULL) /*if null then there were no matches*/
	    continue;

	if (pmatchHead == NULL)
	    pmatchHead = matchs = allocatestruct(match);
	else {
	    matchs->next = allocatestruct(match);
	    matchs = matchs->next;
	    }

	matchs->slot = trav;
	matchs->next = NULL;

	/*For each drive down the net we create a new copy */
	for (links=trav->outflow; links != NULL; links=links->next) {

	    lrs = allocatestruct(lr);

	    /*If we have a binder then add it to the list*/
	    if (links->binder != NULL) {
		lrs->locals = ValuesCopy(links->binder, 1);
		lrs->locals->oe.origin = fact;
		lrs->locals->next = ValuesCopy(rs->bindings, ALL);
		} 
	    else
		lrs->locals = ValuesCopy(rs->bindings, ALL);

	    lrs->next = NULL;
	    AssertFlow(rs, links->path, lrs, links->side); /*normal pattern*/
	    } /*for*/

	ReturnValues(rs->bindings);
	rs->bindings = NULL;

	} /*for each pattern in the list*/

    return(pmatchHead); 

} /*DataFlow*/

/*
 *MergeLr - Merges two token lists.
 */
struct lr *MergeLr(lr1, lr2)
struct lr *lr1, *lr2;
{
    struct lr *new; /*new lr being created*/
    struct values *trav;

    /* See if all the variables values that should exist, indeed exist*/
    if (VarEqual(lr1->locals->next, lr2->locals->next) == NO)
	 return(NULL);

    new = allocatestruct(lr);
    new->locals = ValuesCopy(lr1->locals, ALL);

    for (trav=new->locals; trav->next != NULL; trav = trav->next);

    trav->next = ValuesCopy(lr2->locals, ALL);
	 
    return(new);

} /*MergeLr*/


/*
 *VarEqual - are two lists of variables equal 
 */
VarEqual(list1, list2)
struct values *list1, *list2;
{
    struct values *trav, *pv;

    for (trav=list2; trav != NULL;) {
	switch (trav->opcode) {
	    case SBIND :
		for (pv=list1; pv != NULL; pv = pv->next) {
		    if ((pv->name == trav->name) && (pv->opcode == SBIND)) {
			if (pv->type != trav->type)
			    return(NO);
			else switch (pv->type) {
			    case NUMERIC :
				if (pv->numeric != trav->numeric) 
				    return(NO);

				break;

			    case WORD :
				if (strcmp(pv->string,trav->string) != 0) 
				    return(NO);

				break;

			    default :
				EH("Compare OUT of bounds: MergeLr");
				return(NO);
			    } /*switch*/
			    
			break; /*Out of the for loop*/

			} /*If names are the same*/

		    } /*for elements in list two*/

		trav = trav->next;

		break;
			
	    default :
		trav = trav->next;
		break;
	    } /*switch*/

	} /*for trav*/

    return(YES);

} /*VarEqual*/

/*
 *AssertFlow - when an assert is performed, this function is called to drive
 *	       to create the proper execution sequence.
 */
AssertFlow(rs, pnode, new, sside)
struct defrs *rs;
struct decidernode *pnode; /*starting node to drive data down*/
struct lr *new; /*the new set of lrs being added to this nodes sside*/
int sside; /*the start side for the run*/
{
    struct lr *plr, *dlr;
    struct lr *cyclelr;
    struct lr *newlr;

    if (pnode->outflow == NULL) {
	AddActivation(rs, pnode, new->locals);
	ReturnLrs(new);
	return(YES);
	} /*if*/

    cyclelr = (sside == RHS) ? pnode->lhs : pnode->rhs;

    for (dlr=new; dlr != NULL; dlr=dlr->next) 
	for (plr = cyclelr; plr != NULL; plr = plr->next) 
	    if ((newlr=MergeLr(dlr, plr)) == NULL)
		continue;
	    else
		AssertFlow(rs, pnode->outflow, newlr, pnode->side);

    for (dlr = new; dlr->next != NULL; dlr = dlr->next);

    if (sside == RHS) {
	dlr->next = pnode->rhs;
	pnode->rhs = new;
	} /*if*/
    else {
	dlr->next = pnode->lhs;
	pnode->lhs = new;
	}

    return(YES);

} /*AssertFlow*/

/*
 *AddActivation - if a rule is ready to fire (activated) then it is placed in
 *                the agenda. This is done using this function.
 */
AddActivation(rs, pnode, values)
struct defrs *rs;
struct decidernode *pnode; /*RHS of the rule*/
struct values *values; /*variable bindings*/
{
    struct activation *new, *trav, *last;

    new = allocatestruct(activation);
    new->variables = ValuesCopy(values, ALL);
    new->id = rs->Agenda.id++; /*give it an activation number*/
    new->rhs = pnode; /*assign its RHS*/
    new->priority = pnode->rhs->locals->numeric;
    imgr.stmt_name = pnode->rhs->locals->name;

    if (rs->Agenda.list == NULL) {
	new->next = rs->Agenda.list; /*add it to the list*/
	rs->Agenda.list = new;
	}
    else {

	for (trav = rs->Agenda.list; trav != NULL; trav = trav->next) {
	    if (trav->priority <= new->priority) {
		if (rs->Agenda.list != trav) {
		    last->next = new;
		    new->next = trav;
		    } /*rs*/
		else  {
		    new->next = rs->Agenda.list; /*add it to the list*/
		    rs->Agenda.list = new;
		    } /*else*/

		break;
		} /*if <= */

	    if (trav->next == NULL) {
		trav->next = new;
		new->next = NULL;
		break;
		}

	    last = trav;
	   } /*for trav != NULL*/

	} /*else */

    return(YES);

} /*AddActivation*/
