/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: eo.c
 *
 * Date: 6-16-87
 *
 */

#include <stdio.h>
#include "common.h"
#include "comm.h"

/*
 *CurrentEO
 */
CurrentEO()
{
    if (imgr.current_eo == NULL) {
	EH("No current expert object.");
	return(NO);
	}

    return(YES);

} /*CurrentEO*/

/*
 *CreateInBoxes
 */
CreateInBoxes(eos)
struct defeo *eos;
{
    struct mailbox *out;
  
    for (;eos != NULL; eos = eos->next)
	for(out = eos->out_boxes;out != NULL; out = out->next)
	    AddEO_to_EO_Boxes(out->eo_name, eos->name, IN_BOX);

} /*CreateInBoxes*/

/*
 *AddEO_to_EO_Boxes
 */
AddEO_to_EO_Boxes(eo_name, eo_new, direction)
char *eo_name;
char *eo_new;
int direction;
{
    struct mailbox *trav;
    struct defeo *eo;
    struct eo_id temp;

    temp.status = WORD;
    temp.ni.name = eo_name;

    /*Get the EO*/
    if ((eo = GetEO_by_name(&temp)) == NULL) {
	sprintf(error_sb,"SE:: EO <%s> does not exist.", eo_name);
	EH(error_sb);
	return(NO);
	}

    trav = (direction == IN_BOX) ? eo->in_boxes : eo->out_boxes;

    for (; trav != NULL; trav=trav->next)
	if (strcmp(eo_new, trav->eo_name) == 0)
	    return(NO);

    trav = allocatestruct(mailbox);

    /*out_box eo id still has to be set later on*/

    /*
      We begin execution locally first, so that the initial state
      of the expert object can be developed, then from there it will
      begin to check its in_boxes for new data. So initially we set
      its in_boxes status to MESG_END which means initially that
      it doesn't know anything about anyone else (cycle one).
     */

    trav->status = MESG_END;
    trav->eo_name = eo_new;
    trav->eo_id = (direction == IN_BOX) ? GetEO_ID(eo_new) : -1;
    trav->mesgs = INIT_BUF;
    *trav->mesgs = NULL; /*no mail*/
    trav->next = (direction == IN_BOX) ? eo->in_boxes : eo->out_boxes;

    if (direction == IN_BOX)
	eo->in_boxes = trav;
    else
	eo->out_boxes = trav;

    return(YES);

} /*AddEO_to_EO_Boxes*/

/*
 *AddEO_to_Ruleset
 */
AddEO_to_Ruleset(name)
char *name;
{
    struct eo_id *trav;

    for (trav=imgr.current_rs->out_boxes; trav != NULL; trav=trav->next)
	if (name == trav->ni.name)
	    return(NO);

    trav = allocatestruct(eo_id);

    trav->status = WORD;
    trav->ni.name = name;

    trav->next = imgr.current_rs->out_boxes;
    imgr.current_rs->out_boxes = trav;

    return(YES);

} /*AddEO_to_Ruleset*/

/*
 * RS_Exists
 */
RS_Exists(name)
char *name;
{
    struct defrs *trav;

    for (trav = imgr.defrs_lst; trav != NULL; trav = trav->next)
	 if (name == trav->name)
	     return(YES);

    return(NO);

} /*RS_Exists*/
