/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: execute.c - contains all of the functions necessary for the evaluation of
 *            the right hand side of a rule.
 *
 * Date: 6-16-87
 *
 */

#include <stdio.h>
#include "common.h"
#include "comm.h"

/*
 *ExecuteRHS - Executes all the functions on the RHS of a rule.
 */
ExecuteRHS(eo, commands, locals)
struct defeo *eo;
struct lr *commands;
struct values *locals;
{
    for (; commands != NULL; commands = commands->next) 
	rhs[commands->locals->whoset](eo, commands->locals->next, locals);
	
} /*ExecuteRHS*/

/*
 *MSendMesg - send a message to a list of EOs.
 */
MSendMesg(from, mesg, eo_ids)
struct defeo *from;
char *mesg;
struct eo_id *eo_ids;
{
    int eo_id;

    for(; eo_ids != NULL; eo_ids = eo_ids->next) {

	if (eo_ids->status == NUMERIC)
	    eo_id = eo_ids->ni.id;
	else 
	    eo_id = GetEO_ID(eo_ids->ni.name);

	SendMesg(from, mesg, eo_id);
	} /*for*/

    return(YES);

} /*MSendMesg*/

/*
 *ExecuteStop - shut down the expert object.
 */
ExecuteStop(eo, command, locals)
struct defeo *eo;
struct values *command; /*list of variables to retract*/
struct values *locals;  /*Local variables list bound: state*/
{
    struct mailbox *out;

    eo->status = EO_STOPPED;

    for (out = eo->out_boxes; out != NULL; out = out->next)
	SendMesg(eo, mesg_stopped, out->eo_id);

    return(YES);

} /*ExecuteStop*/

/*
 *ExecuteRetract - executes the retract function.
 */
ExecuteRetract(eo, command, locals)
struct defeo *eo;
struct values *command; /*list of variables to retract*/
struct values *locals;  /*Local variables list bound: state*/
{
    struct values *pv;
    char *mesg;
    struct eo_id *eos;


    eos = command->oe.eo_id;

    /*For each variable listed, retract the fact it is bound to*/
    for (;command != NULL; command = command->next) {

	if ((pv=LocalBind(command->string, SBIND, locals)) == NULL) {
	    sprintf(error_sb,"Retract Failed: ?%s not found.", command->string);
	    EH(error_sb);
	    return(NO);
	    }

	if (pv->oe.origin == NULL) {
sprintf(error_sb, "Retract Failed: ?%s not bound to a fact.", command->name);
	    EH(error_sb);
	    return(NO);
	    } /*if*/

	mesg=CreateToken(MESG_RETRACT,(struct element *)NULL,pv->oe.origin->id);
	MSendMesg(eo, mesg, command->oe.eo_id);

        } /*for each command argument*/

    return(YES);

} /*ExecuteRetract*/

/*
 *ExecutePrintOut - Executes the RHS command "printout"
 */
ExecutePrintOut(eo, command, locals)
struct defeo *eo;
struct values *command;
struct values *locals;
{
    struct values *pv;

    for (;command != NULL; command=command->next) {

	switch (command->opcode) {
	    case NUMERIC :
		printf("%f", command->numeric);
		break;

	    case WORD :
		printf("%s", command->string);
		break;

	    case SBIND :
		if ((pv=LocalBind(command->string, SBIND, locals)) == NULL) {
sprintf(error_sb,"Variable %s not found for PRINTOUT command.",command->string);
		    EH(error_sb);
		    return(NO);
		    }

		if (pv->type == NUMERIC)
		    printf("%f", pv->numeric);
		else
		    printf("%s", pv->string);
		break;

	    } /*switch*/

	printf(" ");

	} /*for each element of the printout*/

    printf("\n");
    
} /*ExecutePrintOut*/

/*
 *ExecuteAssert - executes the RHS command ExecuteAssert
 */
ExecuteAssert(eo, command, locals)
struct defeo *eo;
struct values *command; /*command arguments*/
struct values *locals; /*local variables list*/
{
    struct element *last;
    struct element *pelenew, *pelehead;
    struct values *pv; /*return values from LocalBind*/
    char *mesg;

    pelehead = NULL;

    do {

	if (command->opcode == EO_IDS) {
	    mesg = CreateToken(MESG_ASSERT, pelehead, COMPUTE_FACT_ID(eo));
	    MSendMesg(eo, mesg, command->oe.eo_id);
	    break;
	    } /*if*/

	if (pelehead == NULL) 
	    pelenew = pelehead = allocatestruct(element);
	else {
	    last = pelenew;
	    pelenew->next = allocatestruct(element);
	    pelenew = pelenew->next;
	    } /*else*/

	/*create the appropriate elements*/

	switch (command->opcode) {

	    case WORD :
		pelenew->type = WORD;
		pelenew->string = command->string;
		break;

	    case NUMERIC :
		pelenew->type = NUMERIC;
		pelenew->numeric = command->numeric;
		break;

	    case SBIND :
		if ((pv=LocalBind(command->string, SBIND, locals)) == NULL) {
sprintf(error_sb,"Variable %s not found for ASSERT command.", command->string);
		    EH(error_sb);
		    goto ASSERT_ERROR;
		    }

		pelenew->type = pv->type;

		if (pelenew->type == WORD)
		    pelenew->string = pv->string;
		else
		    pelenew->numeric = pv->numeric;

		break;

	    default :
		EH("ASSERT FIELD OUT OF BOUNDS IN SWITCH");
		return(NO);
		break;
	    } /*switch*/

	command = command->next;
	} while (command != NULL);

	pelenew->next = NULL;
		
	return(YES);
		
ASSERT_ERROR: /*Error, clean up*/
	pelenew->next = NULL;
	ReturnElements(pelehead);
	return(NO);

} /*ExecuteAssert*/

/*
 *RetractFlow - retracts a fact from the working memory.
 */
RetractFlow(eo, fact)
struct defeo *eo;
struct fact *fact; /*fact that we are retracting*/
{
    struct match *pms; /*Pattern Matches*/
    struct datalink *flow; /*links from each pattern*/
    struct defrs *rs_trav;

    /*



    HAS TO BE WORKED ON -- RETRACT FROM ALL RULE SETS



    */

    for (rs_trav = eo->defrs_locals;rs_trav != NULL; rs_trav = rs_trav->next)
	CleanAgenda(rs_trav, fact->id, YES); 
    
    /*For each pattern the fact matched*/
    for (pms=fact->list; pms != NULL; pms=pms->next) 
	for (flow=pms->slot->outflow; flow != NULL; flow = flow->next) 
	    RemoveId(flow->path, flow->side, fact->id);

    return(YES);

} /*RetractFlow*/

/*
 *RemoveId
 */
RemoveId(des, side, id)
struct decidernode *des;
int side;
int id;
{
    struct lr *head, *binds, *last, *temp;

    for (;des->outflow != NULL;des=des->outflow){

	head = binds = (side == RHS) ? des->rhs : des->lhs;

	/*For each binding list in the side*/
	for (last=NULL; binds != NULL; ) 
	    if (AnyId(id, binds->locals) == NO) {
		last = binds;
		binds = binds->next;
		}
	    else {

		/*Remove locals pair*/

		temp = binds;

		if (last == NULL) 
		    binds = head = binds->next;
		else 
		    binds = last->next = binds->next;

		FreeLr(temp);

		} /*Each binding list*/

	if (side == RHS)
	    des->rhs = head;
	else
	    des->lhs = head;

	side = des->side; /*Next side we are entering from*/
    } /*for*/

} /*RemoveId*/
	
/*
 *AnyId - checks a token list to see if a particular system id exists.
 */
AnyId(id, values)
int id; /*ID we are looking for*/
struct values *values; /*List of values*/
{

    for (;values != NULL; values = values->next) {
	if (values->opcode != SPECIAL)
	    return(NO);

	if (values->whoset == id)
	    return(YES);
	} /*for*/

    return(NO);

} /*AnyId*/
