/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: fact.c - fact related functions.
 *
 * Date: 6-16-87
 *
 */

#include <stdio.h>
#include "common.h"
#include "comm.h"

/*
 *RemoveFact - removes a fact from the fact list.
 */
RemoveFact(eo, fact)
struct defeo *eo;
struct fact *fact;
{
    if (fact == eo->factlist) {

	eo->factlist = eo->factlist->next;

	if (eo->factlist != NULL)
	    eo->factlist->prev = NULL;
	}
    else {
	if (fact->next != NULL) 
	    fact->next->prev = fact->prev;

	fact->prev->next = fact->next;
	}

    if (fact == eo->factptr)
	eo->factptr = eo->factptr->prev;

    if (fact == eo->factnew)
	eo->factnew = eo->factnew->prev;
	    
    ReturnElements(fact->entry);

    free(fact);

 } /*For each command*/

/*
 *RetractFact - the interactive retract command.
 */
RetractFact()
{
    int factid;
    struct fact *trav;

    GetToken(YES);

    while (token.opcode == NUMERIC) {

	factid = (int)token.numeric;

	for (trav=imgr.current_eo->factlist; trav != NULL; trav = trav->next)
	    if (factid == trav->id)
		break;

	if (trav != NULL) {
	    RetractFlow(imgr.current_eo, trav);
	    RemoveFact(imgr.current_eo, trav);
	    } /*trav*/

	GetToken(YES);

	} /*for each fact id given*/

} /*RetractFact*/

/*
 *AddFacts - Add a new Deffacts statement to the db.
 *      cond -> 0 : Add deffacts statement to its structure.
 *	cond -> 1 : Add the name to the list of known deffact statements.
 *      cond -> 2 : Add the parsed facts to the deffacts structure.
 */
AddFacts(cond, name, stmt, facts)
int cond;
char *name; 
char *stmt; 
char **facts;
{
    struct deffacts *newfacts;

    newfacts=imgr.current_eo->deffacts;

    for (; newfacts != NULL; newfacts = newfacts->next)
	if (newfacts->name == name) {
	    if (cond != 1)
		break;

	    sprintf(error_sb, "Deffact %s redeclared.", name);
	    EH(error_sb);
	    return(NO);
	    }
    
    switch (cond) {
	case 0 : /*add stmt*/
		newfacts->stmt = MemAlloc(strlen(stmt) + 1);
		strcpy(newfacts->stmt, stmt);
		break;

	case 1 : /*add name*/
		newfacts = allocatestruct(deffacts);
		newfacts->name = name;
		newfacts->next = imgr.current_eo->deffacts;
		imgr.current_eo->deffacts = newfacts;
		break;

	case 2 : /*add parsed facts*/
		newfacts->facts = facts;
		break;
	} /*switch*/

    return(YES);
} /*AddFacts*/

/*
 *ProcessFacts
 */
ProcessFacts(eo)
struct defeo *eo;
{

    /*loop while still entries in agenda or facts to be driven*/
    while (eo->factptr != NULL) {

	if (eo->factnew != NULL) {
	    while (eo->factptr != NULL) {
		eo->factptr->list = Dataflow_EO(eo, eo->factptr);
		eo->factptr = eo->factptr->next;
		} /*each fact*/

	    while (eo->factnew->next != NULL)
		eo->factnew = eo->factnew->next;

	    } /*if facts*/
	} /*while*/

    return(YES);

} /*ProcessFacts*/

/*
 *FactExists - does a fact exist in the factlist.
 */
FactExists(eo, fact)
struct defeo *eo;
struct element *fact;
{
    struct fact *pfacttmp;

    for (pfacttmp=eo->factlist; pfacttmp != NULL; pfacttmp=pfacttmp->next)
	if (Equal(fact, pfacttmp->entry) == YES)
	    return( YES );

    return( NO );

} /*FactExists*/

/*
 *GetFact
 */
struct fact *GetFact(eo, fact_id)
struct defeo *eo;
int fact_id;
{
    struct fact *trav;

    for (trav = eo->factlist; trav != NULL; trav = trav->next)
	if (trav->id == fact_id)
	    break;

    return(trav);

} /*GetFact*/
