/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: init.c - contains the code for the initialization of the system when it starts up.
 *
 * Date: 6-16-87
 *
 */

#include <stdio.h>
#include "common.h"
#include "init.h"
#include "comm.h"

/*
 *Init()
 */
Init(argc, argv)
int argc;
char *argv[];
{
    int i;

    /*Set the mode for the interface manager*/
    imgr.watch = YES ;

    sprintf(mesg_start, "%d", MESG_START);
    sprintf(mesg_end, "%d", MESG_END);
    sprintf(mesg_stopped, "%d", MESG_STOPPED);

    /*Always initialize the hash table first, before all other initializations*/
    for (i=0; i < HASHSIZE; i++)
	ht[i] = NULL;

    rhs[ RHSASSERT ] = ExecuteAssert;
    rhs[ RHSPRINTOUT ] = ExecutePrintOut;
    rhs[ RHSRETRACT ] = ExecuteRetract;
    rhs[ RHSSTOP ] = ExecuteStop;

    laFill = laEmpty = laCt = 0;

    imgr.error = NO;

    imgr.eo_ids = 1;
    imgr.current_eo = NULL;
    imgr.defeo_lst = NULL;

    imgr.current_rs = NULL;
    imgr.defrs_lst = NULL;

    imgr.pmatchList = NULL;

    if (argc == 2) {
	imgr.fn = HashString(argv[1]);
	imgr.ln = 1;
	}
    else {
	imgr.fn = NULL;
	imgr.ln = 0;
	} /*Interactive mode*/

    ASSERT = HashString("assert");
    DECLARE = HashString("declare");
    DEFEO = HashString("defeo");
    DEFFACTS = HashString("deffacts");
    DEFRS = HashString("defrs");
    DEFRULE = HashString("defrule");
    PRINTOUT = HashString("printout");
    PRIORITY = HashString("priority");
    RETRACT = HashString("retract");
    SBNULL = HashString("");
    SBAND = HashString("and");
    STOP = HashString("stop");

} /*Init*/

/*
 *CreateRS (see struct definition for comments about defrs declaration)
 */
struct defrs *CreateRS()
{
    struct defrs *new;

    new = allocatestruct(defrs);

    new->priority = RS_PRIORITY_DEFAULT;

    new->Agenda.id = 0;
    new->Agenda.list = NULL;

    new->out_boxes = NULL;
    new->bindings = NULL;
    new->patterns = NULL;
    new->defrules = NULL;
    new->next = NULL;

    return(new);

} /*CreateRS*/

/*
 *CreateEO
 */
struct defeo *CreateEO()
{
    struct defeo *new;

    if (imgr.current_eo != NULL)
	imgr.current_eo->defrs_locals = imgr.defrs_lst;

    imgr.defrs_lst = NULL;
    imgr.current_rs = NULL;

    new = allocatestruct(defeo);

    new->total_rules = 0;

    new->status = EO_INIT;

    new->id = imgr.eo_ids++;

    new->in_boxes = NULL;
    new->out_boxes = NULL;

    new->defrs_locals = NULL;

    return(new);

} /*CreateEO*/

/*
 *InitEO_Factlist
 */
InitEO_Factlist(new)
struct defeo *new;
{

    new->factid = 1;

    /*initial-fact*/
    new->factlist = allocatestruct(fact);
    new->factlist->id = new->factid;
    new->factlist->list = NULL;
    new->factlist->next = new->factlist->prev = NULL;
    new->factlist->entry = allocatestruct(element);
    new->factlist->entry->type = WORD;
    new->factlist->entry->string = HashString("initial-fact");
    new->factlist->entry->id = new->factid++;
    new->factlist->entry->next = NULL;
    new->factptr = new->factnew = new->factlist;

    return(YES);

} /*InitEO_Factlist*/
