/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: inter.c - contains the code for the interactive interface to the system.
 *
 * Date: 6-16-87
 *
 */

#include <stdio.h>
#include "common.h"
#include "inter.h"
#include "comm.h"

/*
 *Prompt
 */
Prompt()
{
    static prompt[ PROMPT_LEN ];
    char *eo;
    char *rs;

    if (imgr.update_prompt == YES) {
	if (imgr.current_eo != NULL)
	    eo = imgr.current_eo->name;
	else
	    eo = "<>";

	if (imgr.current_rs != NULL)
	    rs = imgr.current_rs->name;
	else
	    rs = "";

	imgr.update_prompt = NO;

	sprintf(prompt, "<%s><%s> -> ", eo, rs);
	}

    if (imgr.show_prompt == YES)
	printf(prompt);

} /*Prompt*/
    
/*
 *FetchCommand - fetchs a command from the input stream.
 */
FetchCommand(fp)
FILE *fp; /*input file name*/
{
    int ct;
    int ch;
    int paren_ct;

    imgr.error = NO;
    imgr.ct = 0;

    DoForever {

	ct = 0;

	/*First burn off all white space */
	for (ch=getc(fp);IsFiller(ch)||(ch == EOL)||(ch == ';'); ch=getc(fp)) {
	
	    if (ch == ';') {
		while ((ch=getc(fp) != EOL) && (ch != EOF)); 
		continue;
		}

	    if (ch == EOL) 
		Prompt();

	    } /*for*/

	if (ch == EOF)
	    return(EOF);

	if ((imgr.input[ct] = ch) != '(') {

	    EH("Missing '(' .");

	    while (((ch = getc(fp)) != EOL) && (ch != EOF));

	    if (ch == EOF)
		return(EOF);

	    ungetc(EOL, fp);

	    imgr.ct = 0;
	    imgr.error = NO;
	    continue;
	    }

	imgr.input[ct++] = ch;

 	/* Count parenthesis: matching opening and closing till all are closed*/

	for (paren_ct = 1; paren_ct != 0; ct++) {
	    imgr.input[ct] = getc(fp);
	  
	    if (imgr.input[ct] ==  ')')
		paren_ct--;
	    else if (imgr.input[ct] == '(')
		paren_ct++;
	    else if (imgr.input[ct] == EOF)
		return(EOF);

	    } /*for*/

	imgr.input[ct] = EOS;

	break;
	} /*DoForever*/

    return(YES);

} /*FetchCommand*/

/*
 *InterActive - execute the interactive interpreter.
 */ 
InterActive()
{
    
    printf("Interactive mode.\n");

    imgr.update_prompt = YES;
    imgr.show_prompt = YES;

    Prompt();

    /*Set up the parameters for reading in the input characters*/
    imgr.input = MemAlloc( BUFFER_SIZE );
    imgr.current_eo = NULL;
    imgr.current_rs = NULL;
    imgr.save = YES;

    DoForever {

	if (FetchCommand(stdin) == EOF) {
	    sprintf(error_sb, "Use the (%s) command to exit.", EXITCMD);
	    EH(error_sb);
	    }

	if (imgr.error == NO)
	    ExecuteCommand();

	} /*all commands*/

    free(imgr.input);

} /*InterActive*/

/*
 *ExecuteCommand() - Execute the interpreter level commands.
 */
ExecuteCommand()
{
    extern int ch;		/*input character to scanner*/
    int low;			/*used to hold the low factid*/
    char *mesg;
    struct deffacts *dftmp;
    struct defrule  *drtmp;
    struct element *pele;

    ch = GetIA();		/*fetch next character*/

    imgr.error = NO;

    GetToken(YES);

    if (token.opcode != LPAREN)
	EH("Missing '(' ");

    GetToken(YES);

    if (token.opcode == WORD) {

	/*Figure out which command was read in and execute it*/

	if (strcmp( EXITCMD , token.string) == 0) 
	    if (imgr.save == YES)
	        exit(0);
	    else  {
		printf("The present rules and facts have not been saved.\n");
		imgr.save = YES;
		return(NO);
		} /*else*/

	if (strcmp(FACTSCMD, token.string) == 0) {
	    if (CurrentEO() == YES)
		DisplayFacts(imgr.current_eo);
	    } /*if*/
	else if (token.string == ASSERT) {

	    if (CurrentEO() == NO)
		return(NO);

	    GetToken(YES);

	    if (token.opcode != WORD) 
		EH("Bad Format: Assert: (assert \"<fact>\").");
	    else {
		if ((pele=ProcessNewString(token.string)) != NULL) {
        mesg = CreateToken(MESG_ASSERT, pele, COMPUTE_FACT_ID(imgr.current_eo));
		    SendMesg(imgr.current_eo, mesg, imgr.current_eo->id);
		    } /*if*/
		} /*else if*/

	    } /*if ASSERT*/
	else if (token.string == RETRACT) {
	    RetractFact();
	    } /*if RETRACT*/
	else if (strcmp(CREATERSCMD, token.string) == 0) 
	    Inter_CreateRS();
	else if (strcmp(CREATEEOCMD, token.string) == 0)
	    Inter_CreateEO();
	else if (strcmp(LOADCMD, token.string) == 0)  {
	    imgr.show_prompt = NO;
	    imgr.save = NO;
	    Load();
	    imgr.show_prompt = YES;
	    imgr.update_prompt = YES;
	    }
	else if (strcmp(SAVECMD, token.string) == 0)  {
	    imgr.save = YES;
	    Save();
	    }
	else if (strcmp(RESETCMD, token.string) == 0) 
	    Reset();
	else if (strcmp(RUNCMD, token.string) == 0) 
	    Run();
	else if (token.string == DEFFACTS) {

	    if (CurrentEO() == NO)
		return(NO);

	    /*Flag that the state has been changed*/
	    imgr.save = NO;
	    low = imgr.current_eo->factid;

	    dftmp = imgr.current_eo->deffacts;

	    if (ParseFacts() == NO) {
		if (dftmp == imgr.current_eo->deffacts) 
		    return(NO);

		dftmp = imgr.current_eo->deffacts;
		imgr.current_eo->deffacts = imgr.current_eo->deffacts->next;	
		free(dftmp);
		return(NO);
		} /*clean up if an error occurs*/

	    AddFacts(0, imgr.stmt_name, imgr.input, (char **)NULL);

	    } /*DEFFACTS*/
	else if (token.string == DEFEO )  {
	    GetToken(YES);
	    ParseEO();
	    }
	else if (token.string == DEFRULE) {

	    if (CurrentEO() == NO)
		return(NO);

	    if (imgr.current_rs == NULL) {
		EH("No current rule set.");
		return(NO);
		}

	    imgr.save = NO; /*record change of state*/

	    drtmp = imgr.current_rs->defrules;

	    if (ParseRule() == NO) {

		if (drtmp == imgr.current_rs->defrules)
		    return(NO);

		drtmp = imgr.current_rs->defrules;
		imgr.current_rs->defrules = imgr.current_rs->defrules->next;
		free(drtmp);
		return(NO);
		}

	    AddRule(imgr.stmt_name, imgr.input, NO); /*add the rule to the db*/

	    } /*DEFRULE*/
	else if (strcmp(token.string, PPCMD) == 0)
	    PrettyPrint();
	else if (strcmp(token.string, WATCHCMD) == 0) 
	    imgr.watch = YES;
	else if (strcmp(token.string, UNWATCHCMD) == 0)
	    imgr.watch = NO;
	else if (strcmp(token.string, AGENDACMD) == 0)
	    PrintAgenda_RS(imgr.current_rs);
	else if (strcmp(token.string, EDITRSCMD) == 0) {
	    EditRS();
	    imgr.update_prompt = YES;
	    }
	else if (strcmp(token.string, EDITEOCMD) == 0) {
	    EditEO();
	    imgr.update_prompt = YES;
	    }
	else if (strcmp(token.string, PNETCMD) == 0) 
	    PrintNet(imgr.current_rs);
	else {
	    sprintf(error_sb, "Unknown command %s.", token.string);
	    EH(error_sb);
	    }
	}
    else 
	EH("Unknown command format.");

} /*ExecuteCommand*/

/*
 *Load - loads files into the system.
 */
Load()
{
    FILE *fp;
    char *temp;
    int ct;

    /*Save the old input buffer, for now we are reading in a new file*/
    temp = MemAlloc(strlen(imgr.input) + 1);
    strcpy(temp, imgr.input);

    for (GetToken(YES); token.opcode == WORD; GetToken(YES)) {

	if (token.opcode != WORD) {
	    EH("Bad argument given to Load command.");
	    return(NO);
	    }


	if ((fp = fopen(token.string, "r")) == NULL) {
	    sprintf(error_sb, "%s doesn't exist.", token.string);
	    EH(error_sb);
	    }
	else {
	    ct = imgr.ct; /*save old character pointer*/

	    while (FetchCommand(fp) != EOF) {
		if (imgr.error == NO)
		    ExecuteCommand();
		imgr.error = NO;
		} /*for each command*/

	    imgr.ct = ct; /*restore the old state*/
	    strcpy(imgr.input, temp);
	    ch = imgr.input[ct - 1];
	    fclose(fp);
	    }
	}  /*for*/

    free(temp); /*clean up the details of life*/

} /*Load*/

/*
 * Save - save the present values stored in the input buffers.
 */
Save()
{
    FILE *fp;
    int i;
    int d;

    GetToken(YES);

    if (token.opcode != WORD) {
	EH("Missing filename in save");
	return(NO);
	}

    if ((fp=fopen(token.string, "r")) != NULL) {
	printf("%s already exists.\n", token.string);
	fclose(fp);
	i = strlen(token.string);
	d = 1;
	if (IsNumeric(token.string[i - 1])) {
	    sscanf(token.string + i - 1, "%d", &d);
	    token.string[i - 1] = EOS;
	    d++;
	    } /*Is ended by a number*/

	sprintf(error_sb, "%s%d", token.string, d);
	token.string = error_sb;
	} /*already exists*/

    SaveFile(token.string);

} /*Save*/


/*
 *SaveFile - save the contents of the system in the file name given.
 */
SaveFile(name)
char *name; /*file name to save the environment in*/
{
    FILE *fp; /*file to be opened*/
    struct deffacts *facts;
    struct defrule *rule;

    printf("Saving file under %s.\n", name);

    if ((fp=fopen(name, "w")) == NULL) {

	sprintf(error_sb, "%s could not be opened for write.", name);
	EH(error_sb);
	return(NULL);

	} /*if*/

    /*Write all deffacts to the file*/
    for (facts=imgr.current_eo->deffacts; facts != NULL; facts = facts->next) {
	printf("Saving deffacts: %s.\n", facts->name);
	fprintf(fp, "%s\n", facts->stmt);
	}

    /*Write all defrules to the file*/
    for (rule=imgr.current_rs->defrules; rule != NULL; rule = rule->next) {
	printf("Saving rule: %s.\n", rule->name);
	fprintf(fp, "%s\n", rule->stmt);
	}

    printf("All rules and deffacts have been written to %s\n", token.string);
    fclose(fp);

} /*SaveFile*/

/*
 *AddRule - Add a new defrule to the db.
 *
 *	 cond == 0: Add the defrule's statement to its structure.
 *	 cond == 1: Add the name of the defrule to the list of known defrules.
 */
AddRule(name, stmt, cond)
char *name, *stmt;
int cond;
{
    struct defrule *newrule;

    /*Does it exist?*/
    for (newrule=imgr.current_rs->defrules; newrule != NULL; newrule=newrule->next)
	if (strcmp(newrule->name, name) == 0) {
	    if (cond == 0)
		break;

	    sprintf(error_sb, "Rule %s redeclared.", name);
	    EH(error_sb);
	    return(NO);
	    }
    
    switch (cond) {
	case 0 :
		if (newrule == NULL) {
		    EH("Rule not found - can not add statement.");
		    return(NO);
		    }

		newrule->stmt = MemAlloc(strlen(stmt) + 1);
		strcpy(newrule->stmt, stmt);
		break;

	case 1 :
		imgr.current_dr = newrule = allocatestruct(defrule);
		newrule->priority = DEFRULE_PRIORITY_DEFAULT;
		newrule->name = name;
		newrule->next = imgr.current_rs->defrules;
		imgr.current_rs->defrules = newrule;
		break;

	} /*switch*/

    return(YES);

} /*AddRule*/

/*
 *EditEO
 */
EditEO()
{
    struct defeo *eo;
    struct eo_id name;

    GetToken(YES);

    if (token.opcode != WORD) {
	EH("Expected a word.");
	return(NO);
	}

    name.status = WORD;
    name.ni.name = token.string;

    if ((eo = GetEO_by_name(&name)) == NULL) {
	sprintf(error_sb, "Expert object <%s> doesn't exist.", token.string);
	EH(error_sb);
	return(NO);
	}

    if (imgr.current_eo != NULL) {
	imgr.current_eo->defrs_locals = imgr.defrs_lst;
	}

    imgr.current_eo = eo;
    imgr.defrs_lst = eo->defrs_locals;
    imgr.current_rs = NULL;

    return(YES);

} /*EditEO*/

/*
 *EditRS
 */
EditRS()
{
    struct defrs *rs;

    GetToken(YES);

    if (token.opcode != WORD) {
	EH("Expected a word.");
	return(NO);
	}

    if (CurrentEO() == NO)
	return(NO);

    for (rs = imgr.defrs_lst; rs != NULL; rs = rs->next) 
	if (strcmp(rs->name, token.string) == 0) {
	    imgr.current_rs = rs;
	    break;
	    }

    if (rs == NULL) {
	sprintf(error_sb, "Rule set <%s> not found.", token.string);
	EH(error_sb);
	return(NO);
	}

    GetToken(YES);

    if (token.opcode == NUMERIC)
	rs->priority = token.numeric;

    return(YES);

} /*EditRS*/

/*
 *Inter_CreateEO
 */
Inter_CreateEO()
{
    struct defeo *eo;
    struct eo_id name;

    GetToken(YES);

    if (token.opcode != WORD) {
	EH("A name of an expert object must be of type word.");
	return(NO);
	}

    name.status = WORD;
    name.ni.name = token.string;

    if (GetEO_by_name(&name) != NULL) {
	sprintf(error_sb, "EO <%s> already exists.", token.string);
	EH(error_sb);
	}

    eo = CreateEO();
    eo->next = imgr.defeo_lst;
    imgr.defeo_lst = eo;
    eo->name = token.string;

    return(YES);

} /*Inter_CreateEO*/

/*
 *Inter_CreateRS()
 */
Inter_CreateRS()
{
    struct defrs *rs;

    if (CurrentEO() == NO)
	return(NO);

    GetToken(YES);

    if (token.opcode != WORD) {
	EH("A name of an rule set must be of type word.");
	return(NO);
	}

    if (RS_Exists(token.string) == YES) {
	sprintf(error_sb, "Rule set <%s> already exists.", token.string);
	EH(error_sb);
	return(NO);
	}

    rs = CreateRS();
    rs->name = token.string;
    rs->next = imgr.defrs_lst;
    imgr.defrs_lst = rs;

    GetToken(YES);

    /*optional priority field*/
    if (token.opcode == NUMERIC)
	rs->priority = token.numeric;
    else
	rs->priority = RS_PRIORITY_DEFAULT;

    return(YES);

} /*Inter_CreateRS*/
