/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: main.c - contain the main functions needed for he global execution of the system.
 *
 * Date: 6-16-87
 *
 */

#include <stdio.h>
#include "common.h"		/*Should always be the first header*/
#include "main.h"
#include "comm.h"

extern FILE *fopen();

main(argc, argv) 
int argc;
char *argv[];
{
    struct defeo *eos;

    Init(argc, argv);

    if (argc == 1) {
	InterActive();
	exit(0);
	}
	
    if ((imgr.fp = fopen(argv[1], "r")) == NULL) {
	EH("file doesn't exist");
	return(NO);
	}

    ch = getc(imgr.fp);

    Parser();

    Reset();

    Run();

    return(YES);

} /*main*/

/*
 *Reset
 */
Reset()
{
    int ct;
    int index = 1;

    if (CurrentEO() == NO)
 	return(NO);      

    imgr.current_eo->defrs_locals = imgr.defrs_lst;

    CleanNets();

    /*Set the new offset for the fact ids*/
    ct = imgr.eo_ids;

    do {
	ct = (int)(ct / 10);
	index *= 10;
	} while (ct > 1);

    /*This whole idea has to be refined for interactive mode*/
    /*Since EO's can be added and deleted between runs*/
    imgr.factid_disp = index; /*Create a buffer of factor of 10 for now*/

    SetUp_EOs(imgr.defeo_lst);

    CreateInBoxes(imgr.defeo_lst);

    return(YES);

} /*Reset*/

/*
 *Run
 */
Run()
{
    struct defeo *eos;

    if (CurrentEO() == NO)
 	return(NO);      

    imgr.current_eo->defrs_locals = imgr.defrs_lst;

    imgr.status = SYSTEM_CONT;

    while (imgr.status == SYSTEM_CONT) {

	imgr.status = SYSTEM_DONE;

	for (eos=imgr.defeo_lst; eos != NULL; eos = eos->next)
	    StartComm(eos);

	} /*execution loop*/

    PrintSummary();

    printf("Execution has completed.\n");

    return(YES);

} /*Run*/

/*
 *SetUp_EOs
 */
SetUp_EOs(eos)
struct defeo *eos;
{
    struct mailbox *setids;
    struct defeo *eo_lst;
    struct deffacts *stmt;
    char **facts;

    for (eo_lst=imgr.defeo_lst; eo_lst != NULL; eo_lst = eo_lst->next) {

	eo_lst->status = EO_INIT;
	eo_lst->total_rules = 0;

	/*Give ids to all of the outboxes eo names*/
	for (setids= eo_lst->out_boxes; setids != NULL; setids=setids->next)
	    setids->eo_id = GetEO_ID(setids->eo_name);

	/*For each expert object assert the new facts from the deffacts stmts*/
	for (stmt=eo_lst->deffacts; stmt != NULL; stmt = stmt->next)
	    for (facts=stmt->facts; *facts != NULL; ++facts)
		Assert(eo_lst, *facts, -1);

	} /*for each expert object*/

    return( YES );

} /*SetUp_EOs*/

/*
 * Do until no expert object has any input in its in_boxes.
 * Meaning, for one cycle of the execution, nothing new is assert or
 * retracted in the system.
 */

/*
 * Cycle of execution:
 * 1) Process all in messages for each object, when an object has received
 *    all of its inputs, then process the agenda.
 *
 * 2) Find the next expert object that needs processing.
 */

