 /*
  * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
  * 
  * FILE: opnet.c - holds the routines for construction of the net.
  *
  * Date: 6-16-87
  *
  */

#include <stdio.h>
#include "common.h"

/*
 *AddPattern - adds a pattern to he list of patterns for the net.
 */
struct pattern *AddPattern(new)
struct ruleslot *new;
{
    struct pattern *plist;
    struct pattern *past;
    struct ruleslot *rtemp; /*traverses the old pattern*/
    struct ruleslot *ntemp; /*temporary for traversing the new pattern*/
    struct match *pmatch; /*add the pattern to the match list*/

    past = NULL;

    plist = imgr.current_rs->patterns;

    while (plist != NULL) {

	rtemp = plist->pattern; /*take out the next pattern*/

	ntemp = new; /*ntemp goes through the new pattern each time*/

        /*compare the two patterns to see if there the same*/
	while ((rtemp != NULL) && (ntemp != NULL)) {
	    if (rtemp->opcode != ntemp->opcode)
		break;

	    if (rtemp->opcode == NUMERIC) {
		if (rtemp->numeric != ntemp->numeric)
		    break;
		} /*if*/
	    else if (rtemp->string != ntemp->string)
		break;

	    if (rtemp->string != ntemp->string)
		break;

	    rtemp = rtemp->next;
	    ntemp = ntemp->next;

	    } /*while still a list up there*/

	/*If the two patterns were equal then return a pointer to the old one*/
	if ((rtemp == NULL) && (ntemp == NULL)) {

	    pmatch = allocatestruct(match);
	    pmatch->slot = plist;
	    pmatch->next = imgr.pmatchList;
	    imgr.pmatchList = pmatch;
	    return(plist);
	    }

	past = plist;
	plist = plist->next; /*next plist to check out*/
	}  /*while still more patterns*/

    /*If there were no patterns that matched the new one then create one*/
    if (imgr.current_rs->patterns == NULL) 
	past = imgr.current_rs->patterns = allocatestruct(pattern);
    else {
	past->next = allocatestruct(pattern);
	past = past->next;
	}

    past->pattern = new;
    past->outflow = NULL;
    past->next = NULL;

    pmatch = allocatestruct(match);
    pmatch->slot = past;
    pmatch->next = imgr.pmatchList;
    imgr.pmatchList = pmatch;

    return(past); /*new pattern struct created*/

} /*AddPattern*/

/*
 *CreateNet - build the dataflow graph for the LHS of a rule.
 */
struct decidernode *CreateNet(lhs_list, input_side)
struct node *lhs_list;
int input_side;	
{
    int tally;		/*How many input flows have been added*/
    struct decidernode *pnode; /*the present node being worked on*/
    struct lr *plrtemp; /*created side of a node for single pattern rules*/
    struct pattern *ptemp; /*Temp for pattern return*/
    struct values *pvalues; /*The values created for the constant flow:1 input*/

    tally = 0;

    pnode = allocatestruct(decidernode);
    pnode->outflow = NULL;
    pnode->lhs = pnode->rhs = NULL;
    pnode->side =  input_side;

    while (lhs_list != NULL) {

	if (tally == 1)
	    input_side = RHS;

	if (tally > 1) {
	    pnode->outflow = allocatestruct(decidernode);
	    pnode = pnode->outflow;
    	    pnode->outflow = NULL;
	    pnode->lhs = pnode->rhs = NULL;
	    pnode->side = LHS;
	    } /*tally < 1*/

	if (lhs_list->opcode == PATTERN) {
	    ptemp = AddPattern(lhs_list->pattern);

	    AddDecidernode(pnode, ptemp, input_side, '+', lhs_list->string);

	    ReturnNode(lhs_list->right);

	    lhs_list->right = NULL;
	    }
	else
	    EH("System Error: CreateNet case out of bounds");

	lhs_list = lhs_list->bottom;
	tally++;
	} /*While lhs_list*/

    if (tally == 0)
	EH("Invalid LHS of rule.");

    if (tally == 1) {
	pvalues = allocatestruct(values);
	pvalues->next = NULL;
	pvalues->whoset = -1;
	pvalues->name = SBNULL;
	pvalues->opcode = SPECIAL;
	pvalues->string = SBAND;

	plrtemp = allocatestruct(lr);
	plrtemp->next = NULL;
	plrtemp->locals = pvalues;
	pnode->rhs = plrtemp;
	} /*if only one pattern on LHS of rule*/

    return(pnode);

} /*CreateNet*/

/*
 *AddDecidernode - add a node from a pattern to the net.
 */
AddDecidernode(pnode, pattern, input_side, boolean, binder)
struct decidernode *pnode; 
struct pattern *pattern;
int input_side;
char boolean;
char *binder;
{
    struct datalink *pdatalink;
    struct values *pvalues;

    if (pattern->outflow == NULL) {
	pattern->outflow = allocatestruct(datalink);
	pdatalink = pattern->outflow;
	}
    else {
	/*We already have flow to nodes, so add another one*/
	pdatalink  = pattern->outflow;

	while (pdatalink->next != NULL)
	    pdatalink = pdatalink->next;

	pdatalink->next = allocatestruct(datalink);
	pdatalink = pdatalink->next;

	} /*else LHS or RHS*/

    /*If the pattern is bound to a variable, then store that information*/
    if (binder != SBNULL) {
 	pvalues = allocatestruct(values);
	pvalues->opcode = SBIND;
	pvalues->string = pvalues->name = binder;
	pvalues->type = WORD;
	pvalues->next = NULL;
	pdatalink->binder = pvalues;
	}
    else
	pdatalink->binder = NULL;

    pdatalink->next = NULL;
    pdatalink->path = pnode;
    pdatalink->side = input_side; /*1 -> LHS; 0 -> RHS; */
    pdatalink->boolean = boolean;

} /*AddDecidernode*/
