 /*
  * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
  * 
  * FILE: parser.c - contains the PPS parser
  *
  * Date: 6-16-87
  *
  */

#include <stdio.h>
#include "common.h"
#include "comm.h"

/*
 *Parser - Parses the input during batch mode.
 */
Parser()
{

    GetToken(YES);

    while (token.opcode != QUIT) {

	if (token.opcode != LPAREN) {
	    EH("Expected a '('");
	    return(NO);
	    }

	GetToken(YES);

	if (token.opcode != WORD)  {
	    EH("Expected an identifier after '('");
	    return(NO);
	    }

	if (token.string == DEFEO) {
	    GetToken(YES);
	    if (ParseEO() == NO) 
		return(NO);
	    } /*DEFEO*/
	else {
	    sprintf(error_sb, "%s is not a valid construct.", token.string);
	    EH(error_sb);
	    return(NO);
	    }

	} /*while not EOF*/

    return(YES);

} /*Parser*/

/*
 *ParseMessages
 */
struct eo_id *ParseMessages()
{
    struct eo_id *new, *trav;

    /*Defaults to itself -> present expert object*/
    if (token.opcode != SENDER) {
	new = allocatestruct(eo_id);
	new->ni.name = imgr.current_eo->name;
	new->status = WORD;
	AddEO_to_Ruleset(new->ni.name);
	AddEO_to_EO_Boxes(imgr.current_eo->name, new->ni.name, OUT_BOX);
	return(new);
	} /*SENDER*/

    new = NULL;

    GetToken(YES);

    while (token.opcode == WORD) {

	trav = allocatestruct(eo_id);
	trav->next = new;

	new = trav;
	new->ni.name = token.string;
	new->status = WORD;

	AddEO_to_Ruleset(token.string);
	AddEO_to_EO_Boxes(imgr.current_eo->name, new->ni.name, OUT_BOX);

	GetToken(YES);

	if (token.opcode != COMMA)
	    break;

	GetToken(YES);

	if (token.opcode != WORD) {
	    EH("malformed statement, expected a name after ','");
	    return(NO);
	    }

	} /*while*/

    return(new);

} /*ParseMessages*/

/*
 *ParseEO
 */
ParseEO()
{

    if (token.opcode != WORD) {
	EH("Expected the name of the expert object.");
	return(NO);
	} /*if*/

    if (imgr.current_eo != NULL)
	imgr.current_eo->defrs_locals = imgr.defrs_lst;

    imgr.current_eo = CreateEO();
    imgr.current_eo->name = token.string;

    imgr.defrs_lst = NULL;
    imgr.current_rs = NULL;

    imgr.current_eo->next = imgr.defeo_lst;
    imgr.defeo_lst = imgr.current_eo;

    GetToken(YES);

    if (token.opcode == WORD) 
	GetToken(YES);

    if (token.opcode != LPAREN) {
	EH("Malformed expert object definition.");
	return(NO);
	}


    while (token.opcode == LPAREN) {

	GetToken(YES);

	if (token.opcode != WORD) {
	    EH("Malformed expert object definition.");
	    goto PARSE_EO_ERROR;
	    }

	if (token.string == DEFRS) {
	    if (ParseRS() == NO)
		goto PARSE_EO_ERROR;
	    }
	else if (token.string == DEFFACTS) {
	    if (ParseFacts() == NO) 
		goto PARSE_EO_ERROR;
	    }
	else {
	    EH("Unsupported declaration.");
	    goto PARSE_EO_ERROR;
	    }
	} /*while*/

    if (token.opcode != RPAREN) {
	EH("Missing closing ')' in expert object definition.");
	goto PARSE_EO_ERROR;
	}

    GetToken(YES);

    imgr.current_eo->defrs_locals = imgr.defrs_lst;

    return(YES);

PARSE_EO_ERROR :

    /*Should free rulesets and the expert object*/
    return(NO);

} /*ParseEO*/

/*
 *ParseRS
 */
ParseRS()
{
    struct defrs *rs;

    /*Add another rule set to the main list of rule sets*/
    rs = imgr.current_rs = CreateRS();

    rs->next = imgr.defrs_lst;
    imgr.defrs_lst = rs;

    rs->priority = RS_PRIORITY_DEFAULT;
    imgr.current_rs->out_boxes = NULL;
    imgr.current_rs->patterns = NULL;

    /*Parse the new rule set*/
    GetToken(YES); /*Looking for rule set name*/

    if (token.opcode != WORD) {
	EH("Missing rule set name");
	goto ERROR_DEFRS;
	}

    if (RS_Exists(token.string) == YES) {
	sprintf(error_sb, "Redeclaration of rule set %s", token.string);
	EH(error_sb);
	goto ERROR_DEFRS;
	}

    rs->name = token.string;

    printf("Compiling rule set %s.\n", rs->name);

    GetToken(SAVE);

    if (token.opcode == WORD) /*Is the comment string there? */
	GetToken(YES);

    GetToken(YES);

    while (token.opcode != RPAREN) {

	if (token.opcode != LPAREN) {
	    EH("Missing '(' in defrule statement.");
	    goto ERROR_DEFRS;
	    }

	GetToken(YES);

	if (token.opcode != WORD)  {
	    RS_FORMAT:
	    EH("Illegal format.");
	    goto ERROR_DEFRS;
	    }

	if (token.string == DECLARE) {

	    GetToken(YES);

	    if (token.opcode != LPAREN)
		goto RS_FORMAT;

	    GetToken(YES);

	    if (token.opcode != WORD) 
		goto RS_FORMAT;

	    if (token.string == PRIORITY) {
		GetToken(YES);

		if (token.opcode != NUMERIC)
		    goto RS_FORMAT;

		imgr.current_rs->priority = token.numeric;

		GetToken(YES);
		}
	    else {
		EH("Unknown type of declare statement in defrs");
		goto ERROR_DEFRS;
		}

	    if (token.opcode != RPAREN)
		GetToken(YES);
	    } /*declare*/
	else if (token.string == DEFRULE) 
	    ParseRule();
	else {
	    EH("Unknown statement in defrs");
	    goto ERROR_DEFRS;
	    }
	} /*while*/

    GetToken(YES);

    return(YES);

ERROR_DEFRS:

    imgr.defrs_lst = imgr.defrs_lst->next;
    free(imgr.current_rs);
    return(NO);

    /*
     *should free the other elements of the bad ruleset ....***********
     */
} /*ParseRS*/

/*
 *ParseRule
 */
ParseRule()
{
    extern struct decidernode *CreateNet();
    struct decidernode *state;
    struct node *pnode = NULL;
    struct decidernode *in;
    extern int NodeId;

    ReturnMatch(imgr.pmatchList);
    imgr.pmatchList = NULL;

    GetToken(YES); /*Looking for rule name*/

    if (token.opcode != WORD) {
	EH("Missing Rulename");
	return(NO);
	}

    imgr.stmt_name = token.string;

    /*if NO then the rulename already existed*/
    if (AddRule(token.string, NULL, YES) == NO)
	return(NO);

    printf("Compiling %s .\n", imgr.stmt_name);

    GetToken(SAVE);

    if (token.opcode == WORD) { /*Read the comment string if its there*/
	GetToken(YES);
	}

  DoForever { /*Parse Declarations*/

    if (token.opcode != LPAREN)
	break;

    GetToken(SAVE);

    if (token.opcode != WORD)
	break;
    if (token.string != DECLARE)
	break;

    GetToken(SAVE);

    if (token.opcode != LPAREN) 
	break;

    GetToken(YES); /*Burn off the lookahead*/
    GetToken(YES);
    GetToken(YES);
    GetToken(YES); /*Get Next Token*/

    if (token.opcode != WORD) {
	EH("Missing declarations command after '('");
	return(NO);
	}

    if (token.string == PRIORITY) {

	GetToken(YES);

	if (token.opcode != NUMERIC) {
	    EH("A Rules Priority must be a numeric");
	    return(NO);
	    }
		
	imgr.current_dr->priority = token.numeric;

	} /*Priority*/
    else {
	EH("<declare command> unknown!");
	return(NO);
	}

    GetToken(YES);

    if (token.opcode != RPAREN) {
	EH("Missing ')' ");
	return(NO);
	}

    GetToken(YES);

    if (token.opcode != RPAREN) {
	EH("Missing ')' ");
	return(NO);
	}

    GetToken(SAVE);
    } /*DoForever*/

    GetToken(YES); /*Start parsing right on the '(' */

    if ((pnode = ParseLHS(YES)) == NULL)
	return(NO);

    state = allocatestruct(decidernode);

    state->rhs = state->lhs = NULL;
    state->outflow = NULL;
    state->side = END;

    if ((in = CreateNet(pnode, LHS)) == NULL)
	return(NO);

    in->outflow = state;
    in->side = LHS;
    /*Now Parsing the RHS of the rule*/

    if (token.opcode != SEPARATOR) {
	EH("Missing '=>' in format.");
	return(NO);
	}

    GetToken(YES);

    /*Add the RHS to the final node of the net.*/
    state->lhs = NULL;
    state->rhs = allocatestruct(lr);
    state->rhs->locals = allocatestruct(values);
    state->rhs->locals->name = imgr.stmt_name;
    state->rhs->locals->opcode = WORD;
    state->rhs->locals->numeric = imgr.current_dr->priority;
    state->rhs->locals->next = NULL;

    if ((state->rhs->next = ParseRHS()) == NULL)
	return(NO);

    if (token.opcode != RPAREN) {
	EH("Every rule must have an enclosing set of parenthesis.");
	return(NO);
	}

    GetToken(YES);

    ReturnNode(pnode); /*Give back all that space*/
    pnode = NULL;

    ReturnMatch(imgr.pmatchList);
    imgr.pmatchList = NULL;

    printf("Parse Rule successfully executed. +++++++++++\n");
    if (imgr.error == NO)
	return(YES); /*Successful execution*/
    else
	return(NO);

} /*ParseRule*/

/*
 *ParseFacts - Parse a Deffacts statement.
 */
ParseFacts()
{
    char **facts;
    int ct = 0;
    static char fact[ MAXFACTLEN ];
    char **temp;

    GetToken(YES);

    if (token.opcode != WORD) {
	EH("Missing Deffacts name.");
	return(NO);
	}

    imgr.stmt_name = token.string;

    /*If it returns NO then the name already existed*/
    if (AddFacts(1, token.string, SBNULL, (char **)NULL) == NO)
	return(NO);

    printf("Reading in Deffacts %s\n", token.string);

    GetToken(YES);

    if (token.opcode == WORD) {
	printf("Processing facts: %s.\n", token.string);
	GetToken(YES);
	} /*comment parsing*/

    /*Create a list of strings that are the facts from the deffacts stmt*/
    facts = INIT_BUF;

    for (*facts = NULL; token.opcode != RPAREN; ct = 0, GetToken(YES)) {

	if (token.opcode != LPAREN) {
	    EH("Missing '(' in deffacts stmt.");
	    goto CleanUp_Deffacts;
	    }

	GetToken(YES);

	for (;token.opcode != RPAREN; GetToken(YES)) {

	    if (token.opcode == WORD) 
		sprintf(fact + ct, "%s ", token.string);
	    else
		sprintf(fact + ct, "%6.2f ", token.numeric);

	    while (fact[ct] != '\0')
		ct++;

	    } /*for fact*/

	facts = AddMessage(facts, MakeString(fact));

	} /*while for facts*/

    GetToken(YES); /*burn off last ')'*/

    if (*facts == NULL) {
	EH("Invalid deffacts statement");
	free(facts);
	return(NO);
	}

    AddFacts(2, imgr.stmt_name, SBNULL,  facts);

    return(YES);

CleanUp_Deffacts:

    for (temp = facts; *temp != NULL; temp++)
	free(*temp);

    free(facts);

    return(NO);

} /*ParseFacts*/
		
/*
 *ParseLHS - Parse the left hand side of a rule.
 */
struct node *ParseLHS()
{
    struct node *head, *rtnode;

    /*Create the head of the LHS list of patterns*/
    rtnode = head = allocatestruct(node);
    head->bottom = NULL;

    /*Loop for all patterns in the left hand side*/
    DoForever {

	/*Parse all ?<var> <- statements  (binder)*/
	if (token.opcode == SBIND) {

	    rtnode->opcode = PATTERN;
	    rtnode->string = token.string;

	    GetToken(YES);

	    if (token.opcode != BINDER) {
		EH("Malformed statement: expected '<-'");
	 	return(NULL);
		} /*if*/
	
	    GetToken(YES);

	    if (token.opcode != LPAREN) {
	      sprintf(error_sb, "Expected ?%s<-( <pattern> ) ", rtnode->string);
		EH(error_sb);
		return(NULL);
		} /*if*/

	    /*Parse the pattern now*/
	    if ((rtnode->pattern = ParsePattern()) == NULL)
		return(NULL);

	    } /*if SBIND*/
	else {

	    /*We have a pattern, so parse it*/
	    if (token.opcode != LPAREN)
		break;

	    /*Set the correct type of code*/
	    rtnode->opcode = PATTERN;
	    rtnode->string = SBNULL;
  	    rtnode->right = NULL;

	    if ((rtnode->pattern = ParsePattern()) == NULL)
		return(NULL);

	    } /*else*/

	if ((token.opcode == SBIND) || (token.opcode == LPAREN)) {
	    rtnode->bottom = allocatestruct(node);
	    rtnode = rtnode->bottom;
	    rtnode->bottom = NULL;
	    } /*if*/	

	} /*DoForever*/

    return(head);

} /*ParseLHS*/

/*
 *ParseRHS - Parse the Right Hand Side of a rule.
 */
struct lr *ParseRHS()
{
    struct lr *plrHead, *plrtemp; /*temp to traverse lr struct. */
    struct values *nextslot; /*temp to traverse values struct*/

    nextslot = NULL;
    plrHead = plrtemp = NULL;

    DoForever {

	if (plrtemp == NULL) {
	    plrtemp = plrHead = allocatestruct(lr);
	    }
	else {
	    plrtemp->next = allocatestruct(lr);
	    plrtemp = plrtemp->next;
	    } /*else*/

	plrtemp->locals = allocatestruct(values);
	nextslot = plrtemp->locals;
	
	if (token.opcode != LPAREN)
	    break;

	GetToken(YES);

	nextslot->opcode = token.opcode;
	nextslot->string = token.string;

	if (token.opcode != WORD) {
	    EH("Unknown RHS format.");
	    goto PARSE_ERROR;
	    } /*if*/

	if (token.string == STOP) {
	    nextslot->whoset = RHSSTOP;

	    GetToken(YES);
	    }
	else if (token.string == ASSERT) {

	    nextslot->whoset = RHSASSERT;

	    GetToken(YES);

	    if (token.opcode != LPAREN) {
		EH("Expected '(' in assert statement.");
		goto PARSE_ERROR;
		}

	    GetToken(YES);

	    if (token.opcode == RPAREN) {
		EH("Asserting nothing is an illegal statement.");
		goto PARSE_ERROR;
		}
		
	    /*Parse the entire assert statement*/
	    while (token.opcode != RPAREN) {
		nextslot->next = allocatestruct(values);
		nextslot = nextslot->next;
		nextslot->opcode = token.opcode;

		switch (token.opcode) {

			case SBIND :
			case WORD :
				nextslot->string = token.string;
				break;

			case NUMERIC :
				nextslot->numeric = token.numeric;
				break;

			default :
				EH("Unsupported feature found in assert.");
				return(NULL);
				break;
			} /*assert case*/

		GetToken(YES);
		
		} /*while not RPAREN*/

	    GetToken(YES);

	    nextslot->next = allocatestruct(values);
	    nextslot = nextslot->next;
	    nextslot->opcode = EO_IDS;
	    nextslot->oe.eo_id = ParseMessages();

	    } /*ASSERT*/
	else if (token.string == RETRACT) {

	    nextslot->whoset = RHSRETRACT;

	    GetToken(YES);

	    if (token.opcode != LPAREN) {
		EH("Missing '(' before list of variables.");
		goto PARSE_ERROR;
		}

	    GetToken(YES);

	    if (token.opcode != SBIND) {
		EH("Need at least one rule binded variable");
		goto PARSE_ERROR;
		}

	    do {
		nextslot->next = allocatestruct(values);
		nextslot = nextslot->next;
		nextslot->opcode = token.opcode;
		nextslot->string = token.string;

		GetToken(YES);

		} while (token.opcode == SBIND);

	    GetToken(YES);

	    plrtemp->locals->next->oe.eo_id = ParseMessages();

	   } /*RETRACT*/
	else if (token.string == PRINTOUT) {

	    nextslot->whoset = RHSPRINTOUT;

	    GetToken(YES);

	    if (token.opcode != LPAREN) {
		EH("Expected '(' in printout statement.");
		return(NULL);
		}

	    GetToken(YES);

	    if (token.opcode == RPAREN) {
		EH("A printout without any statement is illegal.");
		goto PARSE_ERROR;
		}
		
	    /*Parse the entire assert statement*/
	    while (token.opcode != RPAREN) {
		nextslot->next = allocatestruct(values);
		nextslot = nextslot->next;
		nextslot->opcode = token.opcode;

		switch (token.opcode) {
			case SBIND :
			case WORD :
			
				nextslot->string = token.string;
				break;

			case NUMERIC :

				nextslot->numeric = token.numeric;
				break;

			default :
			
				EH("Unsupported feature found in printout.");
				goto PARSE_ERROR;
				break;
			} /*printout case*/

	  	    GetToken(YES);
		
		    } /*while not RPAREN*/

	if (token.opcode != RPAREN) {
	    EH("Expected ')' after retract variables.");
	    goto PARSE_ERROR;
	    }

	GetToken(YES);

	    } /*printout*/
	else {
	    EH("Unsupported RHS function.");
	    goto PARSE_ERROR;
	    }


	if (token.opcode != RPAREN) {
	    EH("Expected ')' ");
	    goto PARSE_ERROR;
	    } /*if*/

	GetToken(YES);

	nextslot->next = NULL;

	if (token.opcode != LPAREN)
	    break;
	} /*Do for each RHS statement*/

    if (plrHead == NULL) {
	EH("You cannot have a nonempty LHS of a rule.");
	goto PARSE_ERROR;
	}

    plrtemp->next = NULL;
    return(plrHead);

PARSE_ERROR :
    EH("Illegal RHS of rule");

    /*Give back any allocated space*/
    if (plrHead != NULL)
	plrtemp = NULL;

    if (nextslot != NULL)
	nextslot->next = NULL;

    ReturnLrs(plrHead);
    return(NULL);

} /*ParseRHS*/

/*
 *ParsePattern - Parse a Left Hand Side pattern slot.
 */
struct ruleslot *ParsePattern()
{
    struct ruleslot *head, *trav;
    int op = 0;

    /*head will point to the parsed pattern*/
    head = trav = NULL;

    GetToken(YES);	/*Get the first element of the pattern*/

    /*Parse everything up to the next ')' */
    while (token.opcode != RPAREN) {

	if (head == NULL)
	    trav = head = allocatestruct(ruleslot);
	else {
	    trav->next = allocatestruct(ruleslot);
	    trav = trav->next;
	    }

	/*Case each possible token type possible after connectors*/
	switch (token.opcode) {
	    case SBIND :
	    case WORD :
		trav->opcode = token.opcode;
		trav->string = token.string;
		break;

	    case NUMERIC :
		trav->opcode = token.opcode;
		trav->numeric = token.numeric;
		break;

	    default: 
		EH("Unknown pattern format found");
		trav->next = NULL;
		ReturnRuleSlot(head);
		return(NULL);
		break;

		} /*switch*/

	GetToken(YES);		/*Parse a token.*/


	} /*while != RPAREN*/

    trav->next = NULL;

    GetToken(YES);	/*Parse off the RPAREN*/

    return(head); /*Return the parsed slot*/

} /*ParsePattern*/
