/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: print.c - contains all of the functions responsible for the printing of data
 *       to the user's screen.
 *
 * Date: 6-16-87
 *
 */

#include <stdio.h>
#include "common.h"
#include "comm.h"


/*
 *PrintAgenda_RS
 */
PrintAgenda_RS(rs)
struct defrs *rs;
{
    struct activation *act;
    int ct;

    ct = 0;

    printf("---------------Agenda RS:<%s>--------------\n", rs->name);

    for (act = rs->Agenda.list; act != NULL; act = act->next) {
	printf("Rule Name: %s; ", act->rhs->rhs->locals->name);
	printf("Activation id: %d; ", act->id);
	printf("Priority %6.2f.",  act->priority);
	PrintIds(act->variables);
	printf("\n");
	} 

    printf("------------------------------------\n");

} /*PrintAgenda_RS*/ 

/*
 *PrintIds
 */
PrintIds(locals)
struct values *locals;
{
    char SP = ' ';

    printf(" [");

    for (;locals != NULL; locals=locals->next)  {
	if (locals->opcode != SPECIAL)
	    break;

	if(locals->whoset >= 0)
	    printf("%cf-%d", SP, locals->whoset);

	SP = ',';
	} /*for*/

    printf(" ]");

} /*PrintIds*/

/*
 *DisplayFacts -- Prints out the fact list.
 */
DisplayFacts(eo)
struct defeo *eo;
{
    struct fact *pfacttrav;
    struct element *peletrav;

    printf("=======================================================\n");
    printf("Fact list presently contains:\n");
    printf("=======================================================\n");

    for(pfacttrav = eo->factlist;pfacttrav != NULL; pfacttrav=pfacttrav->next) {

	printf("Id: %d; ", pfacttrav->id);

	for (peletrav=pfacttrav->entry;peletrav != NULL;peletrav=peletrav->next)
	    switch(peletrav->type) {
		case WORD :
			printf("%s ", peletrav->string);
			break;

		case NUMERIC :
			printf("%f ", peletrav->numeric);
			printf(" ");
			break;

		default :
			EH("Display facts: Unknown fact type");
			break;
		} /*switch*/

	printf("\n");
	} /*for*/

    printf("=======================================================\n");

    return(YES);

} /*DisplayFacts*/

/*
 *PrintNet - prints out the net to be displayed.
 */
PrintNet(rs)
struct defrs *rs;
{
    struct pattern *trav;

    if (!(DEBUG))
	return;

    for (trav=rs->patterns; trav != NULL; trav=trav->next) {
	printf("Starting to Print Pattern:\n");
	PrintPattern(trav->pattern);
	PrintFlow(trav->outflow);
	}

    return(YES);

} /*PrintNet*/

/*
 *PrintPatterns
 */
PrintPattern(pattern)
struct ruleslot *pattern;
{
    if (!(DEBUG))
	return;

    printf("Print Pattern: \n");
    for (;pattern != NULL;pattern=pattern->next) {

	printf("opcode = %o\n", pattern->opcode);

	switch (pattern->opcode) {

		case SBIND :
		case WORD :
	    		printf("->string = %s\n", pattern->string);
			break;

		case NUMERIC :
	    		printf("->numeric = %f\n", pattern->numeric);
			break;
		} /*switch*/
	} /*for*/
    printf("============Print Pattern ==========done\n");
} /*PrintPattern*/

/*
 *PrintFlow - Prints out the flow path from a pattern to the RHS of the rule.
 */
PrintFlow(flow)
struct datalink *flow;
{  
    if (!(DEBUG))
	return;

    printf("PrintFlow\n");
    for (;flow != NULL; flow=flow->next) {
	printf("\nOne path is: \n");
	printf("Entering from the %s\n", (flow->side == RHS) ? "RHS" : "LHS");
	PrintDeciderNodes(flow->path, 1);
	}
} /*PrintFlow*/

/*
 *PrintDeciderNodes - Prints out the chain nodes.
 */
PrintDeciderNodes(flow, level)
struct decidernode *flow;
int level;
{

    if (!(DEBUG))
	return;

    if (flow == NULL)
	return;

    printf("PrintDeciderNodes: level %d.\n", level);

    if (flow->side == END)
	printf("This is the contents of the RHS of the rule.\n");
    else
	printf("Enters next from the %s.\n",(flow->side==RHS) ? "RHS" : "LHS");

    printf("\nValues in RHS are: \n");
    PrintLr(flow->rhs);

    printf("\nValues in LHS are: \n");
    PrintLr(flow->lhs);
    printf("===============================level %d\n", level);

    PrintDeciderNodes(flow->outflow, level + 1);
} /*PrintDeciderNode*/

/*
 *PrintLr - Print out the side of a decidernode (input buffer).
 */
PrintLr(lists)
struct lr *lists;
{
    if (!(DEBUG))
	return;

    printf("PrintLr\n");

    for(;lists != NULL;lists=lists->next) {
	printf("Values Entry =============\n");
	PrintValues(lists->locals);
	}
    
    return(0);
}/*PrintLr*/

/*
 *PrintValues - Print out a list of values structures.
 */
PrintValues(pvalues)
struct values *pvalues;
{
    if (!(DEBUG))
	return;

    for(;pvalues != NULL;pvalues=pvalues->next) {
	printf("opcode %o; type = %d; ", pvalues->opcode, pvalues->type);
	printf("whoset = %d;", pvalues->whoset);
	switch (pvalues->opcode) {
		case SBIND :
		case WORD :
			printf("string = %s", pvalues->string);
			break;

		case NUMERIC :
			printf("numeric = %s", pvalues->numeric);
			break;
		} /*switch*/
	printf("\n");
	} /*for*/
} /*PrintValues*/
	
/*
 *PrintNodes - Print a list of node structures.
 */
PrintNodes(pnode)
struct node *pnode;
{
    if (!(DEBUG))
	return;

    if (pnode == NULL) {
	printf("NULL\n");
	return;
	}

     printf("PrintNodes\n");

     printf("opcode = %o\n", pnode->opcode);
     switch (pnode->opcode) {
	case WORD :
		printf("->string = %s\n", pnode->string);
		break;

	case NUMERIC :
		printf("->numeric = %f\n", pnode->numeric);
		break;

	case PATTERN :
		printf("PrintRuleSlot\n");
		PrintRuleSlot(pnode->pattern);
		break;

	} /*switch*/

    printf("right -> \n");
    PrintNodes(pnode->right);
    printf("bottom ->\n");
    PrintNodes(pnode->bottom);
    printf("==================PrintNodes================done\n");
} /*PrintNodes*/

/*
 *PrintRuleSlot - Print out the rule slot.
 */ 
PrintRuleSlot(slot)
struct ruleslot *slot;
{
    if (!(DEBUG))
	return;

    if (slot == NULL)
	return;

    printf("rs:opcode %o\n", slot->opcode);
    switch (slot->opcode) {
	case WORD :
		printf("->string = %s\n", slot->string);
		break;

	case NUMERIC :
		printf("->numeric = %f\n", slot->numeric);
		break;

	} /*switch*/
   
    PrintRuleSlot(slot->next);

} /*PrintRuleSlot*/

/*
 *PrettyPrint - print out the appropriate data object.
 */
PrettyPrint()
{
    struct defrule *rules;
    struct deffacts *facts;
    struct defrs *rs;
    struct defeo *eos;

    GetToken(YES);

    /*Fetch the name*/
    if (token.opcode != WORD) {
	EH("Can't print: bad argument");
	return(NO);
	} /*if*/

    if (token.string == DEFEO) {

	printf("--- expert objects ---\n");

	for (eos=imgr.defeo_lst; eos != NULL; eos = eos->next)
	    printf("%s\n", eos->name);

	return(YES);
	} /*DEFEO*/
    else if (imgr.current_eo == NULL) {
	EH("No current expert object.");
	return(NO);
	} /*No current eo*/
    else if (token.string == DEFRS) {

	printf("--- rule sets ---\n");

	for (rs=imgr.defrs_lst; rs != NULL; rs = rs->next)
	    printf("%s\n", rs->name);

	} /*DEFRS*/
    else if (token.string == DEFFACTS) {
	
	printf("--- deffacts ---\n");

	for (facts= imgr.current_eo->deffacts;facts != NULL;facts = facts->next)
	     printf("%s\n", facts->name);

	} /*DEFFACTS*/
    else if (token.string == DEFRULE) {

	printf("--- defrules ---\n");

	for (rules= imgr.current_rs->defrules;rules != NULL;rules = rules->next)
	    printf("%s\n", rules->name);

	} /*DEFRULE*/
    else if (strcmp(token.string, "in_boxes") == 0) {
	printf("In boxes are: \n");
	PrintBoxes(imgr.current_eo->in_boxes);

	} /*in_boxes*/
    else if (strcmp(token.string, "out_boxes") == 0) {
	printf("Out boxes are: \n");
	PrintBoxes(imgr.current_eo->in_boxes);

	} /*out_boxes*/
    else if (strcmp(token.string, "summary") == 0) 
	PrintSummary();
    else {

    /*First see if it is a rule name, if so print it and exit*/
    for (rules = imgr.current_rs->defrules; rules != NULL; rules=rules->next) 
	if (strcmp(rules->name, token.string) == 0) {
	    printf("Rule name <%s> \n", token.string);
	    printf("%s\n", rules->stmt);
	    return(YES);
	    } /*if*/

    /*If not a rule name, then see if its a deffacts name, print, exit*/
    for (facts=imgr.current_eo->deffacts; facts != NULL; facts = facts->next)
	if (strcmp(facts->name, token.string)== 0){
	    printf("%s\n", facts->stmt);
	    return(YES);
	    } /*if*/
	   
    } /*not one of the above*/

    return(YES);

} /*PrettyPrint*/

/*
 *PrintMesgs
 */
PrintMesgs(mesgs)
char **mesgs;
{
    if (*mesgs == NULL)
	printf("No messages.\n");
    else for (;*mesgs != NULL;++mesgs)
	printf("mesg: <%s> \n", *mesgs);

    return(YES);

} /*PrintMesgs*/

/*
 *PrintBoxes
 */
PrintBoxes(boxes)
struct mailbox *boxes;
{

    for (;boxes != NULL; boxes = boxes->next) {

	printf("eo <%s>: id <%d>: status: ", boxes->eo_name, boxes->eo_id);

	switch (boxes->status) {
		case MESG_STOPPED :
				printf("stopped.\n");
				break;

		case MESG_INIT :
				printf("initialization.\n");
				break;

		case MESG_END :
				printf("received.\n");
				break;

		case MESG_START :
				printf("in progress.\n");
				break;

		default :
				printf("unknown: se.\n");
				break;
		} /*switch*/

	PrintMesgs(boxes->mesgs);
	printf("----------------------------done.\n");
	} /*for*/

    return(YES);

} /*PrintBoxes*/

/*
 *PrintSummary
 */
PrintSummary()
{
    struct defeo *eos;
    long ct;

    printf("--------------------Execution Summary--------------------\n");

    for (eos=imgr.defeo_lst; eos != NULL; eos = eos->next) {
	printf("Expert Object <%s>: ", eos->name);

	printf(" status <");
	switch (eos->status) {
		case EO_READY :
			printf("ready");
			break;

		case EO_STOPPED :
			printf("stopped");
			break;

		case EO_INIT :
			printf("initialized");
			break;

		case EO_PROCESSING :
			printf("running");
			break;

		case EO_WAITING :
			printf("waiting");
			break;

		default :
			printf("unknown");
			break;
		}

	ct = eos->total_rules;
	printf("> : %d rule%c fired.\n", ct, (ct == 1) ? ' ': 's');

	} /*for*/

    printf("---------------------------------------------------------\n");

    return(YES);

} /*PrintSummary*/
