/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: runtime.c - contains some of the more generic runtime functions.
 *
 * Date: 6-16-87
 *
 */

#include <stdio.h>
#include "common.h"
#include "comm.h"

/*
 *CleanAgenda
 */
CleanAgenda(rs, id, cond)
struct defrs *rs;
int id; /*Id to remove from the list*/
int cond;
{
    struct activation *search, *last, *temp;

    for (search=rs->Agenda.list; search != NULL;) {

	if (AnyId(id, search->variables) == NO) {
	    last = search;
	    search = search->next;
	    continue;
	    } /*AnyId*/

	if (search == rs->Agenda.list)
	    rs->Agenda.list = search->next;
	else
	    last->next = search->next;

	temp = search;
	search = search->next;
	
	ReturnValues(temp->variables);
	free(temp);

	} /*for*/

} /*CleanAgenda*/

/*
 *HASHTABLE: Our hash table for all of the strings in the system,
 *           for only one of any string can exist in the system,
 *           so comparison between strings can be done by simply 
 *           seeing if they have the same memory address.
 */

/*
 *MakeString - create a new string.
 */
char *MakeString(str)
char *str;
{
    char *new;

    new = (char *)malloc(strlen(str) + 1);

    if (new == NULL)
	EH("OUT OF DYNAMIC SPACE!!!!!!");

    strcpy(new, str);

    return(new);

} /*MakeString*/

/*
 *HashString - returns the systems address for the string given as
 *             its parameter.
 */
char *HashString(string)
char *string;
{
    int entry;
    struct draw *pdrawPeek;
    char *calc;
    int tally;

    entry = 0;
    tally = 3;

    calc = string;

    while ((*calc != EOS) && (tally-- != 0))
	entry += (int)*(calc++);

    entry = entry % HASHSIZE;

    for (pdrawPeek=ht[entry]; pdrawPeek != NULL; pdrawPeek=pdrawPeek->next) 
	if (strcmp(string, pdrawPeek->entry) == 0)
	    break;

    if (pdrawPeek == NULL) { /*Isn't in the hash table*/
	pdrawPeek = allocatestruct(draw);
	pdrawPeek->entry = MakeString(string);
	pdrawPeek->next = ht[entry];
	ht[entry] = pdrawPeek;
	} /*if*/

    return(pdrawPeek->entry); /*Return pointer to the new string*/
} 

/*
 *Equal - are two facts equal.
 */
Equal(peleold, pelenew)
struct element *peleold, *pelenew;
{
    while ((peleold != NULL) && (pelenew != NULL)) {
	if (peleold->type != pelenew->type)
	    return(NO);

	switch (peleold->type) {
	    case WORD :
		if (peleold->string != pelenew->string)
		    return(NO);
		break;

	    case NUMERIC :
		if (peleold->numeric != pelenew->numeric)
		    return(NO);
		break;

	    default :
		EH("Equal: unknown element type!");
		break;
	    } /*switch*/
	
	peleold = peleold->next;
	pelenew = pelenew->next;
	} /*while still equal length*/

/*Think about this: can this if be done faster*/

    if ((pelenew == peleold) && (pelenew == NULL))
	return( YES );

    return( NO );
} /*Equal*/

/*
 *EH - prints error messages to the user screen.
 */
EH(msg)
char *msg;
{
    imgr.error = YES;
    token.opcode = EOF;

    if(imgr.fn == NULL) {
	printf("ERROR: %s\n", msg);
	} /*interactive mode*/
    else { 
	if (imgr.stmt_name != NULL)
	    fprintf(stderr, "In rule %s ::: ", imgr.stmt_name);

	fprintf(stderr,"Fatal: ln %d in %s: %s\n",imgr.ln, imgr.fn, msg);
	} /*else*/

} /*EH*/

/*
 *FlushMessages
 */
char **FlushMessages(mesgs)
char **mesgs;
{
    while (*mesgs != NULL) 
	mesgs = RemoveMessage(mesgs);

    return(mesgs);

} /*FlushMessages*/

/*
 *CleanNets
 */
CleanNets()
{
    struct defeo *eos;
    struct defrs *rs;
    struct mailbox *boxes, *tmpbox;
    struct fact *facts, *tmpfact;

    imgr.total_rules = 0;

    for (eos=imgr.defeo_lst; eos != NULL; eos = eos->next) {

	/*first the out boxes*/
	for (boxes=eos->out_boxes; boxes != NULL;boxes = boxes->next)
	    boxes->mesgs = FlushMessages(boxes->mesgs);

	/*secondly, we flush the in boxes*/
	/*We recreate all the in_boxes so this means we give back any
	  old ones first*/
	for (boxes=eos->in_boxes; boxes != NULL;) {

	    boxes->mesgs = FlushMessages(boxes->mesgs);
	    free(boxes->mesgs);

	    tmpbox = boxes;
	    boxes = boxes->next;

	    free(tmpbox);
	    } /*for boxes != NULL*/

	eos->in_boxes = NULL;

	/*now we flush the fact list*/
	for (facts=eos->factlist; facts != NULL;) {
	    tmpfact = facts;
	    facts = facts->next;
	    ReturnElements(tmpfact->entry);
	    ReturnMatch(tmpfact->list);
	    } /*for all the facts*/

	InitEO_Factlist(eos);

	/*now we reset each rule sets opnet*/
	for (rs=eos->defrs_locals; rs != NULL; rs = rs->next) 
	    ResetNet(rs->patterns);

	} /*for each eo*/

    CreateInBoxes(imgr.defeo_lst);

} /*CleanNets*/

/*
 *ResetNet
 */
ResetNet(net)
struct pattern *net;
{
    struct lr *ReturnLrs();
    struct decidernode *path;
    struct datalink *paths;
    int side;

    for (;net != NULL; net=net->next) 
	for (paths = net->outflow; paths != NULL; paths=paths->next) 
	    for (side=paths->side, path=paths->path; path != NULL;) {

		if (side == RHS)
		    path->rhs = ReturnLrs(path->rhs);
		else
		    path->lhs = ReturnLrs(path->lhs);

		side = path->side;
		path = path->outflow;
		} /*for*/

    return(YES);

} /*ResetNet*/
