/*
 * PPS - The Parallel Production System - Frank Lopez - pps_questions@siliconvalleyone.com
 * 
 * FILE: scanner.c - This file contains the scanner routines
 *
 * Date: 6-16-87
 *
 */
#include <stdio.h>
#include "common.h"
#include "scanner.h"

GetIA()
{
    if (imgr.input[imgr.ct] == EOS)
	return(EOF);

    return((int)imgr.input[imgr.ct++]);
} /*GetIA*/

UnGetScanChar(ch, fp)
int ch;
FILE *fp;
{
    if (imgr.fn == NULL)
	imgr.ct--;
    else
	ungetc(ch, fp);
}

GetScanChar()
{
    if (imgr.fn == NULL)
	return(GetIA());
    else
	return(getc(imgr.fp));

} /*GetScanChar*/

char *GetWord(mode)
int mode;
{
    char wordsb[ MAXWORDLEN ];
    int ct;		/*counts the number of letters read*/

   for(ct=0; IsWord(ch); ct++, ch = GetScanChar())
	wordsb[ct] = ch;

    wordsb[ct] = EOS;

    return(HashString(wordsb));
}

/*
 *GetToken -- gets the next token from the input buffer.
 */
GetToken(la)
int la; 
{
    int ct;
    int sign;
    float fin;
    char wordsb[ MAXWORDLEN ];

    if (la == YES ) {
	if (laCt > 0) {
	    token.string = LaTokens[laEmpty].string;
	    token.numeric = LaTokens[laEmpty].numeric;
	    token.opcode = LaTokens[laEmpty].opcode;
	    laCt--;
	    laEmpty++;
	    if (laEmpty >= LASIZE)
		laEmpty = 0;
	    return(0);
	    }
	} /*la == YES*/

    DoForever {
	if (ch == EOL) {
	    imgr.ln = imgr.ln + 1;
	    ch = GetScanChar();
	    }
	else if (IsFiller(ch)) {
	    while (IsFiller(ch))
		ch = GetScanChar();
	    }
	else if (ch == ';') {
	    while ((ch != EOL) && (ch != EOF))
		ch = GetScanChar();
	    }
	else 
	    break; /*Found a valid parsing token*/
	} /*Take care of nontoken strings*/

    if (IsAlpha(ch)) {
	token.opcode = WORD;
	token.string = GetWord();
	}
    else
    switch (ch) {
	case ',' :
		token.opcode = COMMA;
		ch = GetScanChar();
		break;

	case '<' :
		if ((ch = GetScanChar()) != '-') {
		    EH("Unknown format '<????'");
		    return(NO);
		    }

		ch = GetScanChar();
		token.opcode =  BINDER;
		break;

	case '"' :
		for (ct=0; (wordsb[ct] = GetScanChar()) != '"';) {

		    if (wordsb[ct] == EOF) {
			EH("End of file found in string");
			return(0);
			}

		    if (wordsb[ct] == EOL)
			EH("Line return found in string");

		    ct++;

		    if (ct >= MAXWORDLEN)
			EH("String length limit exceeded.");
		    }

		ch = GetScanChar();
		wordsb[ct] = EOS;
		token.opcode = WORD;
		token.string = MakeString(wordsb);
		break;

	case '(' :
		token.opcode = LPAREN;
		ch = GetScanChar();
		break;

	case ')' :
		token.opcode = RPAREN;
		ch = GetScanChar();
		break;

	case EOF :
		token.opcode = QUIT;
		break;

	case '=' :

		if ((ch = GetScanChar()) != '>') {
		    EH("Illegal form found '='");
		    break;
		    }

		token.opcode = SEPARATOR;
		ch = GetScanChar();

		break;

	case '?' :
		ch = GetScanChar();

		if (IsAlpha(ch)) {
		    token.string = GetWord();
		    token.opcode = SBIND;
		    }
		else
		    token.opcode = SFREE;

		break;
	default :
		sign = 1;

		if (IsNumeric(ch) || (ch == '-')) {

		    if (ch == '-') {
			sign = -1;
			ch = GetScanChar();
			} /*ch == '-'*/

		    if (ch == '>') {
			token.opcode = SENDER;
			ch = GetScanChar();
			}
		    else if (IsNumeric(ch)) {

		        token.opcode = NUMERIC; 

			UnGetScanChar(ch, imgr.fp);

			if (imgr.fn == NULL) 
			    sscanf(imgr.input + imgr.ct, "%f", &fin);
			else 
			    fscanf(imgr.fp, "%f", &fin);

			token.numeric = fin * sign;

			/*Parse off the read float*/
			for (ch=GetScanChar(); IsNumeric(ch); ch=GetScanChar());

			if (ch == '.') {
			    ch = GetScanChar();

			    if (!(IsNumeric(ch)))
				break;

			    for (; IsNumeric(ch); ch=GetScanChar());

			    if ((ch == 'e') || (ch == 'E')) 
				ch = GetScanChar();

			    for (; IsNumeric(ch); ch=GetScanChar());

			    } /*ch == '.'*/
			} /*IsNumeric*/
		    else {
			EH("-???? is not supported");
			goto GETTOKEN_ERROR;
			}
		    } /*Is numeric or starts with '-' */
		else {
		    if ((ch == EOS) && (imgr.fp == NULL)) {
			token.opcode = EOF;
			break;
			}

		    sprintf(error_sb, "Unknown character found octal %o\n", ch);
		    EH(error_sb);
		    goto GETTOKEN_ERROR;
		    } /*Not Numeric, nor Production*/
		break;
	    } /*switch*/

    if (la == SAVE) {
	if (laCt == LASIZE)		/*DEBUGGING*/
	    EH("Look Ahead buffer is overextended by one.");

	LaTokens[laFill].string = token.string;
	LaTokens[laFill].numeric = token.numeric;
	LaTokens[laFill].opcode = token.opcode;

	laFill++;
	if (laFill >= LASIZE)
	    laFill = 0;

	laCt++; }

    return(YES);

GETTOKEN_ERROR:

    EH("Unknown input token format found.");
    return(NO);

} /*GetToken*/
