/*
 *assert.c - code for the assert function.
 */

#include <stdio.h>
#include <parallel.h>
#include "common.h"

/*
 *ProcessNewString - Take a character string as input and parse it into
 *		     an internal fact format.
 */
struct element *ProcessNewString(string)
char *string;
{
    struct element *pelenew;
    struct element *peleret;
    struct scanner in;

    /*turn on the scanner string mode*/

    /*now have the scanner use this input string*/
    InitScanner(&in);
    in.input = string;
    in.ln = -1;
    in.ch = GetScanChar(&in); /*prime the scanner*/
    
    /*peleret is the pointer to the start of the fact structure*/
    peleret = NULL;

    /*For each token in the assert string*/
    for (GetToken(&in, YES); in.token.opcode != QUIT; GetToken(&in, YES)) {

	    if (peleret != NULL) {
		pelenew->next = allocatestruct(element);
		pelenew = pelenew->next;
		}
	    else
		pelenew = peleret = allocatestruct(element);

	    pelenew->type =in.token.opcode;

	    /*facts only have two types of elements*/
	    switch (in.token.opcode) {
		case NUMERIC :
			pelenew->numeric = in.token.numeric;
			break;

		case WORD :
			pelenew->string =in.token.string;
			break;

		default :
			EH(&in, "Illegal assert format.");
			goto PNSEXIT;
			break;
		} /*switch*/

	} /*for length of assert*/

    if (peleret != NULL)
	pelenew->next = NULL;

PNSEXIT :
    return(peleret);

} /*ProcessNewString*/

/*
 *Assert 
 *	Given a fact as a character string, this function asserts it into
 *      the working memory.
 */
Assert(rs, string, id)
struct defrs *rs;
char *string;
int id;
{

    if (Assertx(rs, ProcessNewString(string), id) == NULL)
	return(NO); /*the string already existed*/

    return(YES); /*assert did take place*/

} /*Assert*/

/*
 *Assertx - assert a fact from the main system.
 */
struct fact *Assertx(rs, newfact, id)
struct defrs *rs;
struct element *newfact;
int id;
{
    struct fact *new;

    if (newfact == NULL)
	return(NULL);

    /*first see if the fact already exists, if it does we do not assert it
      again since it is already known
     */
    if (FactExists(rs, newfact) == YES) 
	return(NULL);

    /*now we create the fact*/
    new = allocatestruct(fact);

    /*
      fact id is computed either as the given id (computed in another EO) or
      as a computed EO which is a fact id for this EO, with this EO's id as
      the least significant positions of the id (see documentation)
     */

    new->id = id;
    new->entry = newfact;
    new->next = NULL;

    new->next = rs->factlist;
    rs->factlist = new;

    new->list = Dataflow(rs, new);

    return(new); /*The fact has been asserted*/

} /*Assert a fact*/
