/*
 *common.h -- Common declaration file.
 */
#define DEBUG 0			/*Debug prompt*/

#define RS_PRIORITY_DEFAULT 0.0
#define DEFRULE_PRIORITY_DEFAULT 0.0

#define COMPUTE_FACT_ID(eo) ((eo->factid++ * imgr.factid_disp) + eo->id)
#define FREEMEM(ptr) share_free((char *)ptr)
#define MALLOCMEM(size) (char *)share_malloc(size)

#define NULLSCANNER (struct scanner *)NULL

extern char *share_malloc();

#define RHSASSERT 0
#define RHSPRINTOUT 1
#define RHSRETRACT 2
#define RHSSTOP 3
#define MAXRHS 4

#define SAVE 2			/*Save the present token: LA*/
#define YES 1
#define NO 0
#define ALL -1

#define RHS 0
#define LHS 1
#define END 2

#define MAXFACTLEN 256		/*Maximum fact length*/
#define MAXWORDLEN 256		/*Maximum word length*/
#define HASHSIZE 91		/*Hash table size*/
#define LASIZE 5		/*Size of the look ahead parser buffer*/

#define EOS '\0'
#define EOL '\n'

#define IsNumeric(ch) ((ch <= '9') && (ch >= '0'))
#define IsAlpha(ch) (((ch >= 'a') && (ch <= 'z')) || ((ch <= 'Z')&&(ch >= 'A')))
#define IsWord(ch) (IsNumeric(ch) || IsAlpha(ch)|| ((ch == '_') || (ch == '-')))
#define IsFiller(ch) ((ch == ' ') || (ch == '\r') || (ch == '\t'))

#define DoForever while(1)

#define allocatestruct(new) (struct new *)MemAlloc(sizeof(struct new))

/*		Globals Everyone should know about.		*/

/*runtime*/
extern char error_sb[];
extern struct interface imgr;
extern struct opt_interface opt;

extern struct values *ValuesCopy();
extern struct values *LocalBind();
extern struct values *CreateValue();

extern char *MemAlloc();
extern char *MakeString();

extern struct element *ProcessNewString();
extern struct fact *Assertx();
extern struct fact *GetFact();
extern struct match *Dataflow();
extern struct values *Compare();

extern struct mailbox *CreateMailbox();
extern struct defeo *CreateEO();
extern struct defrs *CreateRS();
extern struct eo_id *ParseMessages();

extern struct ruleslot *ParsePattern();
extern struct test *ParseTest();
extern struct lr *ParseRHS();
extern struct node *ParseLHS();

/*Global character strings : hashed*/
extern char *ASSERT;
extern char *DECLARE;
extern char *DEFEO;
extern char *DEFFACTS;
extern char *DEFRS;
extern char *DEFRULE;
extern char *PRINTOUT;
extern char *PRIORITY;
extern char *RETRACT;
extern char *SBNULL;
extern char *SBAND;
extern char *STOP;

/*Hash Table*/
extern char *HashString();
extern struct draw *ht[];
extern LOCK *ht_locks[];
extern struct opthash *variables;

/*RHS execution*/
extern int (*rhs[])(); /*Table to the RHS commands*/

/*Opcodes*/
#define WORD		01
#define LPAREN 		02
#define RPAREN		03
#define NUMERIC		04
#define SBIND		05
#define SFREE		06
#define COMMA		07

#define SPECIAL		010
#define QUIT		011

#define BINDER		020
#define SEPARATOR	021
#define ANY		022
#define PATTERN		023
#define EO_IDS		024
#define SENDER		025

/*structures*/

/*la_token: the look ahead token list for the parser*/
struct la_token {
	char *string;
	int opcode;
	int numeric;
	struct la_token *next;
}; /*la_token*/

/*token_main: the token for the look ahead list*/
struct token_main {
	char *string;	/*If string the returned here*/
	float numeric;	/*If numeric value is here*/
	int opcode;		/*Tokens octal representation: opcode*/
}; /*token_main*/

/*struct scanner*/
struct scanner {
    /*look ahead stuff*/
    int laFill;
    int laCt;
    int laEmpty;
    struct la_token LaTokens[ LASIZE ]; /*Lookahead buffer*/

    FILE *fp;
    char *fn;
    char *input;
    int ch;
    int ct;
    int ln;
    struct token_main token;
}; /*scanner*/

/*agenda: Holds the runtime agenda*/
struct agenda {
	int id; /*next agenda id*/
	struct activation *list; /*list of rule activations*/
}; /*struct agenda*/

/*struct defeo is for the expert object tasks*/
struct defeo {
	char *name;
	int id;

	int status;
	long total_rules; 	/*total number of rules fired for this EO*/
	int factid;			/*next fact id*/

	struct mailbox *si; 	/*list of EOs that information is sent to*/

	SEMAPHORE *recv;

	struct eo_id *in_boxes;
	struct eo_id *out_boxes;

	struct defrs *defrs_locals; /*list of rule sets in this expert object*/
	struct deffacts *deffacts;  /*deffacts declared in environment       */
	struct scanner in;

	struct defeo *next; 	/*next eo in the list*/

	/*The following information is needed to determine which rule to fire*/
	struct mailbox *rs_box;
	int rs_priority;
	int rule_priority;

}; /*struct defeo*/

/*interface : the interactive interface*/
struct interface {
	struct scanner in;

	BARRIER *barrier;
	SEMAPHORE  *print;	/*only one task should print at a time*/
	LOCK *lock;		/*mutual exclusion access*/

	int update_prompt;	/*should prompt be updated*/
	int show_prompt;	/*should prompt be echoed*/

	char *stmt_name;	/*current stmt name being parsed*/
	int eo_ids;		  /*eo ids: starts at 1 */

	struct defeo *current_eo; /*current defeo*/
	struct defeo *defeo_lst;  /*list of all defeo*/

	struct defrs *current_rs; /*current defrs*/
	struct defrs *defrs_lst;  /*list of all defrs*/

	struct defrule *current_dr; /*current rule*/

	int watch;		/*watch the execution*/

	char *input;  	/*the input string being parsed in interactive mode*/
	int ct; 	/*present scanner position in parsing that string*/

	int factid_disp;	/*constant factids are multipled by*/

	short save;		/*Checks to see if the buffers have been saved*/
	short error;		/*error flag*/

	struct match *pmatchList;
}; /*Interface_Manager*/

/*struct defrs - the rule set task structure*/
struct defrs {
  	struct scanner in;

	char *name; /*name of the defrs*/
	int priority;

	int status;

	struct defrule *defrules;
	struct defeo *eo;

	struct pattern *patterns;
	struct agenda Agenda;
	struct fact *blackboard;

	SEMAPHORE *out_sema;
	struct mailbox *out_box;

	SEMAPHORE *in_sema;
	struct mailbox *in_box;

	struct defrs *next;

}; /*struct defrs*/

/*activation: holds information pertinent for each rule's activation*/
struct activation {
	char *name;
	int priority;
	int id;			/*activation id*/
	struct decidernode *rhs;  /*rhs of the rule*/
	struct values *variables; /*variable bindings*/
	struct activation *next;  /*next activation in the agenda*/
}; /*struct activation*/

/*datalink: For each pattern there is a list of links to LHS's of various
	    rules that used the pattern, this list is created using datalinks*/
struct datalink {
	char boolean;		
	int id; 		  /*id number for this rules LHS*/
	int tally;
	short side;		  /*Side it enters decidernode from*/
	struct values *binder;    /*If set then we have a variable binding*/
	struct decidernode *path; /*Path to decidernode in net*/
	struct datalink *next;    /*net link in list*/
}; /*struct datalink*/

/*decidernode: used to create the internal net structure*/
struct decidernode {
	int side;
	struct lr *lhs, *rhs;
	struct decidernode *outflow;
}; /*struct decidernode*/

/*deffacts: holds information about a parsed deffacts statement*/
struct deffacts {
	char *name; /*Name of the deffacts statement*/
	char *stmt; /*The actual deffacts string*/
	char **facts; /*facts in character string format*/
	struct deffacts *next;
}; /*struct deffacts*/

/*defrule: holds information about a parsed defrule statement*/
struct defrule {
	char *name; /*Name of the defrule stmt*/
	char *stmt; /*The English form of the defrule stmt.*/
	int priority;
	struct defrule *next;
}; /*struct defrule*/

/*draw: holds one of the draws in the hash table*/
struct draw {
	char *entry;
	struct draw *next;
}; /*struct draw*/

/*element: each element of a fact is held in this structure*/
struct element {
	char *string;
	float numeric;
	int id;
	int type;
	struct element *next;
}; /*struct element*/

/*eo_id: used to keep a list of eo ids.*/
struct eo_id {
	int status;
	char *name;
	int id;
	struct eo_id *next;
	SEMAPHORE *sema;
	struct mailbox *box;
}; /*struct eo_id*/

/*fact: each fact has a header stmt, fact id does this*/
struct fact {
	int id;
	struct element *entry;
	struct match *list; /*the list of patterns this fact matched*/
	struct fact *prev, *next;
}; /*struct fact*/

/*factid: for each pattern a list of fact ids that matched it is created
	  using this structure*/
struct factid {
	int id;
	struct factid *next;
}; /*struct factid*/

/*lr: for each node in the net a list of input tokens is kept for its
	  left and right input, this list is creating using struct lr*/
struct lr {
	struct values *locals; /*variable bindings for this level*/
	struct lr *next; /*next lr*/
}; /*struct lr*/

/*mailbox*/
struct mailbox {
	int id;
	int box_id;
	short status; 
	/*
	Has the first message come? 
	Have all messages come from this eo?
	*/

	LOCK *lock; /*Since the use of an box is only to add or delete a 
		      message (quick), the box is locked (busy waiting). */
	char *name;
	char **mesgs;
	struct mailbox *next;

}; /*struct mailbox*/

/*match: list of all patterns that matched a fact*/
struct match {
	struct pattern *slot; /*pattern matched*/
	struct match *next; /*linked list of matched patterns*/
}; /*struct match*/

/*node: a temporary structure for the LHS of a rule, this allows
	one to perform intermediate operations on the structure,
	and then call CreateNet which will then create the final net*/
struct node {
	char *string; /*string values storage*/
	float numeric; /*numeric values storage*/
	int opcode; /*What type of a node is it*/
	struct node *bottom;  /*next operand in list*/
	struct node *right;  /*next node element*/
	struct ruleslot *pattern; /*pattern parsed*/
  	struct node *next; /*next node*/
}; /*struct node*/

/*pattern: header for a pattern*/
struct pattern {
	struct ruleslot *pattern;
	struct datalink *outflow;
	struct pattern *next;
}; /*struct pattern*/

/*ruleslot: the elements of a pattern*/
struct ruleslot { /*used to represent slot function representations*/
	char *string; /*storage for strings*/
	float numeric; /*storage for numbers*/
	int opcode; /*interpreter opcode*/
	struct ruleslot *next; 
}; /*struct ruleslot*/

/*
 *values is a general purpose structure mostly used for token list construction
 */
struct values {
	char *name;
	char *string;
	float numeric;
	int opcode;
	int type;
	int whoset;

	union {
	struct fact *origin;
	struct eo_id *eo_id;
	} oe;

	struct values *next;

}; /*struct values*/

/* RHS functions*/
extern int ExecuteAssert(); 
extern int ExecuteRetract();
extern int ExecutePrintOut();
extern int ExecuteStop();

#define SYSTEM_MEMORY 150000       /*size of shared memory*/
