/*
 *main.c - contain the main functions needed for he global execution of the
 *         system.
 */

#include <parallel.h>
#include <stdio.h>
#include "common.h"		/*Should always be the first header*/
#include "main.h"
#include "comm.h"

extern FILE *fopen();

main(argc, argv) 
int argc;
char *argv[];
{

    Init();

    InitScanner(&imgr.in);

    if (argc == 1) {
	EH(NULLSCANNER, "Missing input file on command line");
	exit(1);
	} /*if*/
	
    if (argc > 2) 
	imgr.watch = NO;

    imgr.in.fn = argv[1];

    if ((imgr.in.fp = fopen(argv[1], "r")) == NULL) {
	EH(NULLSCANNER, "file doesn't exist");
	return(NO);
	}

    imgr.in.ch = GetScanChar(&imgr.in);

    Parser(&imgr.in);

    Reset();

    if (imgr.error == NO)
	Run();

    return(YES);

} /*main*/

/*
 *Reset
 */
Reset()
{
    int ct;
    int index = 1;

    imgr.current_eo->defrs_locals = imgr.defrs_lst;

    /*Set the new offset for the fact ids*/
    ct = imgr.eo_ids;

    do {
	ct = (int)(ct / 10);
	index *= 10;
	} while (ct > 1);

    /*This whole idea has to be refined for interactive mode*/
    /*Since EO's can be added and deleted between runs*/
    imgr.factid_disp = index; /*Create a buffer of factor of 10 for now*/

    SetUp_EOs(imgr.defeo_lst);

    CreateInBoxes(imgr.defeo_lst);

    return(YES);

} /*Reset*/

/*
 *Run
 */
Run()
{

    imgr.current_eo->defrs_locals = imgr.defrs_lst;

    StartEOs();

    PrintSummary();

    printf("Execution has completed.\n");

    return(YES);

} /*Run*/

/*
 *CleanAgenda
 */
CleanAgenda(rs, id)
struct defrs *rs;
int id; /*Id to remove from the list*/
{
    struct activation *search, *last, *temp;

    for (search=rs->Agenda.list; search != NULL;) {

	if (AnyId(id, search->variables) == NO) {
	    last = search;
	    search = search->next;
	    continue;
	    } /*AnyId*/

	if (search == rs->Agenda.list)
	    rs->Agenda.list = search->next;
	else
	    last->next = search->next;

	temp = search;
	search = search->next;
	
	ReturnValues(temp->variables);
	FREEMEM(temp);

	} /*for*/

} /*CleanAgenda*/

/*
 *HASHTABLE: Our hash table for all of the strings in the system,
 *           for only one of any string can exist in the system,
 *           so comparison between strings can be done by simply 
 *           seeing if they have the same memory address.
 */

/*
 *MakeString - create a new string.
 */
char *MakeString(str)
char *str;
{
    char *new;

    new = MALLOCMEM(strlen(str) + 1);

    if (new == NULL)
	EH(NULLSCANNER, "OUT OF DYNAMIC SPACE!!!!!!");

    strcpy(new, str);

    return(new);

} /*MakeString*/

/*
 *HashString - returns the systems address for the string given as
 *             its parameter.
 */
char *HashString(string)
char *string;
{
    int entry;
    struct draw *pdrawPeek;
    char *calc;
    int tally;

    entry = 0;
    tally = 3;

    calc = string;

    while ((*calc != EOS) && (tally-- != 0))
	entry += (int)*(calc++);

    entry = entry % HASHSIZE;
     
    spin_lock(ht_locks);

    for (pdrawPeek=ht[entry]; pdrawPeek != NULL; pdrawPeek=pdrawPeek->next) 
	if (strcmp(string, pdrawPeek->entry) == 0)
	    break;

    if (pdrawPeek == NULL) { /*Isn't in the hash table*/
	pdrawPeek = allocatestruct(draw);
	pdrawPeek->entry = MakeString(string);
	pdrawPeek->next = ht[entry];
	ht[entry] = pdrawPeek;
	} /*if*/

    spin_unlock(ht_locks);

    return(pdrawPeek->entry); /*Return pointer to the new string*/

} /*HashString*/ 

/*
 *Equal - are two facts equal.
 */
Equal(peleold, pelenew)
struct element *peleold, *pelenew;
{
    while ((peleold != NULL) && (pelenew != NULL)) {
	if (peleold->type != pelenew->type)
	    return(NO);

	switch (peleold->type) {
	    case WORD :
		if (peleold->string != pelenew->string)
		    return(NO);
		break;

	    case NUMERIC :
		if (peleold->numeric != pelenew->numeric)
		    return(NO);
		break;

	    default :
		EH(NULLSCANNER, "Equal: unknown element type!");
		break;
	    } /*switch*/
	
	peleold = peleold->next;
	pelenew = pelenew->next;
	} /*while still equal length*/

/*Think about this: can this if be done faster*/

    if ((pelenew == peleold) && (pelenew == NULL))
	return( YES );

    return( NO );
} /*Equal*/

/*
 *EH - prints error messages to the user screen.
 */
EH(in, mesg)
struct scanner *in;
char *mesg;
{
    imgr.error = YES;

    if (in == NULLSCANNER)
	fprintf(stderr, "ERROR: <%s>.\n", mesg);
    else {
	in->token.opcode = EOF;

	if (imgr.stmt_name != NULL) 
	    fprintf(stderr, "In rule %s ::: ", imgr.stmt_name);

	fprintf(stderr,"Fatal: ln %d in %s: %s\n", in->ln, in->fn, mesg);
	} /*else*/
   
} /*EH*/

/*
 *AddRule - Add a new defrule to the db.
 *
 *	 cond == 0: Add the defrule's statement to its structure.
 *	 cond == 1: Add the name of the defrule to the list of known defrules.
 */
AddRule(name, stmt, cond)
char *name, *stmt;
int cond;
{
    struct defrule *newrule;

    /*Does it exist?*/
    for (newrule=imgr.current_rs->defrules; newrule != NULL; newrule=newrule->next)
	if (strcmp(newrule->name, name) == 0) {
	    if (cond == 0)
		break;

	    sprintf(error_sb, "Rule %s redeclared.", name);
	    EH(NULLSCANNER, error_sb);
	    return(NO);
	    }
    
    switch (cond) {
	case 0 :
		if (newrule == NULL) {
		    EH(NULLSCANNER, "Rule not found - can not add statement.");
		    return(NO);
		    }

		newrule->stmt = MemAlloc(strlen(stmt) + 1);
		strcpy(newrule->stmt, stmt);
		break;

	case 1 :
		imgr.current_dr = newrule = allocatestruct(defrule);
		newrule->priority = DEFRULE_PRIORITY_DEFAULT;
		newrule->name = name;
		newrule->next = imgr.current_rs->defrules;
		imgr.current_rs->defrules = newrule;
		break;

	} /*switch*/

    return(YES);

} /*AddRule*/

/*
 *Init()
 */
Init()
{
    int i;

    /*allocates memory for the system*/
    if (share_malloc_init( SYSTEM_MEMORY ) != 0) {
	EH(NULLSCANNER, "NO SYSTEM MEMORY: share_malloc_init");
	exit(0);
	} /*if*/

    /*Set the mode for the interface manager*/
    imgr.watch = YES ;
    imgr.lock = spin_create( PAR_UNLOCKED );

    if ((imgr.print = semaphore_create( TASK_BLOCK )) == NULL)
	EH(NULLSCANNER, "Semaphore print failed in creation.");

    sprintf(mesg_end, "%d", MESG_END );
    sprintf(mesg_stopped, "%d", MESG_STOPPED );
    sprintf(mesg_eo_stopped, "%d", MESG_EO_STOPPED);
    sprintf(mesg_na, "%d", MESG_NA );
    sprintf(mesg_fire, "%d", MESG_FIRE );

    /*Always initialize the hash table first, before all other initializations*/
    for (i=0; i < HASHSIZE; i++) {
	ht[i] = NULL;
	ht_locks[i] = spin_create( PAR_UNLOCKED );
	} /*for*/

    rhs[ RHSASSERT ] = ExecuteAssert;
    rhs[ RHSPRINTOUT ] = ExecutePrintOut;
    rhs[ RHSRETRACT ] = ExecuteRetract;
    rhs[ RHSSTOP ] = ExecuteStop;

    imgr.error = NO;

    imgr.eo_ids = 1;
    imgr.current_eo = NULL;
    imgr.defeo_lst = NULL;

    imgr.current_rs = NULL;
    imgr.defrs_lst = NULL;


    ASSERT = HashString("assert");
    DECLARE = HashString("declare");
    DEFEO = HashString("defeo");
    DEFFACTS = HashString("deffacts");
    DEFRS = HashString("defrs");
    DEFRULE = HashString("defrule");
    PRINTOUT = HashString("printout");
    PRIORITY = HashString("priority");
    RETRACT = HashString("retract");
    SBNULL = HashString("");
    SBAND = HashString("and");
    STOP = HashString("stop");

} /*Init*/

/*
 *CreateRS (see struct definition for comments about defrs declaration)
 */
struct defrs *CreateRS()
{
    struct defrs *new;

    new = allocatestruct(defrs);

    new->priority = RS_PRIORITY_DEFAULT;
    new->status = EO_READY;

    InitScanner(&new->in);

    new->Agenda.id = 0;
    new->Agenda.list = NULL;

    new->out_box = CreateMailbox();
    new->in_box = CreateMailbox();

    /*semaphores block the task, not the processs*/
    new->out_sema = semaphore_create( TASK_BLOCK );
    new->in_sema = semaphore_create( TASK_BLOCK );

    /*initialize the local data list*/
    new->blackboard = NULL;
    new->patterns = NULL;
    new->defrules = NULL;
    new->next = NULL;

    return(new);

} /*CreateRS*/

/*
 *CreateEO
 */
struct defeo *CreateEO()
{
    struct defeo *new;

    if (imgr.current_eo != NULL)
	imgr.current_eo->defrs_locals = imgr.defrs_lst;

    imgr.defrs_lst = NULL;
    imgr.current_rs = NULL;

    new = allocatestruct(defeo);

    InitScanner(&new->in);

    new->defrs_locals = NULL;
    new->total_rules = 0;
    new->factid = 1;
    new->status = EO_INIT;
    new->deffacts = NULL;
    new->rs_box = NULL;

    new->id = imgr.eo_ids++;

    new->in_boxes = NULL;
    new->recv = semaphore_create( TASK_BLOCK );

    new->out_boxes = NULL;

    return(new);

} /*CreateEO*/

/*
 *CreateMailbox
 */
struct mailbox *CreateMailbox()
{
    struct mailbox *new;
    static box_id = 1000;

    new = allocatestruct(mailbox);

    new->lock = spin_create( PAR_UNLOCKED );
    new->status = MESG_START;
    new->box_id = box_id++;

    new->mesgs = INIT_BUF;
    *new->mesgs = NULL;

    return(new);

} /*CreateMailbox*/

/*
 *InitScanner
 */
InitScanner(in)
struct scanner *in;
{

    in->laCt = in->laFill = in->laEmpty = 0;
    in->ln = 1;
    in->input = NULL;
    in->ct = 0;
    in->fn = NULL;
    in->ch = EOF;

} /*InitScanner*/
