/*
 *parser.c - contains the PPS parser.
 */

#include <parallel.h>
#include <stdio.h>
#include "common.h"
#include "comm.h"

/*
 *Parser - Parses the input during batch mode.
 */
Parser(in)
struct scanner *in;
{

    GetToken(in, YES);

    while (in->token.opcode != QUIT) {

	if (in->token.opcode != LPAREN) {
	    EH(in, "Expected a '('");
	    return(NO);
	    }

	GetToken(in, YES);

	if (in->token.opcode != WORD)  {
	    EH(in, "Expected an identifier after '('");
	    return(NO);
	    }

	if (in->token.string == DEFEO) {
	    GetToken(in, YES);
	    if (ParseEO(in) == NO) 
		return(NO);
	    } /*DEFEO*/
	else {
	    sprintf(error_sb, "%s is not a valid construct.", in->token.string);
	    EH(in, error_sb);
	    return(NO);
	    }

	} /*while not EOF*/

    return(YES);

} /*Parser*/

/*
 *ParseMessages
 */
struct eo_id *ParseMessages(in)
struct scanner *in;
{
    struct eo_id *new, *trav;

    /*Defaults to itself -> present expert object*/
    if (in->token.opcode != SENDER) {
	new = allocatestruct(eo_id);
	new->name = imgr.current_eo->name;
	new->id = imgr.current_eo->id;
	new->status = WORD;
	new->next = NULL;

	AddEO_to_EO_Boxes(imgr.current_eo->name, new->name, OUT_BOX);

	return(new);
	} /*SENDER*/

    new = NULL;

    GetToken(in, YES);

    while (in->token.opcode == WORD) {

	trav = allocatestruct(eo_id);
	trav->next = new;

	new = trav;
	new->name = in->token.string;
	new->status = WORD;

	AddEO_to_EO_Boxes(imgr.current_eo->name, new->name, OUT_BOX);

	GetToken(in, YES);

	if (in->token.opcode != COMMA)
	    break;

	GetToken(in, YES);

	if (in->token.opcode != WORD) {
	    EH(in, "malformed statement, expected a name after ','");
	    return(NO);
	    }

	} /*while*/

    return(new);

} /*ParseMessages*/

/*
 *ParseEO
 */
ParseEO(in)
struct scanner *in;
{

    if (in->token.opcode != WORD) {
	EH(in, "Expected the name of the expert object.");
	return(NO);
	} /*if*/

    if (imgr.current_eo != NULL)
	imgr.current_eo->defrs_locals = imgr.defrs_lst;

    imgr.current_eo = CreateEO();
    imgr.current_eo->name = in->token.string;

    imgr.defrs_lst = NULL;
    imgr.current_rs = NULL;

    imgr.current_eo->next = imgr.defeo_lst;
    imgr.defeo_lst = imgr.current_eo;

    GetToken(in, YES);

    if (in->token.opcode == WORD) 
	GetToken(in, YES);

    if (in->token.opcode != LPAREN) {
	EH(in, "Malformed expert object definition.");
	return(NO);
	}


    while (in->token.opcode == LPAREN) {

	GetToken(in, YES);

	if (in->token.opcode != WORD) {
	    EH(in, "Malformed expert object definition.");
	    goto PARSE_EO_ERROR;
	    }

	if (in->token.string == DEFRS) {
	    if (ParseRS(in) == NO)
		goto PARSE_EO_ERROR;
	    }
	else if (in->token.string == DEFFACTS) {
	    if (ParseFacts(in) == NO) 
		goto PARSE_EO_ERROR;
	    }
	else {
	    EH(in, "Unsupported declaration.");
	    goto PARSE_EO_ERROR;
	    }
	} /*while*/

    if (in->token.opcode != RPAREN) {
	EH(in, "Missing closing ')' in expert object definition.");
	goto PARSE_EO_ERROR;
	}

    GetToken(in, YES);

    imgr.current_eo->defrs_locals = imgr.defrs_lst;

    return(YES);

PARSE_EO_ERROR :

    /*Should FREEMEM rulesets and the expert object*/
    return(NO);

} /*ParseEO*/

/*
 *ParseRS
 */
ParseRS(in)
struct scanner *in;
{
    struct defrs *rs;

    /*Add another rule set to the main list of rule sets*/
    rs = imgr.current_rs = CreateRS();

    rs->next = imgr.defrs_lst;
    imgr.defrs_lst = rs;

    rs->priority = RS_PRIORITY_DEFAULT;

    imgr.current_rs->patterns = NULL;

    /*Parse the new rule set*/
    GetToken(in, YES); /*Looking for rule set name*/

    if (in->token.opcode != WORD) {
	EH(in, "Missing rule set name");
	goto ERROR_DEFRS;
	}

    if (RS_Exists(in->token.string) == YES) {
	sprintf(error_sb, "Redeclaration of rule set %s", in->token.string);
	EH(in, error_sb);
	goto ERROR_DEFRS;
	}

    rs->in_box->name = in->token.string;
    rs->in_box->id = -1;

    rs->out_box->name = imgr.current_eo->name;
    rs->out_box->id = imgr.current_eo->id;

    rs->name = in->token.string;

    printf("Compiling rule set %s.\n", rs->name);

    GetToken(in, SAVE);

    if (in->token.opcode == WORD) /*Is the comment string there? */
	GetToken(in, YES);

    GetToken(in, YES);

    while (in->token.opcode != RPAREN) {

	if (in->token.opcode != LPAREN) {
	    EH(in, "Missing '(' in defrule statement.");
	    goto ERROR_DEFRS;
	    }

	GetToken(in, YES);

	if (in->token.opcode != WORD)  {
	    RS_FORMAT:
	    EH(in, "Illegal format.");
	    goto ERROR_DEFRS;
	    }

	if (in->token.string == DECLARE) {

	    GetToken(in, YES);

	    if (in->token.opcode != LPAREN)
		goto RS_FORMAT;

	    GetToken(in, YES);

	    if (in->token.opcode != WORD) 
		goto RS_FORMAT;

	    if (in->token.string == PRIORITY) {
		GetToken(in, YES);

		if (in->token.opcode != NUMERIC)
		    goto RS_FORMAT;

		imgr.current_rs->priority = in->token.numeric;

		GetToken(in, YES);
		}
	    else {
		EH(in, "Unknown type of declare statement in defrs");
		goto ERROR_DEFRS;
		}

	    if (in->token.opcode != RPAREN)
		GetToken(in, YES);
	    } /*declare*/
	else if (in->token.string == DEFRULE) 
	    ParseRule(in);
	else {
	    EH(in, "Unknown statement in defrs");
	    goto ERROR_DEFRS;
	    }
	} /*while*/

    GetToken(in, YES);

    return(YES);

ERROR_DEFRS:

    imgr.defrs_lst = imgr.defrs_lst->next;
    FREEMEM(imgr.current_rs);
    return(NO);

    /*
     *should FREEMEM the other elements of the bad ruleset ....***********
     */
} /*ParseRS*/

/*
 *ParseRule
 */
ParseRule(in)
struct scanner *in;
{
    extern int NodeId;
    extern struct decidernode *CreateNet();
    struct decidernode *in_node;
    struct decidernode *state;
    struct node *pnode = NULL;

    ReturnMatch(imgr.pmatchList);
    imgr.pmatchList = NULL;

    GetToken(in, YES); /*Looking for rule name*/

    if (in->token.opcode != WORD) {
	EH(in, "Missing Rulename");
	return(NO);
	} /*if*/

    imgr.stmt_name = in->token.string;

    /*if NO then the rulename already existed*/
    if (AddRule(in->token.string, (char *)NULL, YES) == NO)
	return(NO);

    printf("Compiling %s.\n", imgr.stmt_name);

    GetToken(in, SAVE);

    if (in->token.opcode == WORD) /*comment string*/
	GetToken(in, YES);

  DoForever { /*Parse Declarations*/

    if (in->token.opcode != LPAREN)
	break;

    GetToken(in, SAVE);

    if (in->token.opcode != WORD)
	break;

    if (in->token.string != DECLARE)
	break;

    GetToken(in, SAVE);

    if (in->token.opcode != LPAREN) 
	break;

    GetToken(in, YES); /*Burn off the lookahead*/
    GetToken(in, YES);
    GetToken(in, YES);
    GetToken(in, YES); /*Get Next Token*/

    if (in->token.opcode != WORD) {
	EH(in, "Missing declarations command after '('");
	return(NO);
	}

    if (in->token.string == PRIORITY) {

	GetToken(in, YES);

	if (in->token.opcode != NUMERIC) {
	    EH(in, "A Rules Priority must be a numeric");
	    return(NO);
	    } /*if*/
		
	imgr.current_dr->priority = in->token.numeric;

	} /*Priority*/
    else {
	EH(in, "<declare command> unknown!");
	return(NO);
	} /*else*/

    GetToken(in, YES);

    if (in->token.opcode != RPAREN) {
	EH(in, "Missing ')' ");
	return(NO);
	} /*if*/

    GetToken(in, YES);

    if (in->token.opcode != RPAREN) {
	EH(in, "Missing ')' ");
	return(NO);
	} /*if*/

    GetToken(in, SAVE);
    } /*DoForever*/

    GetToken(in, YES); /*Start parsing right on the '(' */

    if ((pnode = ParseLHS(in)) == NULL)
	return(NO);

    state = allocatestruct(decidernode);

    state->rhs = state->lhs = NULL;
    state->outflow = NULL;
    state->side = END;

    if ((in_node = CreateNet(pnode, LHS)) == NULL)
	return(NO);

    in_node->outflow = state;
    in_node->side = LHS;
    /*Now Parsing the RHS of the rule*/

    if (in->token.opcode != SEPARATOR) {
	EH(in, "Missing '=>' in format.");
	return(NO);
	}

    GetToken(in, YES);

    /*Add the RHS to the final node of the net.*/
    state->lhs = NULL;
    state->rhs = allocatestruct(lr);
    state->rhs->locals = allocatestruct(values);
    state->rhs->locals->name = imgr.stmt_name;
    state->rhs->locals->opcode = WORD;
    state->rhs->locals->numeric = imgr.current_dr->priority;
    state->rhs->locals->next = NULL;

    if ((state->rhs->next = ParseRHS(in)) == NULL)
	return(NO);

    if (in->token.opcode != RPAREN) {
	EH(in, "Every rule must have an enclosing set of parenthesis.");
	return(NO);
	}

    GetToken(in, YES);

    ReturnNode(pnode); /*Give back all that space*/
    pnode = NULL;

    ReturnMatch(imgr.pmatchList);
    imgr.pmatchList = NULL;

    printf("Parse Rule successfully executed. +++++++++++\n");
    if (imgr.error == NO)
	return(YES); /*Successful execution*/
    else
	return(NO);

} /*ParseRule*/

/*
 *ParseFacts - Parse a Deffacts statement.
 */
ParseFacts(in)
struct scanner *in;
{
    char fact[ MAXFACTLEN ];
    char **facts;
    char **temp;
    int ct = 0;

    GetToken(in, YES);

    if (in->token.opcode != WORD) {
	EH(in, "Missing Deffacts name.");
	return(NO);
	}

    imgr.stmt_name = in->token.string;

    /*If it returns NO then the name already existed*/
    if (AddFacts(1, in->token.string, SBNULL, (char **)NULL) == NO)
	return(NO);

    printf("Reading in Deffacts %s\n", in->token.string);

    GetToken(in, YES);

    if (in->token.opcode == WORD) {
	printf("Processing facts: %s.\n", in->token.string);
	GetToken(in, YES);
	} /*comment parsing*/

    /*Create a list of strings that are the facts from the deffacts stmt*/
    facts = INIT_BUF;

    for (*facts = NULL; in->token.opcode != RPAREN; ct = 0, GetToken(in, YES)) {

	if (in->token.opcode != LPAREN) {
	    EH(in, "Missing '(' in deffacts stmt.");
	    goto CleanUp_Deffacts;
	    }

	GetToken(in, YES);

	for (;in->token.opcode != RPAREN; GetToken(in, YES)) {

	    if (in->token.opcode == WORD) 
		sprintf(fact + ct, "%s ", in->token.string);
	    else
		sprintf(fact + ct, "%6.2f ", in->token.numeric);

	    while (fact[ct] != '\0')
		ct++;

	    } /*for fact*/

	facts = AddMessage(facts, MakeString(fact));

	} /*while for facts*/

    GetToken(in, YES); /*burn off last ')'*/

    if (*facts == NULL) {
	EH(in, "Invalid deffacts statement");
	FREEMEM(facts);
	return(NO);
	}

    AddFacts(2, imgr.stmt_name, SBNULL,  facts);

    return(YES);

CleanUp_Deffacts:

    for (temp = facts; *temp != NULL; temp++)
	FREEMEM(*temp);

    FREEMEM(facts);

    return(NO);

} /*ParseFacts*/
		
/*
 *ParseLHS - Parse the left hand side of a rule.
 */
struct node *ParseLHS(in)
struct scanner *in;
{
    struct node *head, *rtnode;

    /*Create the head of the LHS list of patterns*/
    rtnode = head = allocatestruct(node);
    head->bottom = NULL;

    /*Loop for all patterns in the left hand side*/
    DoForever {

	/*Parse all ?<var> <- statements  (binder)*/
	if (in->token.opcode == SBIND) {

	    rtnode->opcode = PATTERN;
	    rtnode->string = in->token.string;

	    GetToken(in, YES);

	    if (in->token.opcode != BINDER) {
		EH(in, "Malformed statement: expected '<-'");
	 	return(NULL);
		} /*if*/
	
	    GetToken(in, YES);

	    if (in->token.opcode != LPAREN) {
	      sprintf(error_sb, "Expected ?%s<-( <pattern> ) ", rtnode->string);
		EH(in, error_sb);
		return(NULL);
		} /*if*/

	    /*Parse the pattern now*/
	    if ((rtnode->pattern = ParsePattern(in)) == NULL)
		return(NULL);

	    } /*if SBIND*/
	else {

	    /*We have a pattern, so parse it*/
	    if (in->token.opcode != LPAREN)
		break;

	    /*Set the correct type of code*/
	    rtnode->opcode = PATTERN;
	    rtnode->string = SBNULL;
  	    rtnode->right = NULL;

	    if ((rtnode->pattern = ParsePattern(in)) == NULL)
		return(NULL);

	    } /*else*/

	if ((in->token.opcode == SBIND) || (in->token.opcode == LPAREN)) {
	    rtnode->bottom = allocatestruct(node);
	    rtnode = rtnode->bottom;
	    rtnode->bottom = NULL;
	    } /*if*/	

	} /*DoForever*/

    return(head);

} /*ParseLHS*/

/*
 *ParseRHS - Parse the Right Hand Side of a rule.
 */
struct lr *ParseRHS(in)
struct scanner *in;
{
    struct lr *plrHead, *plrtemp; /*temp to traverse lr struct. */
    struct values *nextslot; /*temp to traverse values struct*/

    nextslot = NULL;
    plrHead = plrtemp = NULL;

    DoForever {

	if (plrtemp == NULL) {
	    plrtemp = plrHead = allocatestruct(lr);
	    }
	else {
	    plrtemp->next = allocatestruct(lr);
	    plrtemp = plrtemp->next;
	    } /*else*/

	plrtemp->locals = allocatestruct(values);
	nextslot = plrtemp->locals;
	
	if (in->token.opcode != LPAREN)
	    break;

	GetToken(in, YES);

	nextslot->opcode = in->token.opcode;
	nextslot->string = in->token.string;

	if (in->token.opcode != WORD) {
	    EH(in, "Unknown RHS format.");
	    goto PARSE_ERROR;
	    } /*if*/

	if (in->token.string == STOP) {
	    nextslot->whoset = RHSSTOP;

	    GetToken(in, YES);
	    } /*if*/
	else if (in->token.string == ASSERT) {

	    nextslot->whoset = RHSASSERT;

	    GetToken(in, YES);

	    if (in->token.opcode != LPAREN) {
		EH(in, "Expected '(' in assert statement.");
		goto PARSE_ERROR;
		} /*if*/

	    GetToken(in, YES);

	    if (in->token.opcode == RPAREN) {
		EH(in, "Asserting nothing is an illegal statement.");
		goto PARSE_ERROR;
		} /*if*/
		
	    /*Parse the entire assert statement*/
	    while (in->token.opcode != RPAREN) {
		nextslot->next = allocatestruct(values);
		nextslot = nextslot->next;
		nextslot->opcode = in->token.opcode;

		switch (in->token.opcode) {

			case SBIND :
			case WORD :
				nextslot->string = in->token.string;
				break;

			case NUMERIC :
				nextslot->numeric = in->token.numeric;
				break;

			default :
				EH(in, "Unsupported feature found in assert.");
				return(NULL);
				break;
			} /*assert case*/

		GetToken(in, YES);
		
		} /*while not RPAREN*/

	    GetToken(in, YES);

	    nextslot->next = allocatestruct(values);
	    nextslot = nextslot->next;
	    nextslot->opcode = EO_IDS;
	    nextslot->oe.eo_id = ParseMessages(in);

	    } /*ASSERT*/
	else if (in->token.string == RETRACT) {

	    nextslot->whoset = RHSRETRACT;

	    GetToken(in, YES);

	    if (in->token.opcode != LPAREN) {
		EH(in, "Missing '(' before list of variables.");
		goto PARSE_ERROR;
		} /*if*/

	    GetToken(in, YES);

	    if (in->token.opcode != SBIND) {
		EH(in, "Need at least one rule binded variable");
		goto PARSE_ERROR;
		} /*if*/

	    do {
		nextslot->next = allocatestruct(values);
		nextslot = nextslot->next;
		nextslot->opcode = in->token.opcode;
		nextslot->string = in->token.string;

		GetToken(in, YES);

		} while (in->token.opcode == SBIND);

	    GetToken(in, YES);

	    plrtemp->locals->next->oe.eo_id = ParseMessages(in);

	   } /*RETRACT*/
	else if (in->token.string == PRINTOUT) {

	    nextslot->whoset = RHSPRINTOUT;

	    GetToken(in, YES);

	    if (in->token.opcode != LPAREN) {
		EH(in, "Expected '(' in printout statement.");
		return(NULL);
		} /*if*/

	    GetToken(in, YES);

	    if (in->token.opcode == RPAREN) {
		EH(in, "A printout without any statement is illegal.");
		goto PARSE_ERROR;
		} /*if*/
		
	    /*Parse the entire assert statement*/
	    while (in->token.opcode != RPAREN) {
		nextslot->next = allocatestruct(values);
		nextslot = nextslot->next;
		nextslot->opcode = in->token.opcode;

		switch (in->token.opcode) {
			case SBIND :
			case WORD :
			
				nextslot->string = in->token.string;
				break;

			case NUMERIC :

				nextslot->numeric = in->token.numeric;
				break;

			default :
			
				EH(in, "Unsupported feature found in printout.");
				goto PARSE_ERROR;
				break;
			} /*printout case*/

	  	    GetToken(in, YES);
		
		    } /*while not RPAREN*/

	if (in->token.opcode != RPAREN) {
	    EH(in, "Expected ')' after retract variables.");
	    goto PARSE_ERROR;
	    }

	GetToken(in, YES);

	    } /*printout*/
	else {
	    EH(in, "Unsupported RHS function.");
	    goto PARSE_ERROR;
	    }


	if (in->token.opcode != RPAREN) {
	    EH(in, "Expected ')' ");
	    goto PARSE_ERROR;
	    } /*if*/

	GetToken(in, YES);

	nextslot->next = NULL;

	if (in->token.opcode != LPAREN)
	    break;

	} /*Do for each RHS statement*/

    if (plrHead == NULL) {
	EH(in, "You cannot have a nonempty LHS of a rule.");
	goto PARSE_ERROR;
	} /*if*/

    plrtemp->next = NULL;
    return(plrHead);

PARSE_ERROR :
    EH(in, "Illegal RHS of rule");

    /*Give back any allocated space*/
    if (plrHead != NULL)
	plrtemp = NULL;

    if (nextslot != NULL)
	nextslot->next = NULL;

    ReturnLrs(plrHead);
    return(NULL);

} /*ParseRHS*/

/*
 *ParsePattern - Parse a Left Hand Side pattern slot.
 */
struct ruleslot *ParsePattern(in)
struct scanner *in;
{
    struct ruleslot *head, *trav;

    /*head will point to the parsed pattern*/
    head = trav = NULL;

    GetToken(in, YES);	/*Get the first element of the pattern*/

    /*Parse everything up to the next ')' */
    while (in->token.opcode != RPAREN) {

	if (head == NULL)
	    trav = head = allocatestruct(ruleslot);
	else {
	    trav->next = allocatestruct(ruleslot);
	    trav = trav->next;
	    }

	/*Case each possible token type possible after connectors*/
	switch (in->token.opcode) {
	    case SBIND :
	    case WORD :
		trav->opcode = in->token.opcode;
		trav->string = in->token.string;
		break;

	    case NUMERIC :
		trav->opcode = in->token.opcode;
		trav->numeric = in->token.numeric;
		break;

	    default: 
		EH(in, "Unknown pattern format found");
		trav->next = NULL;
		ReturnRuleSlot(head);
		return(NULL);
		break;

		} /*switch*/

	GetToken(in, YES);		/*Parse a in->token.*/


	} /*while != RPAREN*/

    trav->next = NULL;

    GetToken(in, YES);	/*Parse off the RPAREN*/

    return(head); /*Return the parsed slot*/

} /*ParsePattern*/

/*
 *AddPattern - adds a pattern to he list of patterns for the net.
 */
struct pattern *AddPattern(new)
struct ruleslot *new;
{
    struct pattern *plist;
    struct pattern *past;
    struct ruleslot *rtemp; /*traverses the old pattern*/
    struct ruleslot *ntemp; /*temporary for traversing the new pattern*/
    struct match *pmatch; /*add the pattern to the match list*/

    past = NULL;

    plist = imgr.current_rs->patterns;

    while (plist != NULL) {

	rtemp = plist->pattern; /*take out the next pattern*/

	ntemp = new; /*ntemp goes through the new pattern each time*/

        /*compare the two patterns to see if there the same*/
	while ((rtemp != NULL) && (ntemp != NULL)) {
	    if (rtemp->opcode != ntemp->opcode)
		break;

	    if (rtemp->opcode == NUMERIC) {
		if (rtemp->numeric != ntemp->numeric)
		    break;
		} /*if*/
	    else if (rtemp->string != ntemp->string)
		break;

	    if (rtemp->string != ntemp->string)
		break;

	    rtemp = rtemp->next;
	    ntemp = ntemp->next;

	    } /*while still a list up there*/

	/*If the two patterns were equal then return a pointer to the old one*/
	if ((rtemp == NULL) && (ntemp == NULL)) {

	    pmatch = allocatestruct(match);
	    pmatch->slot = plist;
	    pmatch->next = imgr.pmatchList;
	    imgr.pmatchList = pmatch;
	    return(plist);
	    }

	past = plist;
	plist = plist->next; /*next plist to check out*/
	}  /*while still more patterns*/

    /*If there were no patterns that matched the new one then create one*/
    if (imgr.current_rs->patterns == NULL) 
	past = imgr.current_rs->patterns = allocatestruct(pattern);
    else {
	past->next = allocatestruct(pattern);
	past = past->next;
	}

    past->pattern = new;
    past->outflow = NULL;
    past->next = NULL;

    pmatch = allocatestruct(match);
    pmatch->slot = past;
    pmatch->next = imgr.pmatchList;
    imgr.pmatchList = pmatch;

    return(past); /*new pattern struct created*/

} /*AddPattern*/

/*
 *CreateNet - build the dataflow graph for the LHS of a rule.
 */
struct decidernode *CreateNet(lhs_list, input_side)
struct node *lhs_list;
int input_side;	
{
    int tally;		/*How many input flows have been added*/
    struct decidernode *pnode; /*the present node being worked on*/
    struct lr *plrtemp; /*created side of a node for single pattern rules*/
    struct pattern *ptemp; /*Temp for pattern return*/
    struct values *pvalues; /*The values created for the constant flow:1 input*/

    tally = 0;

    pnode = allocatestruct(decidernode);
    pnode->outflow = NULL;
    pnode->lhs = pnode->rhs = NULL;
    pnode->side =  input_side;

    while (lhs_list != NULL) {

	if (tally == 1)
	    input_side = RHS;

	if (tally > 1) {
	    pnode->outflow = allocatestruct(decidernode);
	    pnode = pnode->outflow;
    	    pnode->outflow = NULL;
	    pnode->lhs = pnode->rhs = NULL;
	    pnode->side = LHS;
	    } /*tally < 1*/

	if (lhs_list->opcode == PATTERN) {
	    ptemp = AddPattern(lhs_list->pattern);

	    AddDecidernode(pnode, ptemp, input_side, '+', lhs_list->string);

	    ReturnNode(lhs_list->right);

	    lhs_list->right = NULL;
	    }
	else
	    EH(NULLSCANNER, "System Error: CreateNet case out of bounds");

	lhs_list = lhs_list->bottom;
	tally++;
	} /*While lhs_list*/

    if (tally == 0)
	EH(NULLSCANNER, "Invalid LHS of rule.");

    if (tally == 1) {
	pvalues = allocatestruct(values);
	pvalues->next = NULL;
	pvalues->whoset = -1;
	pvalues->name = SBNULL;
	pvalues->opcode = SPECIAL;
	pvalues->string = SBAND;

	plrtemp = allocatestruct(lr);
	plrtemp->next = NULL;
	plrtemp->locals = pvalues;
	pnode->rhs = plrtemp;
	} /*if only one pattern on LHS of rule*/

    return(pnode);

} /*CreateNet*/

/*
 *AddDecidernode - add a node from a pattern to the net.
 */
AddDecidernode(pnode, pattern, input_side, boolean, binder)
struct decidernode *pnode; 
struct pattern *pattern;
int input_side;
char boolean;
char *binder;
{
    struct datalink *pdatalink;
    struct values *pvalues;

    if (pattern->outflow == NULL) {
	pattern->outflow = allocatestruct(datalink);
	pdatalink = pattern->outflow;
	}
    else {
	/*We already have flow to nodes, so add another one*/
	pdatalink  = pattern->outflow;

	while (pdatalink->next != NULL)
	    pdatalink = pdatalink->next;

	pdatalink->next = allocatestruct(datalink);
	pdatalink = pdatalink->next;

	} /*else LHS or RHS*/

    /*If the pattern is bound to a variable, then store that information*/
    if (binder != SBNULL) {
 	pvalues = allocatestruct(values);
	pvalues->opcode = SBIND;
	pvalues->string = pvalues->name = binder;
	pvalues->type = WORD;
	pvalues->next = NULL;
	pdatalink->binder = pvalues;
	}
    else
	pdatalink->binder = NULL;

    pdatalink->next = NULL;
    pdatalink->path = pnode;
    pdatalink->side = input_side; /*1 -> LHS; 0 -> RHS; */
    pdatalink->boolean = boolean;

} /*AddDecidernode*/
