/*
 *task.c
 *
 *Some comments:
 * In this particular implementation I start up one process for each RS and
 * EO in this system. Thus I am simulating one PE for each RS and EO. 
 * After the processes are started, then I start up the tasks for each
 * RS and EO which will make it one process for each task. Since all
 * tasks run till system conclusion their is no problem with this.
 *
 * We also create a barrier equal to the total number of RSs and EOs in
 * the system. Each EO/RS task started first does a barrier() call. Once
 * all tasks are started, then everything starts in sync.
 */

#include <parallel.h>
#include <stdio.h>
#include "comm.h"
#include "common.h"

/*
 *StartEOs - start up all of the EOs
 */
StartEOs()
{
    extern int Start_tasks();
    int ct = 0;
    struct defeo *eo;
    struct defeo *eo_in;
    struct defrs *rs;
    struct eo_id *eos;
    struct eo_id *out;

    for (eo=imgr.defeo_lst; eo != NULL; eo = eo->next) {
	ct++;

	for (eos=eo->out_boxes;eos != NULL; eos = eos->next) {

	    if ((eo_in = GetEO_by_name(eos->name)) == NULL) {
		sprintf(error_sb, "Expert object %s missing.", eos->name);
		EH(NULLSCANNER, error_sb);
		return(NO);
		} /*if*/

	    eos->id = eo_in->id;
	    eos->sema = eo_in->recv;
	    }/*for*/

	for (eos = eo->in_boxes; eos != NULL; eos = eos->next) {

	    if ((eo_in = GetEO_by_name(eos->name)) == NULL) {
		sprintf(error_sb, "Expert object %s missing.", eos->name);
		EH(NULLSCANNER, error_sb);
		return(NO);
		} /*if*/

	    for (out=eo_in->out_boxes;out != NULL;out = out->next) 
		if (out->id == eo->id)
		    break;

	    if (out == NULL) {
		EH(NULLSCANNER, "Missing out box.");
		return(NO);
		}

	    eos->box = out->box;
	    eos->id = eo_in->id;
	    } /*for*/

	    
	for (rs=eo->defrs_locals; rs != NULL; rs = rs->next)
	    ct++;

	} /*for*/

    imgr.barrier = barrier_create(ct, TASK_BLOCK);

    printf("Starting %d processes\n", ct);

    task_init( 200000, ct, Start_tasks, 1024, 0);

} /*StartEOs*/

/*
 *Start_tasks
 */
Start_tasks()
{
    struct defeo *eo;
    struct defrs *rs;
    extern void EO_Exec();

    semaphore_init(imgr.print, TASK_BLOCK);
    semaphore_signal(imgr.print);

    for (eo=imgr.defeo_lst; eo != NULL; eo = eo->next) { /*for*/

	for(rs=eo->defrs_locals; rs != NULL; rs = rs->next)
	    rs->eo = eo;

	if (task_start(EO_Exec, 4096, 1, eo->id) == NULL) { 
	    EH(NULLSCANNER, "task_start failure ... se: quit");
	    exit(1);
	    } /*if*/

	} /*for*/

} /*Start_tasks*/
